/*
  This file is a part of bzip2 and/or libbzip2, a program and
  library for lossless, block-sorting data compression.

  Copyright (C) 1996-2000 Julian R Seward.  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.

  2. The origin of this software must not be misrepresented; you must
     not claim that you wrote the original software.  If you use this
     software in a product, an acknowledgment in the product
     documentation would be appreciated but is not required.

  3. Altered source versions must be plainly marked as such, and must
     not be misrepresented as being the original software.

  4. The name of the author may not be used to endorse or promote
     products derived from this software without specific prior written
     permission.

  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  Julian Seward, Cambridge, UK.
  jseward@acm.org
  bzip2/libbzip2 version 1.0 of 21 March 2000

  This program is based on (at least) the work of:
     Mike Burrows
     David Wheeler
     Peter Fenwick
     Alistair Moffat
     Radford Neal
     Ian H. Witten
     Robert Sedgewick
     Jon L. Bentley

  For more information on these sources, see the manual.

  To get some idea how the block sorting algorithms in this file 
  work, read my paper
     On the Performance of BWT Sorting Algorithms
  in Proceedings of the IEEE Data Compression Conference 2000,
  Snowbird, Utah, USA, 27-30 March 2000.  The main sort in this
  file implements the algorithm called  cache  in the paper.
*/

#ifndef _BZLIB_H
#define _BZLIB_H

#include "cputypes.h"

#ifndef BZ_IMPORT
#define BZ_EXPORT
#endif

/* Constants for the back end. */

#define BZ_MAX_ALPHA_SIZE 258
#define BZ_MAX_CODE_LEN    23

#define BZ_RUNA 0
#define BZ_RUNB 1

#define BZ_N_GROUPS 6
#define BZ_G_SIZE   50
#define BZ_N_ITERS  4

#define BZ_MAX_SELECTORS (2 + (900000 / BZ_G_SIZE))

/* States and modes for compression. */

#define BZ_M_IDLE      1
#define BZ_M_RUNNING   2
#define BZ_M_FLUSHING  3
#define BZ_M_FINISHING 4

#define BZ_S_OUTPUT    1
#define BZ_S_INPUT     2

#define BZ_N_RADIX 2
#define BZ_N_QSORT 12
#define BZ_N_SHELL 18
#define BZ_N_OVERSHOOT ( BZ_N_RADIX + BZ_N_QSORT + BZ_N_SHELL + 2 )

#define BZ_LESSER_ICOST  0
#define BZ_GREATER_ICOST 15

#define MAIN_QSORT_SMALL_THRESH 20
#define MAIN_QSORT_DEPTH_THRESH ( BZ_N_RADIX + BZ_N_QSORT )
#define MAIN_QSORT_STACK_SIZE   100

#define FALLBACK_QSORT_SMALL_THRESH 10
#define FALLBACK_QSORT_STACK_SIZE   100

#define BZ_RUN     0
#define BZ_FLUSH   1
#define BZ_FINISH  2

#define BZ_OK                0
#define BZ_RUN_OK            1
#define BZ_FLUSH_OK          2
#define BZ_FINISH_OK         3
#define BZ_STREAM_END        4
#define BZ_SEQUENCE_ERROR    (-1)
#define BZ_PARAM_ERROR       (-2)
#define BZ_MEM_ERROR         (-3)
#define BZ_DATA_ERROR        (-4)
#define BZ_DATA_ERROR_MAGIC  (-5)
#define BZ_IO_ERROR          (-6)
#define BZ_UNEXPECTED_EOF    (-7)
#define BZ_OUTBUFF_FULL      (-8)
#define BZ_CONFIG_ERROR      (-9)


/* States for decompression. */

#define BZ_X_IDLE        1
#define BZ_X_OUTPUT      2

#define BZ_X_MAGIC_1     10
#define BZ_X_MAGIC_2     11
#define BZ_X_MAGIC_3     12
#define BZ_X_MAGIC_4     13
#define BZ_X_BLKHDR_1    14
#define BZ_X_BLKHDR_2    15
#define BZ_X_BLKHDR_3    16
#define BZ_X_BLKHDR_4    17
#define BZ_X_BLKHDR_5    18
#define BZ_X_BLKHDR_6    19
#define BZ_X_BCRC_1      20
#define BZ_X_BCRC_2      21
#define BZ_X_BCRC_3      22
#define BZ_X_BCRC_4      23
#define BZ_X_RANDBIT     24
#define BZ_X_ORIGPTR_1   25
#define BZ_X_ORIGPTR_2   26
#define BZ_X_ORIGPTR_3   27
#define BZ_X_MAPPING_1   28
#define BZ_X_MAPPING_2   29
#define BZ_X_SELECTOR_1  30
#define BZ_X_SELECTOR_2  31
#define BZ_X_SELECTOR_3  32
#define BZ_X_CODING_1    33
#define BZ_X_CODING_2    34
#define BZ_X_CODING_3    35
#define BZ_X_MTF_1       36
#define BZ_X_MTF_2       37
#define BZ_X_MTF_3       38
#define BZ_X_MTF_4       39
#define BZ_X_MTF_5       40
#define BZ_X_MTF_6       41
#define BZ_X_ENDHDR_2    42
#define BZ_X_ENDHDR_3    43
#define BZ_X_ENDHDR_4    44
#define BZ_X_ENDHDR_5    45
#define BZ_X_ENDHDR_6    46
#define BZ_X_CCRC_1      47
#define BZ_X_CCRC_2      48
#define BZ_X_CCRC_3      49
#define BZ_X_CCRC_4      50

/* Constants for the fast MTF decoder. */
#define MTFA_SIZE 4096
#define MTFL_SIZE 16

typedef struct {
  u8 * next_in;
  u32 avail_in;
  u64 total_in;

  u8 * next_out;
  u32 avail_out;
  u64 total_out;

  void * state;
} bz_stream;

/* Structure holding all the compression-side stuff. */

typedef struct
{
  /* pointer back to the struct bz_stream */
  bz_stream * strm;

  /* mode this stream is in, and whether inputting */
  /* or outputting data */
  s32 mode;
  s32 state;

  /* remembers avail_in when flush/finish requested */
  s32 avail_in_expect;

  /* for doing the block sorting */
  u32 * arr1;
  u32 * arr2;
  u32 * ftab;
  s32  origPtr;

  /* aliases for arr1 and arr2 */
  u32 * ptr;
  u8 * block;
  u16 * mtfv;
  u8 * zbits;

  /* for deciding when to use the fallback sorting algorithm */
  s32 workFactor;

  /* run-length-encoding of the input */
  u32 state_in_ch;
  s32 state_in_len;
  s32 rNToGo;
  s32 rTPos;

  /* input and output limits and current posns */
  s32 nblock;
  s32 nblockMAX;
  s32 numZ;
  s32 state_out_pos;

  /* map of bytes used in block */
  s32 nInUse;
  u8 inUse[256];
  u8 unseqToSeq[256];

  /* the buffer for bit stream creation */
  u32 bsBuff;
  s32 bsLive;

  /* block and combined CRCs */
  u32 blockCRC;
  u32 combinedCRC;

  /* misc administratium */
  s32 verbosity;
  s32 blockNo;
  s32 blockSize100k;

  /* stuff for coding the MTF values */
  s32 nMTF;
  s32 mtfFreq[BZ_MAX_ALPHA_SIZE];
  u8 selector[BZ_MAX_SELECTORS];
  u8 selectorMtf[BZ_MAX_SELECTORS];

  u8 len[BZ_N_GROUPS][BZ_MAX_ALPHA_SIZE];
  s32 code[BZ_N_GROUPS][BZ_MAX_ALPHA_SIZE];
  s32 rfreq[BZ_N_GROUPS][BZ_MAX_ALPHA_SIZE];
  /* second dimension: only 3 needed; 4 makes index calculations faster */
  u32 len_pack[BZ_MAX_ALPHA_SIZE][4];
} EState;

/* Structure holding all the decompression-side stuff. */

typedef struct
{
  /* pointer back to the struct bz_stream */
  bz_stream * strm;

  /* state indicator for this stream */
  s32 state;

  /* for doing the final run-length decoding */
  u8 state_out_ch;
  u32 state_out_len;
  u8 blockRandomised;
  s32 rNToGo;
  s32 rTPos;

  /* the buffer for bit stream reading */
  u32 bsBuff;
  s32 bsLive;

  /* misc administratium */
  s32 blockSize100k;
  u8 smallDecompress;
  s32 currBlockNo;
  s32 verbosity;

  /* for undoing the Burrows-Wheeler transform */
  s32 origPtr;
  u32 tPos;
  s32 k0;
  s32 unzftab[256];
  s32 nblock_used;
  s32 cftab[257];
  s32 cftabCopy[257];

  /* for undoing the Burrows-Wheeler transform (FAST) */
  u32 * tt;

  /* for undoing the Burrows-Wheeler transform (SMALL) */
  u16 * ll16;
  u8 * ll4;

  /* stored and calculated CRCs */
  u32 storedBlockCRC;
  u32 storedCombinedCRC;
  u32 calculatedBlockCRC;
  u32 calculatedCombinedCRC;

  /* map of bytes used in block */
  s32 nInUse;
  u8 inUse[256];
  u8 inUse16[16];
  u8 seqToUnseq[256];

  /* for decoding the MTF values */
  u8 mtfa[MTFA_SIZE];
  s32 mtfbase[256 / MTFL_SIZE];
  u8 selector[BZ_MAX_SELECTORS];
  u8 selectorMtf[BZ_MAX_SELECTORS];
  u8 len[BZ_N_GROUPS][BZ_MAX_ALPHA_SIZE];

  s32 limit[BZ_N_GROUPS][BZ_MAX_ALPHA_SIZE];
  s32 base[BZ_N_GROUPS][BZ_MAX_ALPHA_SIZE];
  s32 perm[BZ_N_GROUPS][BZ_MAX_ALPHA_SIZE];
  s32 minLens[BZ_N_GROUPS];

  /* save area for scalars in the main decompress code */
  s32 save_i;
  s32 save_j;
  s32 save_t;
  s32 save_alphaSize;
  s32 save_nGroups;
  s32 save_nSelectors;
  s32 save_EOB;
  s32 save_groupNo;
  s32 save_groupPos;
  s32 save_nextSym;
  s32 save_nblockMAX;
  s32 save_nblock;
  s32 save_es;
  s32 save_N;
  s32 save_curr;
  s32 save_zt;
  s32 save_zn;
  s32 save_zvec;
  s32 save_zj;
  s32 save_gSel;
  s32 save_gMinlen;
  s32 * save_gLimit;
  s32 * save_gBase;
  s32 * save_gPerm;
} DState;

/* full function prototypes here !!! */

/* utility functions */

s32 BZ2_bzCompressInit( bz_stream * strm, s32 blockSize100k, s32 verbosity,
  s32 workFactor );

s32 BZ2_bzCompress( bz_stream * strm, s32 action );

s32 BZ2_bzCompressEnd( bz_stream * strm );

s32 BZ2_bzDecompressInit( bz_stream * strm, s32 verbosity, s32 is_small );

s32 BZ2_bzDecompress( bz_stream * strm );

s32 BZ2_bzDecompressEnd( bz_stream * strm );

/* compression utils */

void BZ2_blockSort( EState * s );

void BZ2_compressBlock( EState * s, u8 is_last_block );

void BZ2_bsInitWrite( EState * s );

void BZ2_hbAssignCodes( s32 *code, u8 *length, s32 minLen, s32 maxLen,
  s32 alphaSize );

void BZ2_hbMakeCodeLengths( u8 * len, s32 * freq, s32 alphaSize, s32 maxLen );

/* decompression utils */

s32 BZ2_indexIntoF( s32 indx, s32 * cftab );

s32 BZ2_decompress( DState * s );

void BZ2_hbCreateDecodeTables( s32 * limit, s32 * base, s32 * perm,
  u8 * length, s32 minLen, s32 maxLen, s32 alphaSize );

/* more utility functions */

void bz_internal_error ( int errcode );

void prepare_new_block( EState * s );

void init_RL( EState * s );

u8 isempty_RL( EState* s );

void add_pair_to_block( EState * s );

void flush_RL( EState * s );

u8 copy_input_until_stop( EState * s );

u8 handle_compress( bz_stream * strm );

void unRLE_obuf_to_output_FAST( DState * s );

void unRLE_obuf_to_output_SMALL( DState * s );

void bsFinishWrite( EState * s );

void bsW( EState * s, s32 n, u32 v );

void bsPutUInt32( EState * s, u32 u );

void bsPutUChar( EState * s, u8 c );

void makeMaps_e( EState * s );

void generateMTFValues( EState * s );

void sendMTFValues( EState * s );

void makeMaps_d( DState * s );

void fallbackSimpleSort( u32 * fmap, u32 * eclass, s32 lo, s32 hi );

void fallbackQSort3( u32 * fmap, u32 * eclass, s32 loSt, s32 hiSt );

void fallbackSort( u32 * fmap, u32 * eclass, u32 * bhtab, s32 nblock,
  s32 verb );

u8 mainGtU( u32 i1, u32 i2, u8 * block, u16 * quadrant, u32 nblock,
  s32 * budget );

void mainSimpleSort( u32 * ptr, u8 * block, u16 * quadrant, s32 nblock,
  s32 lo, s32 hi, s32 d, s32 * budget );

u8 mmed3( u8 a, u8 b, u8 c );

void mainSort( u32 * ptr, u8 * block, u16 * quadrant, u32 * ftab,
  s32 nblock, s32 verb, s32 * budget );

#endif
