/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#ifndef V3_COMPRESS_H
#define V3_COMPRESS_H

#include "cputypes.h"
#include "buffer.h"

#define V3_COMPRESS_BZIP2  173

#define V3_COMPRESS_MODE_COMP    13
#define V3_COMPRESS_MODE_DECOMP  14

#define V3_COMPRESS_STATE_INIT   42  /* Initialized */
#define V3_COMPRESS_STATE_DATA   43  /* Feeding */
#define V3_COMPRESS_STATE_DONE   44  /* Finished */
#define V3_COMPRESS_STATE_ERROR  45  /* Error, data invalid */

#define V3_COMPRESS_ERROR_NONE     0
#define V3_COMPRESS_ERROR_NOMEM    -1  /* out of memory */
#define V3_COMPRESS_ERROR_CORRUPT  -2  /* compressed data corruption */
#define V3_COMPRESS_ERROR_BUFFER   -3  /* buffer too small */
#define V3_COMPRESS_ERROR_PARAM    -4  /* The user passed a NULL pointer */
#define V3_COMPRESS_ERROR_TYPE     -5  /* invalid compression type */
#define V3_COMPRESS_ERROR_STATE    -6  /* invalid state, wrong order */

typedef struct
{
  void * tmp_data;
  u32 state;
  u32 type;
  u32 mode;
  v3_MUTEX lock;
} v3_COMPRESS_TMP;

/* Compression and decompressiong funcations */

s32 v3CompressInit( v3_BUFFER * output, v3_COMPRESS_TMP * tmp,
  u32 type, u32 level );
  /*
    Initialize the output buffer and setup the temporary data needed to
    perform the type of compression specified.
    level is a number ranging from 1 to 9 representing the amount
    of compression to attempt, 1 being fast, 9 being intensive.
    Returns: V3_PASS on success, or an error code on failure.
  */

s32 v3Compress( v3_BUFFER * output, v3_COMPRESS_TMP * tmp,
  const void * const data, u64 length );
  /*
    Feed length bytes of data into the compression routines. Any compressed
    data will be placed into the output buffer but may not be put into
    the output buffer after every call.
    Note: Make sure your data is in big endian format before using.
    Returns: V3_PASS on success, or an error code on failure.
  */

s32 v3CompressEnd( v3_BUFFER * output, v3_COMPRESS_TMP * tmp );
  /*
    Clear any temporary data and put the last of the compressed data into
    the buffer. Make sure to empty and v3BufferFree the output buffer of your
    data after calling this function.
    Returns: V3_PASS on success, or an error code on failure.
  */

s32 v3DecompressInit( v3_BUFFER * output, v3_COMPRESS_TMP * tmp, u32 type );
  /*
    Initialize the output buffer and setup the temporary data needed to
    perform the type of decompression specified.
    Returns: V3_PASS on success, or an error code on failure.
  */

s32 v3Decompress( v3_BUFFER * output, v3_COMPRESS_TMP * tmp,
  const void * const data, u64 length );
  /*
    Feed length bytes of data into the decompression routines. Any
    decompressed data will be placed into the output buffer but may not be
    put into the output buffer after every call.
    Returns: V3_PASS on success, or an error code on failure.
  */

s32 v3DecompressEnd( v3_BUFFER * output, v3_COMPRESS_TMP * tmp );
  /*
    Clear any temporary data and put the last of the decompressed data into
    the buffer. Make sure to empty and v3BufferFree the output buffer of your
    data after calling this function.
    Returns: V3_PASS on success, or an error code on failure.
  */

/* testing */

s32 v3_TestCompress( void );
  /*
    Test functions in this header.
    Returns: V3_PASS on success, or a negative number corresponding to the
      test that failed.
  */

#endif
