/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#include "config.h"
#include "cosmio.h"
#include "cosmfile.h"
#include "cosmmem.h"

s32 v3ConfigLoad( v3_CONFIG * config, const ascii * filename )
{
  v3_FILE file;
  v3_CONFIG_SECTION * tmp_section;
  u64 length;
  ascii * ptr, * eol, * section, * key, * value;
  s32 error;
  u32 strlen;
  u32 i, j;

  if ( config == NULL )
  {
    return( V3_FILE_ERROR_DENIED );
  }

  v3MutexLock( &config->lock, V3_MUTEX_WAIT );

  if ( filename == NULL )
  {
    if ( config->section_count > 0 )
    {
      /* already in use */
      v3MutexUnlock( &config->lock );
      return( V3_FILE_ERROR_DENIED );
    }
    /* fill in struct */
    config->memory = NULL;
    config->sections = NULL;
    config->section_count = 0;
  }
  else
  {
    v3MemSet( &file, v3u64u32( sizeof( v3_FILE ) ), 0 );

    if ( ( error = v3FileOpen( &file, (ascii *) filename, V3_FILE_MODE_READ,
      V3_FILE_LOCK_READ ) ) != V3_PASS )
    {
      v3MutexUnlock( &config->lock );
      return( error );
    }
    if ( ( error = v3FileLength( &length, &file ) ) != V3_PASS )
    {
      v3FileClose( &file );
      v3MutexUnlock( &config->lock );
      return( error );
    }
    config->sections = NULL;
    config->section_count = 0;
    if ( ( config->memory = (ascii *) v3MemAlloc( length, V3_MEM_NORMAL ) )
      == NULL )
    {
      v3FileClose( &file );
      v3MutexUnlock( &config->lock );
      return( V3_FILE_ERROR_DENIED );
    }

    if ( !v3u64Eq( v3FileRead( config->memory, length, &file ), length ) )
    {
      v3FileClose( &file );
      v3MemFree( config->memory );
      v3MutexUnlock( &config->lock );
      return( V3_FILE_ERROR_DENIED );
    }
    v3FileClose( &file );

    /* parse */
    ptr = config->memory;
    while ( ( section = v3StrCharA( ptr, '[', length ) ) != NULL )
    {
      section++;
      ptr = v3StrCharA( ptr, ']', length );
      *ptr = 0;
      ptr++;
      ptr++;

      if ( config->section_count == 0 )
      {
        if ( ( config->sections = (v3_CONFIG_SECTION *)
          v3MemAlloc( v3u64u32( sizeof( v3_CONFIG_SECTION ) ),
          V3_MEM_NORMAL ) ) == NULL )
        {
          v3MutexUnlock( &config->lock );
          return( V3_FILE_ERROR_DENIED );
        }
        config->section_count = 1;
      }
      else
      {
        if ( ( config->sections = (v3_CONFIG_SECTION *)
          v3MemRealloc( config->sections, v3u64u32(
          ( config->section_count + 1 ) * sizeof( v3_CONFIG_SECTION ) ),
          V3_MEM_NORMAL ) ) == NULL )
        {
          v3MutexUnlock( &config->lock );
          return( V3_FILE_ERROR_DENIED );
        }
        v3MemSet( &config->sections[config->section_count],
          v3u64u32( sizeof( v3_CONFIG_SECTION ) ), 0 );
        config->section_count += 1;
      }
      i = config->section_count - 1;
      config->sections[i].flag = V3_CONFIG_NORMAL;
      config->sections[i].length =
        v3u32u64( v3u64Add( v3StrLengthA( section ), v3u64u32( 1 ) ) );
      config->sections[i].section = section;
      config->sections[i].key_count = 0;
      config->sections[i].keys = NULL;

      /* now add the keys */
      tmp_section = &config->sections[i];
      while ( *ptr != '[' )
      {
        if ( ( eol = v3StrCharA( ptr, '\n', length ) ) == NULL )
        {
          /* no more lines */
          break;
        }
        *eol = 0;
        eol++;
        if ( ( value = v3StrCharA( ptr, '=', length ) ) != NULL )
        {
          key = ptr;
          *value = 0;
          value++;
          ptr = eol;

          /* ignore empty lines */
          while ( *ptr == '\n' )
          {
            ptr++;
          }

          if ( tmp_section->key_count == 0 )
          {
            if ( ( tmp_section->keys = (v3_CONFIG_KEY *)
              v3MemAlloc( v3u64u32( sizeof( v3_CONFIG_KEY ) ),
              V3_MEM_NORMAL ) ) == NULL )
            {
              v3MutexUnlock( &config->lock );
              return( V3_FILE_ERROR_DENIED );
            }
            tmp_section->key_count = 1;
          }
          else
          {
            if ( ( tmp_section->keys = (v3_CONFIG_KEY *)
              v3MemRealloc( tmp_section->keys, v3u64u32(
              ( tmp_section->key_count + 1 ) * sizeof( v3_CONFIG_KEY ) ),
              V3_MEM_NORMAL ) ) == NULL )
            {
              v3MutexUnlock( &config->lock );
              return( V3_FILE_ERROR_DENIED );
            }
            v3MemSet( &tmp_section->keys[tmp_section->key_count],
              v3u64u32( sizeof( v3_CONFIG_KEY ) ), 0 );
            tmp_section->key_count += 1;
          }

          j = tmp_section->key_count - 1;

          strlen = v3u32u64( v3u64Add( v3StrLengthA( key ),
            v3StrLengthA( value ) ) );
          strlen += 2;

          if ( (tmp_section->keys[j].key = (ascii *)
            v3MemAlloc( v3u64u32( strlen ), V3_MEM_NORMAL ) ) == NULL )
          {
            v3MutexUnlock( &config->lock );
            return( V3_FILE_ERROR_DENIED );
          }
          tmp_section->keys[j].key = key;
          tmp_section->keys[j].value = value;
          tmp_section->keys[j].length = strlen;
          tmp_section->keys[j].flag = V3_CONFIG_NORMAL;
        } /* if key/value pair */
      } /* while section */
    } /* while more to file */
  }

  v3MutexUnlock( &config->lock );

  return( V3_PASS );
}

s32 v3ConfigSave( const v3_CONFIG * config, const ascii * filename )
{
  u32 i, j, flag;
  v3_CONFIG_SECTION * tmp_section;
  v3_FILE file;
  s32 error;

  if ( ( config == NULL ) || ( filename == NULL ) )
  {
    return( V3_FILE_ERROR_DENIED );
  }

  v3MemSet( &file, v3u64u32( sizeof( v3_FILE ) ), 0 );

  if ( ( error = v3FileOpen( &file, (ascii *) filename, V3_FILE_MODE_WRITE
    | V3_FILE_MODE_CREATE | V3_FILE_MODE_TRUNCATE, V3_FILE_LOCK_WRITE ) )
    != V3_PASS )
  {
    return( error );
  }

  v3MutexLock( (v3_MUTEX *) &config->lock, V3_MUTEX_WAIT );

  for ( i = 0 ; i < config->section_count; i++ )
  {
    if ( ( config->sections[i].flag == V3_CONFIG_NORMAL )
      || ( config->sections[i].flag == V3_CONFIG_ALLOCATED ) )
    {
      tmp_section = &config->sections[i];
      flag = 0;
      for ( j = 0 ; j < tmp_section->key_count; j++ )
      {
        if ( ( tmp_section->keys[j].flag == V3_CONFIG_NORMAL )
          || ( tmp_section->keys[j].flag == V3_CONFIG_ALLOCATED ) )
        {
          if ( flag == 0 )
          {
            /* this section contains active keys, write header */
            v3PrintAFile( &file, (ascii *) "[%.*s]\n",
              v3u32u64( v3StrLengthA( tmp_section->section ) ),
              tmp_section->section );
            flag = 1;
          }
          /* write the key/value */
          v3PrintAFile( &file, (ascii *) "%.*s=%.*s\n",
            v3u32u64( v3StrLengthA( tmp_section->keys[j].key ) ),
            tmp_section->keys[j].key,
            v3u32u64( v3StrLengthA( tmp_section->keys[j].value ) ),
            tmp_section->keys[j].value );
        }
      }
      if ( flag == 1 )
      {
        v3PrintAFile( &file, (ascii *) "\n" );
      }
    }
  }

  v3FileClose( &file );
  v3MutexUnlock( (v3_MUTEX *) &config->lock );

  return( V3_PASS );
}

s32 v3ConfigSet( v3_CONFIG * config, const ascii * section,
  const ascii * key, const ascii * value )
{
  u32 i, j;
  v3_CONFIG_SECTION * tmp_section;
  u32 length;
  u64 key_len, val_len;

  if ( ( config == NULL ) || ( section == NULL ) || ( key == NULL ) )
  {
    return( V3_FAIL );
  }

  /* calculate length of key+value + 2 \0 */
  key_len = v3u64Add( v3StrLengthA( key ), v3u64u32( 1 ) );
  val_len = v3u64Add( v3StrLengthA( value ), v3u64u32( 1 ) );
  length = v3u32u64( v3u64Add( key_len, val_len ) );

  v3MutexLock( &config->lock, V3_MUTEX_WAIT );

  for ( i = 0 ; i < config->section_count; i++ )
  {
    if ( v3StrCmpA( section, config->sections[i].section,
      v3StrLengthA( section ) ) == 0 )
    {
      /* found the section */
      tmp_section = &config->sections[i];
      for ( j = 0 ; j < tmp_section->key_count; j++ )
      {
        if ( v3StrCmpA( key, tmp_section->keys[j].key,
          v3StrLengthA( key ) ) == 0 )
        {
          /* found an existing key */
          if ( value == NULL )
          {
            /* mark deleted */
            if ( tmp_section->keys[j].flag == V3_CONFIG_ALLOCATED )
            {
              tmp_section->keys[j].length = 0;
              v3MemFree( tmp_section->keys[j].key );
              tmp_section->keys[j].key = NULL;
            }
            tmp_section->keys[j].flag = V3_CONFIG_DELETED;
          }
          else if ( length < tmp_section->keys[j].length )
          {
            /* big enough, copy over */
            v3StrCopyA( tmp_section->keys[j].value, value, val_len );
            if ( tmp_section->keys[j].flag == V3_CONFIG_DELETED )
            {
              tmp_section->keys[j].flag = V3_CONFIG_NORMAL;
            }
          }
          else if ( tmp_section->keys[j].length > 0 )
          {
            /* expand */
            if ( ( tmp_section->keys[j].flag == V3_CONFIG_NORMAL )
              || ( tmp_section->keys[j].flag == V3_CONFIG_DELETED ) )
            {
              /* do allocate, copy key and value */
              if ( ( tmp_section->keys[j].key =
                (ascii *) v3MemAlloc( v3u64u32( length ),
                V3_MEM_NORMAL ) ) == NULL )
              {
                v3MutexUnlock( &config->lock );
                return( V3_FAIL );
              }
              v3StrCopyA( tmp_section->keys[j].key, key, key_len );
              tmp_section->keys[j].value =
                (ascii *) v3MemOffset( tmp_section->keys[j].key, key_len );
              v3StrCopyA( tmp_section->keys[j].value, value, val_len );
              tmp_section->keys[j].flag = V3_CONFIG_ALLOCATED;
              tmp_section->keys[j].length = length;
            }
            else
            {
              /* realloc bigger, fix * value, then copy new value */
              if ( ( tmp_section->keys[j].key = v3MemRealloc(
                tmp_section->keys[j].key, v3u64u32( length ),
                V3_MEM_NORMAL ) ) == NULL )
              {
                v3MutexUnlock( &config->lock );
                return( V3_FAIL );
              }
              tmp_section->keys[j].value = (ascii *) v3MemOffset(
                tmp_section->keys[j].key, key_len );
              v3StrCopyA( tmp_section->keys[j].value, value, val_len );
            }
          }
          else
          {
            /* alloc'd then free'd, create new */
            if ( ( tmp_section->keys[j].key =
              (ascii *) v3MemAlloc( v3u64u32( length ),
              V3_MEM_NORMAL ) ) == NULL )
            {
              v3MutexUnlock( &config->lock );
              return( V3_FAIL );
            }
            v3StrCopyA( tmp_section->keys[j].key, key, key_len );
            tmp_section->keys[j].value =
              (ascii *) v3MemOffset( tmp_section->keys[j].key, key_len );
            v3StrCopyA( tmp_section->keys[j].value, value, val_len );
            tmp_section->keys[j].length = length;
            tmp_section->keys[j].flag = V3_CONFIG_ALLOCATED;
          }
          v3MutexUnlock( &config->lock );
          return( V3_PASS );
        } /* key matched */
      } /* section matched */
      /* section exists, key does not, alloc a new key */
      if ( tmp_section->key_count == 0 )
      {
        if ( ( tmp_section->keys = (v3_CONFIG_KEY *) v3MemAlloc(
          v3u64u32( sizeof( v3_CONFIG_KEY ) ), V3_MEM_NORMAL ) ) == NULL )
        {
          v3MutexUnlock( &config->lock );
          return( V3_FAIL );
        }
        tmp_section->key_count = 1;
      }
      else
      {
        if ( ( tmp_section->keys = (v3_CONFIG_KEY *)
          v3MemRealloc( tmp_section->keys, v3u64u32(
          ( tmp_section->key_count + 1 ) * sizeof( v3_CONFIG_KEY ) ),
          V3_MEM_NORMAL ) ) == NULL )
        {
          v3MutexUnlock( &config->lock );
          return( V3_FAIL );
        }
        v3MemSet( &tmp_section->keys[tmp_section->key_count],
          v3u64u32( sizeof( v3_CONFIG_KEY ) ), 0 );
        tmp_section->key_count += 1;
      }
      /* make another key struct and add it in */
      j = tmp_section->key_count - 1;
      tmp_section->keys[j].length = length;
      if ( ( tmp_section->keys[j].key =
        (ascii *) v3MemAlloc( v3u64u32( length ), V3_MEM_NORMAL ) ) == NULL )
      {
        tmp_section->key_count -= 1;
        v3MutexUnlock( &config->lock );
        return( V3_FAIL );
      }
      v3StrCopyA( tmp_section->keys[j].key, key, key_len );
      tmp_section->keys[j].value =
        (ascii *) v3MemOffset( tmp_section->keys[j].key, key_len );
      v3StrCopyA( tmp_section->keys[j].value, value, val_len );
      tmp_section->keys[j].flag = V3_CONFIG_ALLOCATED;

      v3MutexUnlock( &config->lock );
      return( V3_PASS );
    } /* section matched */
  } /* section loop */
  /* section does not exist, alloc a new one and create a key */
  if ( config->section_count == 0 )
  {
    if ( ( config->sections = (v3_CONFIG_SECTION *)
      v3MemAlloc( v3u64u32( sizeof( v3_CONFIG_SECTION ) ),
      V3_MEM_NORMAL ) ) == NULL )
    {
      v3MutexUnlock( &config->lock );
      return( V3_FAIL );
    }
    config->section_count = 1;
  }
  else
  {
    if ( ( config->sections = (v3_CONFIG_SECTION *)
      v3MemRealloc( config->sections, v3u64u32(
      ( config->section_count + 1 ) * sizeof( v3_CONFIG_SECTION ) ),
      V3_MEM_NORMAL ) ) == NULL )
    {
      v3MutexUnlock( &config->lock );
      return( V3_FAIL );
    }
    v3MemSet( &config->sections[config->section_count],
      v3u64u32( sizeof( v3_CONFIG_SECTION ) ), 0 );
    config->section_count += 1;
  }
  j = config->section_count - 1;
  config->sections[j].flag = V3_CONFIG_ALLOCATED;
  config->sections[j].length = v3u32u64( v3u64Add( v3StrLengthA( section ),
    v3u64u32( 1 ) ) );
  if ( ( config->sections[j].section = (ascii *) v3MemAlloc(
    v3u64u32( config->sections[j].length ), V3_MEM_NORMAL ) ) == NULL )
  {
    v3MutexUnlock( &config->lock );
    return( V3_FAIL );
  }
  v3StrCopyA( config->sections[j].section, section,
    v3u64u32( config->sections[j].length ) );
  if ( ( config->sections[j].keys = (v3_CONFIG_KEY *) v3MemAlloc(
    v3u64u32( sizeof( v3_CONFIG_KEY ) ), V3_MEM_NORMAL ) ) == NULL )
  {
    v3MutexUnlock( &config->lock );
    return( V3_FAIL );
  }
  config->sections[j].key_count = 1;

  tmp_section = &config->sections[j];
  tmp_section->keys[0].length = length;
  if ( ( tmp_section->keys[0].key =
    (ascii *) v3MemAlloc( v3u64u32( length ), V3_MEM_NORMAL ) ) == NULL )
  {
    v3MutexUnlock( &config->lock );
    return( V3_FAIL );
  }
  v3StrCopyA( tmp_section->keys[0].key, key, key_len );
  tmp_section->keys[0].value =
    (ascii *) v3MemOffset( tmp_section->keys[0].key, key_len );
  v3StrCopyA( tmp_section->keys[0].value, value, val_len );
  tmp_section->keys[0].flag = V3_CONFIG_ALLOCATED;

  v3MutexUnlock( &config->lock );
  return( V3_PASS );
}

const ascii * v3ConfigGet( v3_CONFIG * config, const ascii * section,
  const ascii * key )
{
  u32 i, j;
  v3_CONFIG_SECTION * tmp_section;

  if ( config == NULL )
  {
    return( NULL );
  }

  v3MutexLock( &config->lock, V3_MUTEX_WAIT );

  for ( i = 0 ; i < config->section_count; i++ )
  {
    if ( v3StrCmpA( section, config->sections[i].section,
      v3StrLengthA( section ) ) == 0 )
    {
      /* found the section */
      tmp_section = &config->sections[i];
      for ( j = 0 ; j < tmp_section->key_count; j++ )
      {
        if ( ( v3StrCmpA( key, tmp_section->keys[j].key, v3StrLengthA( key ) )
          == 0 ) && ( ( tmp_section->keys[j].flag == V3_CONFIG_NORMAL )
          || ( tmp_section->keys[j].flag == V3_CONFIG_ALLOCATED ) ) )
        {
          v3MutexUnlock( &config->lock );
          return( tmp_section->keys[j].value );
        }
      }
    }
  }

  v3MutexUnlock( &config->lock );
  return( NULL );
}

void v3ConfigFree( v3_CONFIG * config )
{
  u32 i, j;
  v3_CONFIG_SECTION * section;

  if ( config == NULL )
  {
    return;
  }

  v3MutexLock( &config->lock, V3_MUTEX_WAIT );

  for ( i = 0 ; i < config->section_count; i++ )
  {
    section = &config->sections[i];
    for ( j = 0 ; j < section->key_count; j++ )
    {
      if ( section->keys[j].flag == V3_CONFIG_ALLOCATED )
      {
        v3MemFree( section->keys[j].key );
      }
    }
    if ( section->flag == V3_CONFIG_ALLOCATED )
    {
      v3MemFree( section->section );
    }
    v3MemFree( section->keys );
  }
  v3MemFree( config->sections );
  v3MemFree( config->memory );

  config->memory = NULL;
  config->sections = NULL;
  config->section_count = 0;

  v3MutexUnlock( &config->lock );
  v3MutexFree( &config->lock );
}

s32 v3_TestConfig( void )
{
  v3_CONFIG conf;
  const ascii * ptr;

  v3MemSet( &conf, v3u64u32( sizeof( v3_CONFIG ) ), 0 );

  if ( v3ConfigLoad( &conf, NULL ) != V3_PASS )
  {
    return( -1 );
  }

  /* set some keys */
  if ( v3ConfigSet( &conf, (ascii *) "s1", (ascii *) "k1",
    (ascii *) "v11" ) != V3_PASS )
  {
    return( -2 );
  }
  if ( v3ConfigSet( &conf, (ascii *) "s1", (ascii *) "k2",
    (ascii *) "v12" ) != V3_PASS )
  {
    return( -3 );
  }
  if ( v3ConfigSet( &conf, (ascii *) "s2", (ascii *) "k1",
    (ascii *) "v21" ) != V3_PASS )
  {
    return( -4 );
  }
  if ( v3ConfigSet( &conf, (ascii *) "s2", (ascii *) "k2",
    (ascii *) "v22" ) != V3_PASS )
  {
    return( -5 );
  }

  /* check em */
  ptr = v3ConfigGet( &conf, (ascii *) "s1", (ascii *) "k1" );
  if ( v3StrCmpA( ptr, (ascii *) "v11", v3u64u32( 4 ) ) )
  {
    return( -6 );
  }
  ptr = v3ConfigGet( &conf, (ascii *) "s1", (ascii *) "k2" );
  if ( v3StrCmpA( ptr, (ascii *) "v12", v3u64u32( 4 ) ) )
  {
    return( -7 );
  }
  ptr = v3ConfigGet( &conf, (ascii *) "s2", (ascii *) "k1" );
  if ( v3StrCmpA( ptr, (ascii *) "v21", v3u64u32( 4 ) ) )
  {
    return( -8 );
  }
  ptr = v3ConfigGet( &conf, (ascii *) "s2", (ascii *) "k2" );
  if ( v3StrCmpA( ptr, (ascii *) "v22", v3u64u32( 4 ) ) )
  {
    return( -9 );
  }

  /* change and check some */
  if ( v3ConfigSet( &conf, (ascii *) "s1", (ascii *) "k1",
    (ascii *) "v11longer" ) != V3_PASS )
  {
    return( -10 );
  }
  ptr = v3ConfigGet( &conf, (ascii *) "s1", (ascii *) "k1" );
  if ( v3StrCmpA( ptr, (ascii *) "v11longer", v3u64u32( 10 ) ) )
  {
    return( -11 );
  }
  if ( v3ConfigSet( &conf, (ascii *) "s2", (ascii *) "k2",
    (ascii *) "v" ) != V3_PASS )
  {
    return( -12 );
  }
  ptr = v3ConfigGet( &conf, (ascii *) "s2", (ascii *) "k2" );
  if ( v3StrCmpA( ptr, (ascii *) "v", v3u64u32( 2 ) ) )
  {
    return( -13 );
  }

  /* save it and free */
  if ( v3ConfigSave( &conf, (ascii *) "test.cfg" ) != V3_PASS )
  {
    return( -14 );
  }
  v3ConfigFree( &conf );

  /* load it again (and delete the file) */
  if ( v3ConfigLoad( &conf, (ascii *) "test.cfg" ) != V3_PASS )
  {
    return( -15 );
  }
  v3FileDelete( (ascii *) "test.cfg" );

  /* check the loaded version and free it */
  ptr = v3ConfigGet( &conf, (ascii *) "s1", (ascii *) "k1" );
  if ( v3StrCmpA( ptr, (ascii *) "v11longer", v3u64u32( 10 ) ) )
  {
    return( -16 );
  }
  ptr = v3ConfigGet( &conf, (ascii *) "s1", (ascii *) "k2" );
  if ( v3StrCmpA( ptr, (ascii *) "v12", v3u64u32( 4 ) ) )
  {
    return( -17 );
  }
  ptr = v3ConfigGet( &conf, (ascii *) "s2", (ascii *) "k1" );
  if ( v3StrCmpA( ptr, (ascii *) "v21", v3u64u32( 4 ) ) )
  {
    return( -18 );
  }
  ptr = v3ConfigGet( &conf, (ascii *) "s2", (ascii *) "k2" );
  if ( v3StrCmpA( ptr, (ascii *) "v", v3u64u32( 2 ) ) )
  {
    return( -19 );
  }

  /* change some loaded ones */
  if ( v3ConfigSet( &conf, (ascii *) "s1", (ascii *) "k1",
    (ascii *) "v11" ) != V3_PASS )
  {
    return( -20 );
  }
  ptr = v3ConfigGet( &conf, (ascii *) "s1", (ascii *) "k1" );
  if ( v3StrCmpA( ptr, (ascii *) "v11", v3u64u32( 4 ) ) )
  {
    return( -21 );
  }
  if ( v3ConfigSet( &conf, (ascii *) "s2", (ascii *) "k2",
    (ascii *) "v22" ) != V3_PASS )
  {
    return( -22 );
  }
  ptr = v3ConfigGet( &conf, (ascii *) "s2", (ascii *) "k2" );
  if ( v3StrCmpA( ptr, (ascii *) "v22", v3u64u32( 4 ) ) )
  {
    return( -23 );
  }

  /* delete a key */
  if ( v3ConfigSet( &conf, (ascii *) "s1", (ascii *) "k1", NULL ) != V3_PASS )
  {
    return( -24 );
  }
  ptr = v3ConfigGet( &conf, (ascii *) "s1", (ascii *) "k1" );
  if ( ptr != NULL )
  {
    return( -25 );
  }

  v3ConfigFree( &conf );

  return( V3_PASS );
}
