/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#ifndef V3_CONFIG_H
#define V3_CONFIG_H

#include "cputypes.h"
#include "cosmtask.h"

/*
  Dealing with config files

<begin file>
[section1]\n
key1=value1\n
key2=value2\n
\n
[section2]\n
key1=value1\n
key2=value2\n
\n
</end file>

The Rules:
  These are binary files, not text. You're not supposed to have to edit them.
  Config files contain ascii, not unicode.
  No linefeeds (\n) in section/keys/values - those are line separators.
  No '=' or '[' in keys, and avoid ']'.
  Everything is case sensitive.
  Blank lines are ignored, but added at the end of sections for humans.
  Any improperly formatted line is discarded.
  Comment lines don't exist, that's what documentation is for.
  The majority of cases the config will only be read, so optimize for
  reading.
*/

#define V3_CONFIG_DELETED   0  /* Key deleted */
#define V3_CONFIG_NORMAL    1  /* Part of main memory block */
#define V3_CONFIG_ALLOCATED 2  /* Allocated outside the main block */

typedef struct
{
  /* Key/value pairs are done with one alloc/free for both when needed */
  u32 length; /* length of space allocated for key+value and 2 \0 */
  u32 flag;
  ascii * key;
  ascii * value;
} v3_CONFIG_KEY;

typedef struct
{
  u32 length; /* length of space for section name and \0 */
  u32 flag;
  ascii * section;
  v3_CONFIG_KEY * keys;
  u32 key_count;
} v3_CONFIG_SECTION;

typedef struct
{
  ascii * memory;
  v3_CONFIG_SECTION * sections;
  u32 section_count;
  v3_MUTEX lock;
} v3_CONFIG;

/* Config Functions */

s32 v3ConfigLoad( v3_CONFIG * config, const ascii * filename );
  /*
    Attempt to open the filename and read in the config data.
    If filename is NULL, create an empty config.
    Returns: V3_PASS on success, or a V3_FILE_ERROR_* on failure.
  */

s32 v3ConfigSave( const v3_CONFIG * config, const ascii * filename );
  /*
    Write out the config data to the file.
    Returns: V3_PASS on success, or a V3_FILE_ERROR_* on failure.
  */

s32 v3ConfigSet( v3_CONFIG * config, const ascii * section,
  const ascii * key, const ascii * value );
  /*
    Set the value for the section/key pair. If value is NULL then key is
    deleted.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

const ascii * v3ConfigGet( v3_CONFIG * config, const ascii * section,
  const ascii * key );
  /*
    Get the value of the section/key pair. The returned pointer will
    no longer be valid once another v3Config* call is made.
    Returns: A pointer to the value on success, or NULL on failure.
  */

void v3ConfigFree( v3_CONFIG * config );
  /*
    Free the internal config data.
    Returns: nothing.
  */

/* testing */

s32 v3_TestConfig( void );
  /*
    Test functions in this header.
    Returns: V3_PASS on success, or a negative number corresponding to the
      test that failed.
  */

#endif
