/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

/* CPU/OS Layer - CPU and OS specific code is allowed */

#include "cosmmath.h"
#include "cosmfile.h"
#include "cosmtask.h"
#include "cosmmem.h"
#include "cosmio.h"

#include <errno.h>
#include <stdio.h>      /* for remove */
#include <string.h>

#if ( OS_TYPE == OS_MACOS )
#include <Files.h>
#include <Errors.h>
#include <Script.h>
#else
#include <sys/stat.h>   /* for stat and open */
#include <fcntl.h>      /* for open */
#include <sys/types.h>  /* for open and lseek */
#endif

#if ( OS_TYPE == OS_OS2 )
#define INCL_DOSFILEMGR
#define INCL_DOSERRORS
#include <os2.h>
#endif

#define V3_FILE_PATH_UNIX 0
#define V3_FILE_PATH_DOS  1
#define V3_FILE_PATH_MAC  2

#if ( OS_TYPE == OS_WIN32 )
#include <io.h>
#define V3_FILE_PATHMODE V3_FILE_PATH_DOS
#elif ( OS_TYPE == OS_OS2 )
#include <io.h>
#include <sys/stat.h>
#include <direct.h>
#define V3_FILE_PATHMODE V3_FILE_PATH_DOS
#elif ( OS_TYPE == OS_MACOS )
#define V3_FILE_PATHMODE V3_FILE_PATH_MAC
#else /* unixes */
#include <unistd.h>    /* for stat, close, read, write, lseek, ftruncate */
#include <sys/file.h>  /* for flock */
#define V3_FILE_PATHMODE V3_FILE_PATH_UNIX
#endif

/* Setup the size of file handle ( not the max size of a file ) */
#if ( OS_TYPE == OS_TRU64 )
  /* int, off_t, size_t ans ssize_t are 64 bits */
#define V3_FILE64
#else
  /* int, off_t, size_t ans ssize_t are 32 bits */
#undef V3_FILE64
#endif

s32 v3FileOpen( v3_FILE * file, const ascii * filename, u32 mode, u32 lock )
{
#if ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  FSSpec spec;
  short file_ref_num;
#else
  struct stat file_stats;
#endif
  v3_FILENAME native_filename;
  int open_flags;
  int open_mode;
  int open_result;
  s32 mutex_protected;

  if ( file == NULL )
  {
    /* Can't set last_error */
    return( V3_FILE_ERROR_NOTFOUND );
  }

  v3MutexLock( &file->lock, V3_MUTEX_WAIT );

  if ( file->status != V3_FILE_STATUS_CLOSED )
  {
    v3MutexUnlock( &file->lock );
    return( V3_FILE_ERROR_CLOSED );
  }

  /* Check if mode is valid */
  if ( mode & ~( V3_FILE_MODE_EXIST | V3_FILE_MODE_READ | V3_FILE_MODE_WRITE |
    V3_FILE_MODE_APPEND | V3_FILE_MODE_CREATE | V3_FILE_MODE_TRUNCATE |
    V3_FILE_MODE_SYNC ) )
  {
    file->last_error = V3_FILE_ERROR_MODE;
    v3MutexUnlock( &file->lock );
    return( V3_FILE_ERROR_MODE );
  }

  /* Check if lock is valid */
  if ( lock & ~( V3_FILE_LOCK_NONE | V3_FILE_LOCK_READ |
    V3_FILE_LOCK_WRITE ) )
  {
    file->last_error = V3_FILE_ERROR_MODE;
    v3MutexUnlock( &file->lock );
    return( V3_FILE_ERROR_MODE );
  }

  if ( ( ( lock & V3_FILE_LOCK_READ ) == V3_FILE_LOCK_READ ) &&
    ( ( mode & V3_FILE_MODE_WRITE ) == V3_FILE_MODE_WRITE ) )
  {
    /* invalid lock/mode combination */
    file->last_error = V3_FILE_ERROR_MODE;
    v3MutexUnlock( &file->lock );
    return( V3_FILE_ERROR_MODE );
  }

  if ( ( ( mode & V3_FILE_MODE_WRITE ) != V3_FILE_MODE_WRITE ) &&
    ( ( mode & V3_FILE_MODE_TRUNCATE ) == V3_FILE_MODE_TRUNCATE ) )
  {
    /* must be writing to truncate */
    file->last_error = V3_FILE_ERROR_MODE;
    v3MutexUnlock( &file->lock );
    return( V3_FILE_ERROR_MODE );
  }

  /* translate filename */
  if ( v3_FileNativePath( native_filename, filename ) != V3_PASS )
  {
    file->last_error = V3_FILE_ERROR_NAME;
    v3MutexUnlock( &file->lock );
    return( V3_FILE_ERROR_NAME );
  }

  /* Save filename, mode, and locks */
  v3StrCopyA( file->filename, filename,
    v3u64u32( V3_FILE_MAX_FILENAME + 1 ) );
  file->mode = mode;
  file->lockmode = lock;

  /* check for existance only */
  if ( ( mode & V3_FILE_MODE_EXIST ) == V3_FILE_MODE_EXIST )
  {
#if ( OS_TYPE == OS_MACOS )
    /* We try to make an FSSpec ... returns with fnfErr if doesn't exist */
    if ( ( my_err = FSMakeFSSpec( 0, 0, native_filename, &spec ) ) != noErr )
    {
      if ( my_err == fnfErr )
      {
        file->last_error = V3_FILE_ERROR_NOTFOUND;
        v3MutexUnlock( &file->lock );
        return( V3_FILE_ERROR_NOTFOUND );
      }
      /* other error */
      file->last_error = V3_FILE_ERROR_DENIED;
      v3MutexUnlock( &file->lock );
      return( V3_FILE_ERROR_DENIED );
    }
#else
    /* We are using stat to test file existence */
    if ( stat( (const char *) file->filename, &file_stats ) == -1 )
    {
      if ( errno == ENOENT )
      {
        file->last_error = V3_FILE_ERROR_NOTFOUND;
        v3MutexUnlock( &file->lock );
        return( V3_FILE_ERROR_NOTFOUND );
      }
      /* another error as occured */
      file->last_error = V3_FILE_ERROR_DENIED;
      v3MutexUnlock( &file->lock );
      return( V3_FILE_ERROR_DENIED );
    }
#endif
    file->last_error = V3_FILE_ERROR_NONE;
    v3MutexUnlock( &file->lock );
    return( V3_PASS );
  }

  /* convert mode to native value */
  if ( ( mode & V3_FILE_MODE_READ ) == V3_FILE_MODE_READ )
  {
    if ( ( ( mode & V3_FILE_MODE_WRITE ) == V3_FILE_MODE_WRITE ) ||
      ( ( mode & V3_FILE_MODE_APPEND ) == V3_FILE_MODE_APPEND ) )
    {
#if ( OS_TYPE == OS_MACOS )
      open_flags = fsRdWrPerm;
#else
      open_flags = O_RDWR;
#endif
    }
    else
    {
#if ( OS_TYPE == OS_MACOS )
      open_flags = fsRdPerm;
#else
      open_flags = O_RDONLY;
#endif
    }
  }
  else
  {
    if ( ( ( mode & V3_FILE_MODE_WRITE ) == V3_FILE_MODE_WRITE ) ||
       ( ( mode & V3_FILE_MODE_APPEND ) == V3_FILE_MODE_APPEND ) )
    {
#if ( OS_TYPE == OS_MACOS )
      open_flags = fsWrPerm;
#else
      open_flags = O_WRONLY;
#endif
    }
    else
    {
      file->last_error = V3_FILE_ERROR_DENIED;
      v3MutexUnlock( &file->lock );
      return( V3_FILE_ERROR_MODE );
    }
  }

#if ( OS_TYPE != OS_MACOS )
  /* Check any other flags */
  if ( ( mode & V3_FILE_MODE_APPEND ) == V3_FILE_MODE_APPEND )
  {
    open_flags = open_flags | O_APPEND;
  }
#endif
  open_mode = 0;

  if ( ( mode & V3_FILE_MODE_CREATE ) == V3_FILE_MODE_CREATE )
  {
#if ( OS_TYPE == OS_MACOS )
    my_err = FSMakeFSSpec( 0, 0, file->filename, &spec );
    if ( my_err == noErr )
    {
      my_err = FSpCreate( &spec, 'Cosm', 'Cosm', smSystemScript );
      if ( my_err != noErr )
      {
        v3MutexUnlock( &file->lock );
        return( V3_FILE_ERROR_CREATE );
      }
    }
#else
    open_flags = open_flags | O_CREAT;
#if ( defined( S_IRGRP ) )
    /* platform handling user groups */
    open_mode = S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP;
#else
    /* platform not handling user groups -> use old style */
    open_mode = S_IREAD | S_IWRITE | S_IEXEC;
#endif
#endif  /* OS */
  }

#if ( OS_TYPE != OS_MACOS )
#if ( defined( O_SYNC ) )
  if ( ( mode & V3_FILE_MODE_SYNC ) == V3_FILE_MODE_SYNC )
  {
    open_flags = open_flags | O_SYNC;
  }
#endif
#endif

#if ( OS_TYPE == OS_MACOS )
  /* Really try to open the file */
  my_err = FSMakeFSSpec( 0, 0, native_filename, &spec );
  if ( my_err == noErr )
  {
    my_err = FSpOpenDF( &spec, open_mode, &file_ref_num );
    if ( my_err != noErr )
    {
      switch( my_err )
      {
        case opWrErr:
        case permErr:
          file->last_error = V3_FILE_ERROR_WRITEMODE;
          break;
        case afpAccessDenied:
          file->last_error = V3_FILE_ERROR_DENIED;
          break;
        default:
          /* !!! Mac: This error is wrong. */
          file->last_error = V3_FILE_ERROR_DENIED;
          break;
      }
      mutex_protected = file->last_error;
      v3MutexUnlock( &file->lock );
      return( mutex_protected );
    }
  }
  else
  {
    /* !!! Mac: This error is wrong. */
    file->last_error = V3_FILE_ERROR_DENIED;
    mutex_protected = file->last_error;
    v3MutexUnlock( &file->lock );
    return( mutex_protected );
  }
  open_result = file_ref_num;
#else

  /* force binary type open */
#if ( defined( O_BINARY ) )
  open_flags = open_flags | O_BINARY;
#endif

  /* Really try to open the file */
  open_result = open( (const char *) native_filename, open_flags, open_mode );
  if ( open_result == -1 )
  { /* Error during opening */
    switch( errno )
    {
      case ENOENT:
        file->last_error = V3_FILE_ERROR_NOTFOUND;
        break;
      case EACCES:
        file->last_error = V3_FILE_ERROR_DENIED;
        break;
#if ( defined( EROFS ) )
      case EROFS:
        file->last_error = V3_FILE_ERROR_WRITEMODE;
        break;
#endif
      default:
        file->last_error = V3_FILE_ERROR_DENIED;
    }
    mutex_protected = file->last_error;
    v3MutexUnlock( &file->lock );
    return( mutex_protected );
  }
#endif

  /* File correctly open, store the handle */
  file->handle = v3u64u32( open_result );

  /* Now lock if desired */
  if ( lock != V3_FILE_LOCK_NONE )
  {
    if ( v3_FileLock( file, lock ) != V3_PASS )
    {
#if ( defined( V3_FILE64 ) )
      close( (int) file->handle );
#else
      close( (int) v3u32u64( file->handle ) );
#endif
      /* file->last_error set by v3_FileLock on error */
      mutex_protected = file->last_error;
      v3MutexUnlock( &file->lock );
      return( mutex_protected );
    }
  }

  /* whipe and truncate if wanted */
  if ( ( mode & V3_FILE_MODE_TRUNCATE ) == V3_FILE_MODE_TRUNCATE )
  {
    if ( v3_FileTruncate( file, v3u64u32( (u32) 0 ) ) != V3_PASS )
    {
#if ( defined( V3_FILE64 ) )
      close( (int) file->handle );
#else
      close( (int) v3u32u64( file->handle ) );
#endif
      /* file->last_error set by v3_FileTruncate on error */
      mutex_protected = file->last_error;
      v3MutexUnlock( &file->lock );
      return( mutex_protected );
    }
  }

  /* set length, enough data is set to use this function call */
  if ( v3_FileLength( &file->length, file ) != V3_PASS )
  {
    /* not all filesystem have lengths */
    _V3_SET64( file->length, 00000000, 00000000 );
  }

  if ( v3_FileTell( &file->offset, file ) != V3_PASS )
  {
    if ( ( mode & V3_FILE_MODE_APPEND ) == V3_FILE_MODE_APPEND )
    {
      file->offset = file->length;
    }
    else
    {
      _V3_SET64( file->offset, 00000000, 00000000 );
    }
  }

  file->status = V3_FILE_STATUS_OPEN;
  file->last_error = V3_FILE_ERROR_NONE;

  v3MutexUnlock( &file->lock );
  return( V3_PASS );
}

u64 v3FileRead( void * buffer, u64 length, v3_FILE * file )
{
#if ( OS_TYPE == OS_MACOS )
  long in_out_length;
  OSErr my_err;
#endif
  int real_read;
  u64 tmp;

  if ( file == NULL )
  {
    /* We cant set last_error */
    return( v3u64u32( 0 ) );
  }

  v3MutexLock( &file->lock, V3_MUTEX_WAIT );

  if ( file->status != V3_FILE_STATUS_OPEN )
  {
    file->last_error = V3_FILE_ERROR_CLOSED;
    v3MutexUnlock( &file->lock );
    return( v3u64u32( 0 ) );
  }

#if ( OS_TYPE == OS_MACOS )
/* we use the file->handle as refnul, i.e., as a short */
  /* !!! Mac: fix this */
#if ( defined( CPU_64BIT ) )
  in_out_length = (long) length;
  my_err = FSRead( (short) v3u32u64( file->handle ), &in_out_length, buffer );
#else
  _V3_SET64( tmp, 00000000, 7FFFFFFF );
  if ( v3u64Gt( length, tmp ) )
  {
    v3MutexUnlock( &file->lock );
    return( v3u64u32( 0 ) );
  }
  in_out_length = v3u32u64( length );
  my_err = FSRead( (short) file->handle.lo, &in_out_length, buffer );
#endif
  if ( my_err == noErr )
  {
    real_read = in_out_length;
  }
  else
  {
    switch ( my_err )
    {
      case fLckdErr:
      case rfNumErr:
      case afpAccessDenied:
        file->last_error = V3_FILE_ERROR_DENIED;
        break;
      default:
        file->last_error = V3_FILE_ERROR_NOTFOUND;
    }
    /* !!! Mac: do something about the error */
  }
#else /* OS */
#if ( defined( V3_FILE64 ) )
  /* this should be valid since file handle is 64bit, int should be too */
  real_read = read( (int) file->handle, buffer, length );
#else
  _V3_SET64( tmp, 00000000, 7FFFFFFF );
  if ( v3u64Gt( length, tmp ) )
  {
    v3MutexUnlock( &file->lock );
    return( v3u64u32( 0 ) );
  }
  real_read = read( (int) v3u32u64( file->handle ), buffer,
    v3u32u64( length ) );
#endif
  if ( real_read == -1 )
  {
    if ( errno == EACCES )
    {
      file->last_error = V3_FILE_ERROR_DENIED;
    }
    else
    {
      file->last_error = V3_FILE_ERROR_NOTFOUND;
    }
#endif /* OS */

    v3_FileTell( &tmp, file );

    v3MutexUnlock( &file->lock );
    return( v3u64u32( 0 ) );
  }

  if ( v3_FileTell( &tmp, file ) != V3_PASS )
  {
    /* Set offset ourselves */
    file->offset = v3u64Add( file->offset, length );
  }

  v3MutexUnlock( &file->lock );

#if ( defined( V3_FILE64 ) )
  return( real_read );
#else
  return( v3u64u32( (u32) real_read ) );
#endif
}

u64 v3FileWrite( v3_FILE * file, const void * const buffer, u64 length )
{
  u64 result;

  if ( file == NULL )
  {
    /* Can't set last_error */
    return( v3u64u32( 0 ) );
  }

  v3MutexLock( &file->lock , V3_MUTEX_WAIT );

  if ( file->status != V3_FILE_STATUS_OPEN )
  {
    file->last_error = V3_FILE_ERROR_CLOSED;
    v3MutexUnlock( &file->lock );
    return( v3u64u32( 0 ) );
  }

  result = v3_FileWrite( file, buffer, length);

  v3MutexUnlock( &file->lock );

  return( result );
}

s32 v3FileSeek( v3_FILE * file, u64 offset )
{
  s32 result;

  if ( file == NULL )
  {
    /* Can't set last_error */
    return( V3_FILE_ERROR_NOTFOUND );
  }

  v3MutexLock( &file->lock, V3_MUTEX_WAIT );

  if ( file->status != V3_FILE_STATUS_OPEN )
  {
    v3MutexUnlock( &file->lock );
    return( V3_FILE_ERROR_CLOSED );
  }

  result = v3_FileSeek( file, offset );

  v3MutexUnlock ( &file->lock );
  return( result );
}

s32 v3FileTell( u64 * offset, v3_FILE * file )
{
  u64 len;
  s32 result;

  if ( file == NULL )
  {
    /* Can't set last_error */
    return( V3_FILE_ERROR_NOTFOUND );
  }

  v3MutexLock( &file->lock , V3_MUTEX_WAIT );

  if ( file->status != V3_FILE_STATUS_OPEN )
  {
    v3MutexUnlock( &file->lock );
    return( V3_FILE_ERROR_CLOSED );
  }

  if ( ( result = v3_FileTell( &len, file ) ) == V3_PASS )
  {
    *(offset) = len;
  }
  else
  {
    *(offset) = file->offset;
  }

  v3MutexUnlock( &file->lock );
  return( result );
}

s32 v3FileEOF( v3_FILE * file )
{
  u64 current;
  u64 real;
  s32 result;

  if ( file == NULL )
  {
    /* Can't set last_error */
    return( V3_FILE_ERROR_NOTFOUND );
  }

  v3MutexLock( &file->lock, V3_MUTEX_WAIT );

  if ( file->status != V3_FILE_STATUS_OPEN )
  {
    v3MutexUnlock( &file->lock );
    return( V3_FILE_ERROR_CLOSED );
  }

  if ( ( result = v3_FileTell( &current , file ) ) != V3_PASS )
  {
    v3MutexUnlock( &file->lock );
    return( result );
  }

  if ( ( result = v3_FileLength( &real, file ) ) != V3_PASS )
  {
    v3MutexUnlock( &file->lock );
    return( result );
  }

  v3MutexUnlock( &file->lock );

  /* We compare file size and current position to know if we are at EOF */
  if ( v3u64Eq( current, real ) )
  {
    return( V3_FILE_ERROR_EOF );
  }

  /* not EOF */
  return( V3_PASS );
}

s32 v3FileLength( u64 * length, v3_FILE * file )
{
  s32 result;
  u64 len;

  if ( file == NULL )
  {
    /* Can't set last_error */
    return( V3_FILE_ERROR_NOTFOUND );
  }

  v3MutexLock( &file->lock, V3_MUTEX_WAIT );

  if ( file->status != V3_FILE_STATUS_OPEN )
  {
    v3MutexUnlock( &file->lock );
    return( V3_FILE_ERROR_CLOSED );
  }

  if ( ( result = v3_FileLength( &len, file ) ) != V3_PASS )
  {
    v3MutexUnlock( &file->lock );
    return( result );
  }

  *(length) = len;

  v3MutexUnlock( &file->lock );

  return( V3_PASS );
}

s32 v3FileTruncate( v3_FILE * file, u64 length )
{
  s32 result;

  if ( file == NULL )
  {
    /* Can't set last_error */
    return( V3_FILE_ERROR_NOTFOUND );
  }

  v3MutexLock( &file->lock, V3_MUTEX_WAIT );

  if ( file->status != V3_FILE_STATUS_OPEN )
  {
    v3MutexUnlock( &file->lock );
    return( V3_FILE_ERROR_CLOSED );
  }

  if ( ( result = v3_FileTruncate( file, length ) ) != V3_PASS )
  {
    v3MutexUnlock( &file->lock );
    return( result );
  }

  v3MutexUnlock( &file->lock );

  return( V3_PASS );
}

s32 v3FileDelete( const ascii * filename )
{
#if ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  FSSpec spec;
#endif
  v3_FILENAME native_filename;
  v3_FILE wipe_file;
  s32 error;

  /* Convert slashs to native format and put it into native_filename */
  if ( v3_FileNativePath( native_filename, filename ) != V3_PASS )
  {
    return( V3_FILE_ERROR_NAME );
  }

  /* open the file for write only, exclusive access, and zero out */
  v3MemSet( &wipe_file, v3u64u32( sizeof( v3_FILE ) ), 0 );
  error = v3FileOpen( &wipe_file, filename, V3_FILE_MODE_WRITE |
    V3_FILE_MODE_SYNC, V3_FILE_LOCK_WRITE );
  if ( error != V3_PASS )
  {
    if ( error == V3_FILE_ERROR_NOTFOUND )
    {
      /* no such file */
      return( V3_PASS );
    }
    else
    {
      return( error );
    }
  }

  if ( ( error = v3_FileWipe( &wipe_file, v3u64u32( 0 ) ) ) != V3_PASS )
  {
    return( error );
  }

  v3FileClose( &wipe_file );

#if ( OS_TYPE == OS_MACOS )
  my_err = FSMakeFSSpec( 0, 0, native_filename, &spec );
  if ( my_err != noErr )
  {
    return( V3_FILE_ERROR_NOTFOUND );
  }

  my_err = FSpDelete( &spec );
  if ( my_err != noErr )
  {
    if ( my_err == afpAccessDenied )
    {
      return( V3_FILE_ERROR_DENIED );
    }
    else
    {
      /* !!! Mac: fix this, see below */
      return( V3_FILE_ERROR_NOTFOUND );
    }
  }
#else
  if ( remove( (const char *) native_filename ) != 0 )
  {
    if ( errno == ENOENT )
    {
      /* this should have been cought above */
      return( V3_PASS );
    }
    else /* EACCES */
    {
      return( V3_FILE_ERROR_DENIED );
    }
  }
#endif

  return( V3_PASS );
}

s32 v3FileInfo( v3_FILE_INFO * info, const ascii * filename )
{
#if ( OS_TYPE == OS_WIN32 ) 
  struct _stat buf;
#else
  struct stat buf;
#endif
  v3_FILENAME native_filename;
  v3_time unix_epoc;

  if ( ( info == NULL ) || ( filename == NULL ) )
  {
    return( V3_FILE_ERROR_DENIED );
  }

  /* translate filename */
  if ( v3_FileNativePath( native_filename, filename ) != V3_PASS )
  {
    return( V3_FILE_ERROR_NAME );
  }

#if ( OS_TYPE == OS_WIN32 ) 
  if ( _stat( (const char *) native_filename, &buf ) != 0 )
#else
  if ( stat( (const char *) native_filename, &buf ) != 0 )
#endif
  {
    switch( errno )
    {
      case ENOENT:
        return ( V3_FILE_ERROR_NOTFOUND );
        break;
      case EACCES:
      default:
        return( V3_FILE_ERROR_DENIED );
    }
  }

  /* length */
  info->length = v3u64u32( buf.st_size );

#if ( OS_TYPE == OS_WIN32 ) 
  /* type */
  if ( ( buf.st_mode & _S_IFREG ) == _S_IFREG )
  {
    info->type = V3_FILE_TYPE_FILE;
  }
  else if ( ( buf.st_mode & _S_IFDIR ) == _S_IFDIR )
  {
    info->type = V3_FILE_TYPE_DIR;
  }
  else if ( ( buf.st_mode & _S_IFCHR ) == _S_IFCHR )
  {
    info->type = V3_FILE_TYPE_DEVICE;
  }
  else
  {
    info->type = V3_FILE_TYPE_SPECIAL;
  }

  /* rights */
  info->rights = 0;
  if ( ( buf.st_mode & _S_IREAD ) == _S_IREAD )
  {
    info->rights += V3_FILE_RIGHTS_READ;
  }
  if ( ( buf.st_mode & _S_IWRITE ) == _S_IWRITE )
  {
    info->rights += V3_FILE_RIGHTS_WRITE;
    info->rights += V3_FILE_RIGHTS_APPEND;
  }
  if ( ( buf.st_mode & _S_IEXEC ) == _S_IEXEC )
  {
    info->rights += V3_FILE_RIGHTS_EXEC;
  }
#else /* OS */
  /* type */
  if ( ( buf.st_mode & S_IFREG ) == S_IFREG )
  {
    info->type = V3_FILE_TYPE_FILE;
  }
  else if ( ( buf.st_mode & S_IFDIR ) == S_IFDIR )
  {
    info->type = V3_FILE_TYPE_DIR;
  }
  else if ( ( ( buf.st_mode & S_IFCHR ) == S_IFCHR ) ||
    ( ( buf.st_mode & S_IFBLK ) == S_IFBLK ) )
  {
    info->type = V3_FILE_TYPE_DEVICE;
  }
  else
  {
    info->type = V3_FILE_TYPE_SPECIAL;
  }

  /* rights */
  info->rights = 0;
  if ( ( buf.st_mode & S_IRUSR ) == S_IRUSR )
  {
    info->rights += V3_FILE_RIGHTS_READ;
  }
  if ( ( buf.st_mode & S_IWUSR ) == S_IWUSR )
  {
    info->rights += V3_FILE_RIGHTS_WRITE;
    info->rights += V3_FILE_RIGHTS_APPEND;
  }
  if ( ( buf.st_mode & S_IXUSR ) == S_IXUSR )
  {
    info->rights += V3_FILE_RIGHTS_EXEC;
  }
#endif /* OS */

  /* times */
  _V3_SET128( unix_epoc, 00000000, 386D4380, 00000000, 00000000 );

#if ( defined( CPU_64BIT ) )
  info->create.hi = (s64) buf.st_mtime;
#else
  info->create.hi.hi = 0;
  info->create.hi.lo = (u32) buf.st_mtime;
#endif
  _V3_SET64( info->create.lo, 00000000, 00000000 );
  info->create = v3s128Sub( info->create, unix_epoc );
  
#if ( defined( CPU_64BIT ) )
  info->modify.hi = (s64) buf.st_ctime;
#else
  info->modify.hi.hi = 0;
  info->modify.hi.lo = (u32) buf.st_ctime;
#endif
  _V3_SET64( info->modify.lo, 00000000, 00000000 );
  info->modify = v3s128Sub( info->modify, unix_epoc );
  
#if ( defined( CPU_64BIT ) )
  info->access.hi = (s64) buf.st_atime;
#else
  info->access.hi.hi = 0;
  info->access.hi.lo = (u32) buf.st_atime;
#endif
  _V3_SET64( info->access.lo, 00000000, 00000000 );
  info->access = v3s128Sub( info->access, unix_epoc );
  
  return( V3_PASS );
}

s32 v3FileClose( v3_FILE * file )
{
#if ( OS_TYPE == OS_MACOS )
  OSErr my_err;
#endif
  if ( file == NULL )
  {
    return( V3_FILE_ERROR_NOTFOUND );
  }

  v3MutexLock( &file->lock , V3_MUTEX_WAIT );

  /* First check if the file is open */
  if ( file->status != V3_FILE_STATUS_OPEN )
  {
    v3MutexUnlock( &file->lock );
    return( V3_FILE_ERROR_CLOSED );
  }

  /* Release lock */
  v3_FileUnLock( file );
  /* !!! deal with failed case, not good, very very not good */

#if ( OS_TYPE == OS_MACOS )
  my_err = FSClose( (short) v3u32u64( file->handle ) );
  if ( my_err != noErr)
#else
#if ( defined( V3_FILE64 ) )
  if ( close( (int) file->handle ) != 0 )
#else
  if ( close( (int) v3u32u64( file->handle ) ) != 0 )
#endif
#endif /* OS */
  {
    file->last_error = V3_FILE_ERROR_NOTFOUND;
    v3MutexUnlock( &file->lock );
    return( V3_FILE_ERROR_NOTFOUND );
  }

  file->status = V3_FILE_STATUS_CLOSED;

  v3MutexUnlock( &file->lock );
  v3MutexFree( &file->lock );

  return( V3_PASS );
}

/* Directory functions. */

s32 v3DirOpen( v3_DIR * dir, const ascii * dirname, u32 mode )
{
#if ( OS_TYPE == OS_MACOS )
  /* We don't open directories on Mac. We do create them though */
  OSErr my_err;
  FSSpec spec;
  long dir_id;
#elif ( OS_TYPE == OS_OS2 )
  struct stat stat_buf;
  HDIR hdirFindHandle = HDIR_CREATE;
  FILEFINDBUF3 findBuffer = {0};
  ULONG findCount = 1;
#endif
  v3_FILENAME native_dirname;
  s32 error;
  s32 mutex_protected;
  u32 higher_levels;
  v3_FILENAME tmp_native_dirname;
  v3_FILENAME tmp_dirname;
  ascii * currpos, * nativepos;
  u64 size, i;

  if ( dir == NULL )
  {
    /* Can't set last_error */
    return( V3_DIR_ERROR_NOTFOUND );
  }

  v3MutexLock( &dir->lock, V3_MUTEX_WAIT );

  if ( dir->status != V3_DIR_STATUS_CLOSED )
  {
    v3MutexUnlock( &dir->lock );
    return( V3_DIR_ERROR_CLOSED );
  }

  if ( ( ( ( ( mode & V3_DIR_MODE_EXIST ) == V3_DIR_MODE_EXIST ) ? 1 : 0 ) +
    ( ( ( mode & V3_DIR_MODE_READ ) == V3_DIR_MODE_READ ) ? 1 : 0 ) +
    ( ( ( mode & V3_DIR_MODE_CREATE ) == V3_DIR_MODE_CREATE ) ? 1 : 0 ) )
    != 1 )
  {
    /* mode combination is invalid */
    dir->last_error = V3_DIR_ERROR_MODE;
    v3MutexUnlock( &dir->lock );
    return( V3_DIR_ERROR_MODE );
  }

  if ( mode & ~( V3_DIR_MODE_EXIST | V3_DIR_MODE_READ | V3_DIR_MODE_CREATE ) )
  {
    /* Undefined modes used */
    dir->last_error = V3_DIR_ERROR_MODE;
    v3MutexUnlock( &dir->lock );
    return( V3_DIR_ERROR_MODE );
  }

  /* Convert dirname and put it into native_dirname */
  if ( v3_FileNativePath( native_dirname, dirname ) != V3_PASS )
  {
    dir->last_error = V3_DIR_ERROR_NAME;
    v3MutexUnlock( &dir->lock );
    return( V3_DIR_ERROR_NAME );
  }

  if ( ( mode & V3_DIR_MODE_CREATE ) == V3_DIR_MODE_CREATE )
  {
    /* We are trying to create the directory */

#if ( OS_TYPE == OS_MACOS )
  /*
    !!! Note for MacOS porter : this macos block needs to be mangled and moved
    down below into the do-while loop with the other OS cases.
  */
    my_err = FSMakeFSSpec( 0, 0, native_dirname, &spec );
    if ( my_err != noErr )
    {
      dir->last_error = V3_DIR_ERROR_NOTFOUND;
      v3MutexUnlock( &dir->lock );
      return( V3_DIR_ERROR_NOTFOUND );
    }
    my_err = FSpDirCreate( &spec, smSystemScript, &dir_id );
    if ( my_err == noErr )
    {
#if ( defined( CPU_64BIT ) )
      dir->handle = v3MemAlloc( sizeof( long ), V3_MEM_NORMAL );
#else
      dir->handle = v3MemAlloc( v3u64u32( sizeof( long ) ), V3_MEM_NORMAL );
#endif
      *(dir->handle) = dir_id;
    }
    else
    {
      /* mkdir failed */
      dir->last_error = V3_DIR_ERROR_CREATE;
      v3MutexUnlock( &dir->lock );
      return( V3_DIR_ERROR_CREATE );
    }
    dir->last_error = V3_DIR_ERROR_NONE;
    v3MutexUnlock( &dir->lock );
    return( V3_PASS );
  }
#else /* End of MacOS createDir section */
  /*
    Find the first | to find the native part of the filename, if any.
    We do not try to create the native part.
  */
  i = size = v3StrLengthA( dirname );
  nativepos = v3StrCharA( dirname, '|', i );

  /* Make a copy of the path. */
  v3u64Inc( &size );
  if ( v3StrCopyA( tmp_dirname, dirname, size ) != V3_PASS )
  {
    v3MutexUnlock( &dir->lock );
    dir->last_error = V3_DIR_ERROR_CREATE;
    return( V3_DIR_ERROR_CREATE );
  }

  /* Remove a trailing slash. */
  currpos = (ascii *) v3MemOffset( tmp_dirname,
    v3u64Sub( i, v3u64u32( 1 ) ) );
  if ( *currpos == '/' )
  {
    *currpos = '\0';
  }

  /* translate to native path type */
  v3_FileNativePath( tmp_native_dirname, tmp_dirname );

  higher_levels = 1; /* We should make atleast one directory. */

  do
  {
#if ( OS_TYPE == OS_OS2 )
    error = mkdir( (char *) tmp_native_dirname );
#else /* OS */
#if ( defined( S_IRGRP ) )
    /* platform handling user groups */
    error = mkdir( (const char *) tmp_native_dirname, S_IRUSR | S_IWUSR |
      S_IXUSR | S_IRGRP | S_IXGRP );
#else
    /* platform not handling user groups -> use old style */
    error = mkdir( (const char *) tmp_native_dirname, S_IREAD |
      S_IWRITE | S_IEXEC );
#endif /* S_IRGRP */
#endif /* OS */

    if ( error != 0 )
    {
      /*
        Check if mkdir failed because of lacking earlier elements
        in the path.
      */
      if ( errno == ENOENT )
      {
        higher_levels++;
        /* Remove one level from path and try creating it.  */
        do
        {
          v3u64Dec( &i );
          currpos = (ascii *) v3MemOffset( tmp_dirname, i );
        }
        while ( ( *currpos != '/' ) && ( currpos != nativepos ) &&
          ( v3u64Gt( i , v3u64u32( 0 ) ) ) );

        /*
          If we hit 'bottom' of the tree without managing to
          create a directory we have to give up.
        */
        if ( ( currpos == nativepos ) || ( v3u64Eq( i, v3u64u32( 0 ) ) ) )
        {
          v3MutexUnlock( &dir->lock );
          dir->last_error = V3_DIR_ERROR_CREATE;
          return( V3_DIR_ERROR_CREATE  );
        }

        /* Snub the path. */
        *currpos = '\0';
        /* Convert to native. */
        if ( v3_FileNativePath( tmp_native_dirname, tmp_dirname ) != V3_PASS )
        {
          dir->last_error = V3_DIR_ERROR_NAME;
          v3MutexUnlock( &dir->lock );
          return( V3_DIR_ERROR_NAME );
        }
      }
      else
      {
        /* mkdir failed for some other reason then missing parent. */
        dir->last_error = V3_DIR_ERROR_CREATE;
        v3MutexUnlock( &dir->lock );
        return( V3_DIR_ERROR_CREATE );
      }
    }
    else
    {
      /*
        We managed to create a directory, so we can take another shot at
        the next higher level in the tree.
      */
      higher_levels--;
      if ( higher_levels > 0 )
      {
        do
        {
          currpos = (ascii *) v3MemOffset( tmp_dirname, i );
          v3u64Inc( &i );
        } while ( ( *currpos != '\0' ) && ( v3u64Lt( i, size ) ) );

        v3u64Dec( &i );
        *currpos = '/';

        /*
          This means that we reached the end of the list.
          This should not happend.
        */
        if ( v3u64Eq( i, size ) )
        {
          dir->last_error = V3_DIR_ERROR_CREATE;
          v3MutexUnlock( &dir->lock );
          return( V3_DIR_ERROR_CREATE );
        }

        if ( v3_FileNativePath( tmp_native_dirname, tmp_dirname ) != V3_PASS )
        {
          dir->last_error = V3_DIR_ERROR_NAME;
          v3MutexUnlock( &dir->lock );
          return( V3_DIR_ERROR_NAME );
        }
      } /* if (higher_levels > 0 ) */
    }
  /* While we need to create directories higher up in the directory tree. */
  } while ( higher_levels > 0 );

  /* End of mkdir routines. */
#endif /* OS */

    if ( error != 0 )
    {
      /* mkdir failed */
      dir->last_error = V3_DIR_ERROR_CREATE;
      v3MutexUnlock( &dir->lock );
      return( V3_DIR_ERROR_CREATE );
    }

    dir->last_error = V3_DIR_ERROR_NONE;
    v3MutexUnlock( &dir->lock );
    return( V3_PASS );
  }

  /* we are trying to test existence or read the directory */

  v3StrCopyA( dir->dirname, dirname, v3u64u32( V3_FILE_MAX_FILENAME + 1 ) );

#if ( OS_TYPE == OS_MACOS )
  /* !!! do something */
#elif ( OS_TYPE == OS_OS2 )
  /* DIR_MODE_EXIST != DIR_MODE_OPEN plus dirclose() */
  if ( stat( (const char *) native_dirname, &stat_buf ) == 0 )
  {
    if ( stat_buf.st_mode & S_IFDIR )
    {
      dir->last_error = V3_DIR_ERROR_NONE;
    }
    else
    {
      dir->last_error = V3_DIR_ERROR_NOTFOUND;
    }
  }
  else
  {
    dir->last_error = V3_DIR_ERROR_NOTFOUND;
  }

  if ( dir->last_error == V3_DIR_ERROR_NOTFOUND )
  {
    v3MutexUnlock( &dir->lock );
    return( V3_DIR_ERROR_NOTFOUND );
  }

  if ( ( mode & V3_DIR_MODE_EXIST ) == V3_DIR_MODE_EXIST )
  {
    /* and that was all we needed to do */
    mutex_protected = dir->last_error;
    v3MutexUnlock( &dir->lock );
    return( mutex_protected );
  }
  /* exist test succeeded, let's set up for findfirst & findnext processing */
  if ( DosFindFirst( (char *) native_dirname, &hdirFindHandle,
    FILE_NORMAL | FILE_DIRECTORY, &findBuffer, sizeof( FILEFINDBUF3 ),
    &findCount, FIL_STANDARD ) == 0 )
  {
    dir->handle = (DIR *) hdirFindHandle;
    dir->last_error = V3_DIR_ERROR_NONE;
    dir->status = V3_DIR_STATUS_OPEN;
    v3MutexUnlock( &dir->lock );
    return( V3_PASS );
  }
  else
  {
    v3MutexUnlock( &dir->lock );
    return( V3_DIR_ERROR_EOF );
  }
#elif ( OS_TYPE == OS_WIN32 )
  /*
    add the "\*" to the end of the native if we can, check for MAX_PATH
  */
  i = v3StrLengthA( native_dirname );
  if ( ( v3u64Gt( i, v3u64u32( V3_FILE_MAX_FILENAME - 2 ) ) ) ||
    ( v3u64Gt( i, v3u64u32( MAX_PATH - 2 ) ) ) )
  {
    /* too long */
    dir->last_error = V3_DIR_ERROR_NAME;
    v3MutexUnlock( &dir->lock );
    return( V3_DIR_ERROR_NAME );
  }
  else
  {
    /* add the "\*" */
    currpos = (ascii *) v3MemOffset( native_dirname, i );
    *currpos++ = '\\';
    *currpos++ = '*';
    *currpos = '\0';
  }
  /* open it */
  if ( ( dir->handle = FindFirstFile( (const char *) native_dirname,
    &dir->find_data ) ) == INVALID_HANDLE_VALUE )
  {
    switch( GetLastError() )
    {
      case ERROR_DIRECTORY:
      case ERROR_INVALID_NAME:
        dir->last_error = V3_DIR_ERROR_NOTFOUND;
        break;
      default:
        dir->last_error = V3_DIR_ERROR_DENIED;
    }
    mutex_protected = dir->last_error;
    v3MutexUnlock( &dir->lock );
    return( mutex_protected );
  }

#else
  /* Really try to open the directory */
  dir->handle = opendir( (const char *) native_dirname );
  if ( dir->handle == NULL )
  {
    switch( errno )
    {
      case ENOENT:
        dir->last_error = V3_DIR_ERROR_NOTFOUND;
        break;
      case ENOTDIR:
      case EACCES:
      default:
        dir->last_error = V3_DIR_ERROR_DENIED;
    }
    mutex_protected = dir->last_error;
    v3MutexUnlock( &dir->lock );
    return( mutex_protected );
  }
#endif
#if ( OS_TYPE == OS_OS2 )
/* already been taken care of */
#else
  if ( ( mode & V3_DIR_MODE_EXIST ) == V3_DIR_MODE_EXIST )
  {
    /* We were just testing for the directory existence -> close dir */
#if ( OS_TYPE == OS_MACOS )
    /* !!! Mac: close the dir? */
    v3MemFree( dir->handle );
#elif ( OS_TYPE == OS_WIN32 )
    FindClose( dir->handle );
#else
    closedir( dir->handle );
#endif
    v3MutexUnlock( &dir->lock );
    return( V3_PASS );
  }

  dir->last_error = V3_DIR_ERROR_NONE;
  dir->offset = v3u64u32( 0 );
  dir->status = V3_DIR_STATUS_OPEN;
  v3MutexUnlock( &dir->lock );
  return( V3_PASS );
#endif
}

u64 v3DirRead( v3_FILENAME * buffer, u64 length, v3_DIR * dir )
{
#if ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  CInfoPBRec cpb;
  FSSpec spec;
  v3_FILENAME native_dirname;
  Str255 item_name;
  short index = 1;
#elif ( OS_TYPE == OS_OS2 )
  FILEFINDBUF3 FindBuffer = { 0 }; /* Returned from FindFirst/Next */
  ULONG ulFindCount = 1; /* Look for 1 file at a time */
#elif ( OS_TYPE == OS_WIN32 )
  u32 result;
#else
  struct dirent * readdir_result;
#endif
  s32 done;
  u32 ok_to_exit;

  if ( dir == NULL )
  {
    /* Can't set last_error */
    return( v3u64u32( 0 ) );
  }

  v3MutexLock( &dir->lock, V3_MUTEX_WAIT );

  if ( dir->status != V3_DIR_STATUS_OPEN )
  {
    dir->last_error = V3_DIR_ERROR_CLOSED;
    v3MutexUnlock( &dir->lock );
    return( v3u64u32( 0 ) );
  }

  done = 0;
  ok_to_exit = 0;
  dir->last_error = V3_DIR_ERROR_NONE;

#if ( OS_TYPE == OS_MACOS )
  /* Convert slashs to native format and put it into native_dirname */
  if ( v3_FileNativePath( native_dirname, dir->dirname ) != V3_PASS )
  {
    dir->last_error = V3_DIR_ERROR_NAME;
    v3MutexUnlock( &dir->lock );
    return( v3u64u32( 0 ) );
  }

  my_err = FSMakeFSSpec( 0, 0, native_dirname, &spec );
  if ( my_err != noErr )
  {
    dir->last_error = V3_DIR_ERROR_NOTFOUND;
    v3MutexUnlock( &dir->lock );
    return( v3u64u32( 0 ) );
  }

  cpb.dirInfo.ioCompletion = NULL;
  cpb.dirInfo.ioNamePtr = item_name;
  cpb.dirInfo.ioVRefNum = spec.vRefNum;
  cpb.hFileInfo.ioDirID = spec.parID;

  do
  {
    cpb.hFileInfo.ioFDirIndex = index;
    my_err = PBGetCatInfoSync( &cpb );
    if ( my_err != noErr )
    {
      dir->last_error = V3_DIR_ERROR_EOF;
      ok_to_exit = 1;
    }
    else
    {
      v3StrCopyA( buffer[done], (ascii *) p2cstr(item_name),
        v3u64u32( V3_FILE_MAX_FILENAME + 1 ) );
      done = done + 1;
      v3u64Inc( &dir->offset );

#if ( defined( CPU_64BIT ) )
      if ( done == length )
#else
      if ( v3u64Eq( length, v3u64u32( done ) ) )
#endif
      {
        /* done for now */
        ok_to_exit = 1;
      }
    }
  } while ( ok_to_exit == 0 );
#elif ( OS_TYPE == OS_WIN32 )
  while ( ok_to_exit == 0 )
  {
    if ( v3u64Eq( dir->offset, v3u64u32( 0 ) ) )
    {
      /* copy the filename given by readdir into the buffer */
      v3StrCopyA( buffer[done], (ascii *) &dir->find_data.cFileName,
        v3u64u32( V3_FILE_MAX_FILENAME + 1 ) );
      done = done + 1;
      v3u64Inc( &dir->offset );

      if ( v3u64Eq( length, v3u64u32( done ) ) )
      {
        /* Buffer is full */
        ok_to_exit = 1;
      }
    }
    else
    {
      result = FindNextFile( dir->handle, &dir->find_data );
      if ( result == 0 )
      {
        /* failed, probably out of files, either way we are done */
        dir->last_error = V3_DIR_ERROR_EOF;
        ok_to_exit = 1;
      }
      else
      {
        v3StrCopyA( buffer[done], (ascii *) &dir->find_data.cFileName,
          v3u64u32( V3_FILE_MAX_FILENAME + 1 ) );
        done = done + 1;
        v3u64Inc( &dir->offset );

        if ( v3u64Eq( length, v3u64u32( done ) ) )
        {
          /* done for now */
          ok_to_exit = 1;
        }
      }
    }
  }
#elif ( OS_TYPE == OS_OS2 )
  while ( ok_to_exit == 0 )
  {
    if ( DosFindNext( (u32) dir->handle, &FindBuffer,
      sizeof( FILEFINDBUF3 ), &ulFindCount ) != 0 )
    {
      /* failed, probably because we reached end of directory */
      dir->last_error = V3_DIR_ERROR_EOF;
      ok_to_exit = 1;
    }
    else
    {
      /* copy the filename given by readdir into the buffer */
      v3StrCopyA( buffer[done], (ascii *) FindBuffer.achName,
        v3u64u32( V3_FILE_MAX_FILENAME + 1 ) );
      done = done + 1;
      v3u64Inc( &dir->offset );

      if ( v3u64Eq( length, v3u64u32( done ) ) )
      {
        /* done for now */
        ok_to_exit = 1;
      }
    }
  }
#else /* OS */
  while ( ok_to_exit == 0 )
  {
    readdir_result = readdir( dir->handle );

    if ( readdir_result == NULL )
    {
      dir->last_error = V3_DIR_ERROR_EOF;
      ok_to_exit = 1;
    }
    else
    {
      /* copy the filename given by readdir into the buffer */
      v3StrCopyA( buffer[done], (ascii *) readdir_result->d_name,
        v3u64u32( V3_FILE_MAX_FILENAME + 1 ) );
      done = done + 1;
      v3u64Inc( &dir->offset );

      if ( v3u64Eq( length, v3u64u32( done ) ) )
      {
        /* done for now */
        ok_to_exit = 1;
      }
    }
  }
#endif /* OS */

  v3MutexUnlock( &dir->lock );

#if ( defined( CPU_64BIT ) )
  return( done );
#else
  return( v3u64u32( done ) );
#endif
}

s32 v3DirDelete( const ascii * dirname )
{
#if ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  FSSpec spec;
#endif
  v3_FILENAME native_dirname;

  /* Convert slashs to native format and put it into native_dirname */
  if ( v3_FileNativePath( native_dirname, dirname ) != V3_PASS )
  {
    return( V3_DIR_ERROR_NAME );
  }

#if ( OS_TYPE == OS_MACOS )
  my_err = FSMakeFSSpec( 0, 0, native_dirname, &spec );
  if ( my_err != noErr )
  {
    return( V3_DIR_ERROR_NOTFOUND );
  }

  my_err = FSpDelete( &spec );
  if ( my_err != noErr )
  {
    switch ( my_err )
    {
      case afpAccessDenied:
        return( V3_DIR_ERROR_DENIED );
      case fBsyErr:
        return( V3_DIR_ERROR_NOTEMPTY );
      default:
        return( V3_DIR_ERROR_NOTFOUND );
    }
  }
#elif ( OS_TYPE == OS_WIN32 )
  if ( RemoveDirectory( (const char *) native_dirname ) == 0 )
  {
    switch( GetLastError() )
    {
      case ERROR_DIR_NOT_EMPTY:
      case WSAENOTEMPTY:
        return( V3_DIR_ERROR_NOTEMPTY );
      case ERROR_DIRECTORY:
      case ERROR_INVALID_NAME:
        return( V3_DIR_ERROR_NOTFOUND );
      default:
        return( V3_DIR_ERROR_DENIED );
    }
  }
#else /* OS */
#if ( OS_TYPE == OS_OS2 )
  if ( rmdir( (char *) native_dirname ) != 0 )
#else
  if ( rmdir( (const char *) native_dirname ) != 0 )
#endif
  {
    switch( errno )
    {
      case EACCES:
        return( V3_DIR_ERROR_DENIED );
      case ENOENT:
        return( V3_DIR_ERROR_NOTFOUND );
#if ( defined( ENOTEMPTY ) )
      case ENOTEMPTY:
        return( V3_DIR_ERROR_NOTEMPTY );
#endif
      default:
        return( V3_DIR_ERROR_DENIED );
    }
  }
#endif /* OS */

  return( V3_PASS );
}

s32 v3DirClose( v3_DIR * dir )
{
  if ( dir == NULL )
  {
    /* Can't set last_error */
    return( V3_DIR_ERROR_NOTFOUND );
  }

  v3MutexLock( &dir->lock, V3_MUTEX_WAIT );

  if ( dir->status != V3_DIR_STATUS_OPEN )
  {
    v3MutexUnlock( &dir->lock );
    return( V3_DIR_ERROR_CLOSED );
  }

#if ( OS_TYPE == OS_OS2 )
  if ( DosFindClose( (u32) dir->handle ) != NO_ERROR )
#elif ( OS_TYPE == OS_WIN32 )
  if ( FindClose( dir->handle ) == 0 )
#else
  if ( closedir( dir->handle ) != 0 )
#endif
  {
    dir->last_error = V3_DIR_ERROR_NOTFOUND;
    v3MutexUnlock( &dir->lock );
    return( V3_DIR_ERROR_NOTFOUND );
  }

  dir->status = V3_DIR_STATUS_CLOSED;

#if ( OS_TYPE == OS_MACOS )
  v3MemFree( dir->handle );
#endif

  v3MutexUnlock( &dir->lock );
  v3MutexFree( &dir->lock );

  return( V3_PASS );
}

/* byte order functions */

void v3u16Load( u16 * num, const u8 * bytes )
{
  if ( ( num == NULL ) || ( bytes == NULL ) )
    return;

  *(num) = (u16) ( ( (u16) bytes[0] << 8 ) | (u16) bytes[1] );
}

void v3u16Save( u8 * bytes, const u16 * const num )
{
  u16 tmp;

  if ( ( num == NULL ) || ( bytes == NULL ) )
    return;

  tmp = *(num);
  bytes[0] = (u8) ( ( tmp >> 8 ) & 0xFF );
  bytes[1] = (u8) ( tmp & 0xFF );
}

void v3u32Load( u32 * num, const u8 * bytes )
{
  if ( ( num == NULL ) || ( bytes == NULL ) )
    return;

  *(num) = ( (u32) bytes[0] << 24 ) | ( (u32) bytes[1] << 16 ) |
    ( (u32) bytes[2] << 8 ) | (u32) bytes[3];
}

void v3u32Save( u8 * bytes, const u32 * const num )
{
  u32 tmp;

  if ( ( num == NULL ) || ( bytes == NULL ) )
    return;

  tmp = *(num);
  bytes[0] = (u8) ( ( tmp >> 24 ) & 0xFF );
  bytes[1] = (u8) ( ( tmp >> 16 ) & 0xFF );
  bytes[2] = (u8) ( ( tmp >> 8 ) & 0xFF );
  bytes[3] = (u8) ( tmp & 0xFF );
}

void v3u64Load( u64 * num, const u8 * bytes )
{
#if ( defined( CPU_64BIT ) )
  if ( ( num == NULL ) || ( bytes == NULL ) )
    return;

  *(num) = ( (u64) bytes[0] << 56 ) | ( (u64) bytes[1] << 48 ) |
    ( (u64) bytes[2] << 40 ) | ( (u64) bytes[3] << 32 ) |
    ( (u64) bytes[4] << 24 ) | ( (u64) bytes[5] << 16 ) |
    ( (u64) bytes[6] << 8 ) | (u64) bytes[7];
#else
  u64 tmp;

  if ( ( num == NULL ) || ( bytes == NULL ) )
    return;

  tmp.hi = ( (u32) bytes[0] << 24 ) | ( (u32) bytes[1] << 16 ) |
    ( (u32) bytes[2] << 8 ) | (u32) bytes[3];
  tmp.lo = ( (u32) bytes[4] << 24 ) | ( (u32) bytes[5] << 16 ) |
    ( (u32) bytes[6] << 8 ) | (u32) bytes[7];

  *(num) = tmp;
#endif
}

void v3u64Save( u8 * bytes, const u64 * const num )
{
  u64 tmp;

  if ( ( num == NULL ) || ( bytes == NULL ) )
    return;

  tmp = *(num);

#if ( defined( CPU_64BIT ) )
  bytes[0] = (u8) ( ( tmp >> 56 ) & 0xFF );
  bytes[1] = (u8) ( ( tmp >> 48 ) & 0xFF );
  bytes[2] = (u8) ( ( tmp >> 40 ) & 0xFF );
  bytes[3] = (u8) ( ( tmp >> 32 ) & 0xFF );
  bytes[4] = (u8) ( ( tmp >> 24 ) & 0xFF );
  bytes[5] = (u8) ( ( tmp >> 16 ) & 0xFF );
  bytes[6] = (u8) ( ( tmp >>  8 ) & 0xFF );
  bytes[7] = (u8) ( tmp & 0xFF );
#else
  bytes[0] = (u8) ( ( tmp.hi >> 24 ) & 0xFF );
  bytes[1] = (u8) ( ( tmp.hi >> 16 ) & 0xFF );
  bytes[2] = (u8) ( ( tmp.hi >>  8 ) & 0xFF );
  bytes[3] = (u8) ( tmp.hi & 0xFF );
  bytes[4] = (u8) ( ( tmp.lo >> 24 ) & 0xFF );
  bytes[5] = (u8) ( ( tmp.lo >> 16 ) & 0xFF );
  bytes[6] = (u8) ( ( tmp.lo >>  8 ) & 0xFF );
  bytes[7] = (u8) ( tmp.lo & 0xFF );
#endif
}

void v3u128Load( u128 * num, const u8 * bytes )
{
  u128 tmp;

  if ( ( num == NULL ) || ( bytes == NULL ) )
    return;

#if ( defined( CPU_64BIT ) )
  tmp.hi = ( (u64) bytes[0] << 56 ) | ( (u64) bytes[1] << 48 ) |
    ( (u64) bytes[2] << 40 ) | ( (u64) bytes[3] << 32 ) |
    ( (u64) bytes[4] << 24 ) | ( (u64) bytes[5] << 16 ) |
    ( (u64) bytes[6] << 8 ) | (u64) bytes[7];
  tmp.lo = ( (u64) bytes[8] << 56 ) | ( (u64) bytes[9] << 48 ) |
    ( (u64) bytes[10] << 40 ) | ( (u64) bytes[11] << 32 ) |
    ( (u64) bytes[12] << 24 ) | ( (u64) bytes[13] << 16 ) |
    ( (u64) bytes[14] << 8 ) | (u64) bytes[15];
#else
  tmp.hi.hi = ( (u32) bytes[0] << 24 ) | ( (u32) bytes[1] << 16 ) |
    ( (u32) bytes[2] << 8 ) | (u32) bytes[3];
  tmp.hi.lo = ( (u32) bytes[4] << 24 ) | ( (u32) bytes[5] << 16 ) |
    ( (u32) bytes[6] << 8 ) | (u32) bytes[7];
  tmp.lo.hi = ( (u32) bytes[8] << 24 ) | ( (u32) bytes[9] << 16 ) |
    ( (u32) bytes[10] << 8 ) | (u32) bytes[11];
  tmp.lo.lo = ( (u32) bytes[12] << 24 ) | ( (u32) bytes[13] << 16 ) |
    ( (u32) bytes[14] << 8 ) | (u32) bytes[15];
#endif
  *(num) = tmp;
}

void v3u128Save( u8 * bytes, const u128 * const num )
{
  u128 tmp;

  if ( ( num == NULL ) || ( bytes == NULL ) )
    return;

  tmp = *(num);

#if ( defined( CPU_64BIT ) )
  bytes[0]  = (u8) ( ( tmp.hi >> 56 ) & 0xFF );
  bytes[1]  = (u8) ( ( tmp.hi >> 48 ) & 0xFF );
  bytes[2]  = (u8) ( ( tmp.hi >> 40 ) & 0xFF );
  bytes[3]  = (u8) ( ( tmp.hi >> 32 ) & 0xFF );
  bytes[4]  = (u8) ( ( tmp.hi >> 24 ) & 0xFF );
  bytes[5]  = (u8) ( ( tmp.hi >> 16 ) & 0xFF );
  bytes[6]  = (u8) ( ( tmp.hi >>  8 ) & 0xFF );
  bytes[7]  = (u8) ( tmp.hi & 0xFF );
  bytes[8]  = (u8) ( ( tmp.lo >> 56 ) & 0xFF );
  bytes[9]  = (u8) ( ( tmp.lo >> 48 ) & 0xFF );
  bytes[10] = (u8) ( ( tmp.lo >> 40 ) & 0xFF );
  bytes[11] = (u8) ( ( tmp.lo >> 32 ) & 0xFF );
  bytes[12] = (u8) ( ( tmp.lo >> 24 ) & 0xFF );
  bytes[13] = (u8) ( ( tmp.lo >> 16 ) & 0xFF );
  bytes[14] = (u8) ( ( tmp.lo >>  8 ) & 0xFF );
  bytes[15] = (u8) ( tmp.lo & 0xFF );
#else
  bytes[0]  = (u8) ( ( tmp.hi.hi >> 24 ) & 0xFF );
  bytes[1]  = (u8) ( ( tmp.hi.hi >> 16 ) & 0xFF );
  bytes[2]  = (u8) ( ( tmp.hi.hi >>  8 ) & 0xFF );
  bytes[3]  = (u8) ( tmp.hi.hi & 0xFF );
  bytes[4]  = (u8) ( ( tmp.hi.lo >> 24 ) & 0xFF );
  bytes[5]  = (u8) ( ( tmp.hi.lo >> 16 ) & 0xFF );
  bytes[6]  = (u8) ( ( tmp.hi.lo >>  8 ) & 0xFF );
  bytes[7]  = (u8) ( tmp.hi.lo & 0xFF );
  bytes[8]  = (u8) ( ( tmp.lo.hi >> 24 ) & 0xFF );
  bytes[9]  = (u8) ( ( tmp.lo.hi >> 16 ) & 0xFF );
  bytes[10] = (u8) ( ( tmp.lo.hi >>  8 ) & 0xFF );
  bytes[11] = (u8) ( tmp.lo.hi & 0xFF );
  bytes[12] = (u8) ( ( tmp.lo.lo >> 24 ) & 0xFF );
  bytes[13] = (u8) ( ( tmp.lo.lo >> 16 ) & 0xFF );
  bytes[14] = (u8) ( ( tmp.lo.lo >>  8 ) & 0xFF );
  bytes[15] = (u8) ( tmp.lo.lo & 0xFF );
#endif
}

/* low level functions */

s32 v3_FileNativePath( ascii * native, const ascii * path )
{
  ascii * ptr, * src, * dest;
  u64 length;

  length = v3u64u32( V3_FILE_MAX_FILENAME );

  if ( ( ptr = v3StrCharA( path, (ascii) '|', length ) ) != NULL )
  {
    if ( v3u64Gt( v3StrLengthA( path ), length ) )
    {
      return( V3_FAIL );
    }
    src = (ascii *) path;
    dest = native;
    while ( *src != 0 )
    {
      if ( *src == '|' )
      {
        src++;
      }
      else
      {
        *(dest)++ = *(src)++;
      }
    }
  }
  else
  {
    if ( v3StrCopyA( native, path, v3u64Add( length, v3u64u32( 1 ) ) )
      != V3_PASS )
    {
      return( V3_FAIL );
    }
  }

  /* perform post copy substitutions */
#if ( V3_FILE_PATHMODE == V3_FILE_PATH_DOS )
  ptr = native;
  while ( ( ptr = v3StrCharA( ptr, '/', length ) ) != NULL )
  {
    *(ptr) = '\\';
  }
#elif ( V3_FILE_PATHMODE == V3_FILE_PATH_MAC )
  ptr = native;
  while ( ( ptr = v3StrCharA( ptr, '/', length ) ) != NULL )
  {
    *(ptr) = ':';
  }
#endif

  return( V3_PASS );
}

/*
  ###########################  AUDIT MARK  #################################
  ###########################    Duncan    #################################
  ###########################  AUDIT MARK  #################################
*/

u64 v3_FileWrite( v3_FILE * file, const void * const buffer, u64 length )
{
#if ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  long in_length;
#endif
  int real_write;
  u64 tmp;

#if ( OS_TYPE == OS_MACOS )
/* we use the file->handle as refnum, i.e., as a short */
#if ( defined( CPU_64BIT ) )
  in_length = (long) length;
  my_err = FSWrite( (short) v3u32u64( file->handle ), &in_length, buffer );
#else
  _V3_SET64( tmp, 00000000, 7FFFFFFF );
  if ( v3u64Gt( length, tmp ) )
  {
    return( v3u64u32( 0 ) );
  }
  in_length = v3u32u64( length );
  my_err = FSRead( (short) file->handle.lo, &in_length, buffer );
#endif
  if ( my_err == noErr )
  {
    real_write = in_length;
  }
  else
  {
    switch ( my_err )
    {
      case fLckdErr:
      case rfNumErr:
      case wrPermErr:
        file->last_error = V3_FILE_ERROR_DENIED;
        break;
      default:
        file->last_error = V3_FILE_ERROR_NOTFOUND;
    }
  }
#else
#if ( defined( V3_FILE64 ) )
  real_write = write( (int) file->handle, buffer, length );
#else
  _V3_SET64( tmp, 00000000, 7FFFFFFF );
  if ( v3u64Gt( length, tmp ) )
  {
    return( v3u64u32( 0 ) );
  }
  real_write = write( (int) v3u32u64( file->handle ), buffer,
    v3u32u64( length ) );
#endif
  if ( real_write == -1 )
  { /* Error during write() */
    if ( errno == EACCES )
    {
      file->last_error = V3_FILE_ERROR_DENIED;
    }
    else
    {
      file->last_error = V3_FILE_ERROR_NOTFOUND;
    }
#endif /* OS */

    return( v3u64u32( 0 ) );
  }

  if ( ( file->mode & V3_FILE_MODE_SYNC ) == V3_FILE_MODE_SYNC )
  {
#if ( ( OS_TYPE == OS_WIN32 ) || ( OS_TYPE == OS_OS2 ) )
    fflush( NULL );
#else
#if ( defined( V3_FILE64 ) )
    fsync( (int) file->handle );
#else
    fsync( (int) v3u32u64( file->handle ) );
#endif
#endif /* OS */
  }

  v3_FileTell( &tmp, file ); /* Set file->offset */

  if ( v3u64Gt( file->offset, file->length ) )
  { /* We are extending the file length */
    file->length = file->offset;
  }

#if ( defined( V3_FILE64 ) )
  return( real_write );
#else
  return( v3u64u32( (u32) real_write ) );
#endif
}

s32 v3_FileSeek( v3_FILE * file, u64 offset )
{
#if ( OS_TYPE == OS_MACOS)
  OSErr my_err;
#else
  off_t real_offset;
#endif

  if ( ( file->mode & V3_FILE_MODE_APPEND ) == V3_FILE_MODE_APPEND )
  {
    /* Seek in a file open for append is Forbidden */
    return( V3_FILE_ERROR_DENIED );
  }

#if ( OS_TYPE == OS_MACOS )
#if ( defined( CPU_64BIT ) )
  my_err = SetFPos( (short) v3u32u64( file->handle ), fsFromStart,
    (long) offset );
#else
  if ( v3u64Gt( length, v3u64u32( (u32) 0x7FFFFFFF ) ) )
  {
    file->last_error = V3_FILE_ERROR_SEEK;
    return( V3_FILE_ERROR_SEEK );
  }
  my_err = SetFPos( (short) v3u32u64( file->handle ), fsFromStart,
    v3u32u64( offset ) );
#endif
  if ( my_err == noErr )
  {
    return( V3_PASS );
  }
  else
  {
    file->last_error = V3_FILE_ERROR_NOTFOUND;
    return( V3_FILE_ERROR_NOTFOUND );
  }
#else /* OS */
#if ( defined( V3_FILE64 ) )
  real_offset = lseek( (int) file->handle, (off_t) offset, SEEK_SET );
#else
  if ( v3u64Gt( offset, v3u64u32( (u32) 0x7FFFFFFF ) ) )
  {
    file->last_error = V3_FILE_ERROR_SEEK;
    return( V3_FILE_ERROR_SEEK );
  }
  real_offset = lseek( (int) v3u32u64( file->handle ),
    (off_t) v3u32u64( offset ), SEEK_SET );
#endif

  if ( real_offset == -1 )
  { /* error during the lseek */
    file->last_error = V3_FILE_ERROR_SEEK;
    return( V3_FILE_ERROR_SEEK );
  }

#if ( defined( V3_FILE64 ) )
  file->offset = real_offset;
#else
  file->offset = v3u64u32( (u32) real_offset );
#endif

  return( V3_PASS );

#endif /* OS */
}

s32 v3_FileTell( u64 * offset, v3_FILE * file )
{
#if ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  long file_pos;
#else
  int real_offset;
#endif

#if ( OS_TYPE == OS_MACOS )
#if ( defined( CPU_64BIT ) )
  file_pos = (long) *offset;
#else
  file_pos = v3u32u64( *offset );
#endif
  my_err = GetFPos( (short) v3u32u64( file->handle ), &file_pos );
  if ( my_err == noErr )
  {
    *offset = v3u64u32( file_pos );
    return( V3_PASS );
  }
  else
  {
    file->last_error = V3_FILE_ERROR_NOTFOUND;
    return( file->last_error );
  }
#else /* OS */
#if ( defined( V3_FILE64 ) )
  real_offset = lseek( (int) file->handle, 0 , SEEK_CUR );
#else
  real_offset = lseek( (int) v3u32u64( file->handle ), 0 , SEEK_CUR );
#endif
  if ( real_offset == -1 )
  {
    file->last_error = V3_FILE_ERROR_TELL;
    return( V3_FILE_ERROR_TELL );
  }
#if ( defined( V3_FILE64 ) )
  *(offset) = real_offset;
#else
  *(offset) = v3u64u32( (u32) real_offset );
#endif

  return( V3_PASS );
#endif /* OS */
}

s32 v3_FileLength( u64 * length, v3_FILE * file )
{
#if ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  long file_length;
#else
  struct stat file_stats;
#endif

#if ( OS_TYPE == OS_MACOS )
  my_err = GetEOF((short) v3u32u64( file->handle ), &file_length);
  if ( my_err == noErr )
  {
    *length = v3u64u32( file_length );
    v3MutexUnlock( &file->lock );
    return( V3_PASS );
  }
  else
  {
    v3MutexUnlock( &file->lock );
    file->last_error = V3_FILE_ERROR_LENGTH;
    return( V3_FILE_ERROR_LENGTH );
  }
#else  /* OS */
#if ( defined( V3_FILE64 ) )
  if ( fstat( file->handle, &file_stats ) == -1 )
#else
  if ( fstat( v3u32u64( file->handle ), &file_stats ) == -1 )
#endif
  {
    file->last_error = V3_FILE_ERROR_LENGTH;
    return( V3_FILE_ERROR_LENGTH );
  }
#if ( defined( V3_FILE64 ) )
  *(length) = (u64) file_stats.st_size;
#else
  *(length) = v3u64u32( (u32) file_stats.st_size );
#endif
#endif /* OS */

  return( V3_PASS );
}

s32 v3_FileTruncate( v3_FILE * file, u64 length )
{
#if ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  long new_length;
#endif

  v3_FileWipe( file, length );

#if ( ( OS_TYPE == OS_WIN32 ) || ( OS_TYPE == OS_OS2 ) )
  if ( chsize( (int) v3u32u64( file->handle ),
    (size_t) v3u32u64( length ) ) != 0 )
  {
    if ( errno == EBADF )
    {
      file->last_error = V3_FILE_ERROR_NOTFOUND;
      return( file->last_error );
    }
    else if ( errno == EACCES )
    {
      file->last_error = V3_FILE_ERROR_DENIED;
      return( file->last_error );
    }
  }
#elif ( OS_TYPE == OS_MACOS )
#if ( defined( CPU_64BIT ) )
  /* !!! Mac: fix */
  new_length = (long) length;
#else
  new_length = v3u32u64( length );
#endif
  my_err = SetEOF( (short) v3u32u64( file->handle ), new_length );
  if ( my_err != noErr )
  {
    switch( my_err )
    {
      case fLckdErr:
      case rfNumErr:
      case wrPermErr:
        file->last_error = V3_FILE_ERROR_DENIED;
        break;
      default:
        file->last_error = V3_FILE_ERROR_NOTFOUND;
    }
    return( file->last_error );
  }
#else
#if ( defined( V3_FILE64 ) )
  if ( ftruncate( (int) file->handle, (size_t) length ) != 0 )
#else
  if ( ftruncate( (int) v3u32u64( file->handle ),
    (size_t) v3u32u64( length ) ) != 0 )
#endif /* file64 */
  {
    /* error during the ftruncate */
    if ( errno == EBADF )
    {
      file->last_error = V3_FILE_ERROR_NOTFOUND;
      return( file->last_error );
    }
    else if ( errno == EINVAL )
    {
      file->last_error = V3_FILE_ERROR_WRITEMODE;
      return( file->last_error );
    }
  }
#endif /* OS */

  file->length = length;
  v3_FileSeek( file, length );

  return( V3_PASS );
}

s32 v3_FileLock( v3_FILE * file, u32 lock )
{
#if ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  FSSpec spec;
  v3_FILENAME native_filename;
#endif

#if ( OS_TYPE == OS_LINUX )
  if ( lock == V3_FILE_LOCK_READ )
  {
#if ( defined( V3_FILE64 ) )
    if ( flock( (int) file->handle, LOCK_SH | LOCK_NB ) == -1 )
#else
    if ( flock( (int) v3u32u64( file->handle ), LOCK_SH | LOCK_NB ) == -1 )
#endif
    {
      file->last_error = V3_FILE_ERROR_READLOCK;
      return( file->last_error );
    }
  }
  else
  {
    if ( lock == V3_FILE_LOCK_WRITE )
    {
#if ( defined( V3_FILE64 ) )
      if ( flock( (int) file->handle, LOCK_EX | LOCK_NB ) == -1 )
#else
      if ( flock( (int) v3u32u64( file->handle ), LOCK_EX | LOCK_NB ) == -1 )
#endif
      {
        file->last_error = V3_FILE_ERROR_WRITELOCK;
        return( file->last_error );
      }
    }
  }
#elif ( ( OS_TYPE == OS_SOLARIS ) || ( OS_TYPE == OS_SUNOS ) )
  struct flock lock_desc;

  if ( lock == V3_FILE_LOCK_READ )
  {
    /* With thus values, we should lock all the file */
    lock_desc.l_type = F_RDLCK;
    lock_desc.l_whence = 0;
    lock_desc.l_start = 0;
    lock_desc.l_len = 0;

#if ( defined( V3_FILE64 ) )
    if ( fcntl( (int) file->handle, F_SETLK, &lock_desc ) == -1 )
#else
    if ( fcntl( (int) v3u32u64( file->handle ), F_SETLK, &lock_desc ) == -1 )
#endif
    {
      file->last_error = V3_FILE_ERROR_READLOCK;
      return( file->last_error );
    }
  }
  else
  {
    if ( lock == V3_FILE_LOCK_WRITE )
    {
      /* With thus values, we should lock all the file */
      lock_desc.l_type = F_WRLCK;
      lock_desc.l_whence = 0;
      lock_desc.l_start = 0;
      lock_desc.l_len = 0;

#if ( defined( V3_FILE64 ) )
      if ( fcntl( (int) file->handle, F_SETLK, &lock_desc ) == -1 )
#else
      if ( fcntl( (int) v3u32u64( file->handle ), F_SETLK,
        &lock_desc ) == -1 )
#endif
      {
        file->last_error = V3_FILE_ERROR_WRITELOCK;
        return( file->last_error );
      }
    }
  }
#elif ( OS_TYPE == OS_MACOS )
/* We only have write locks */
  /* Convert slashs to native format and put it into native_filename */
  if ( v3_FileNativePath( native_filename, file->filename ) != V3_PASS )
  {
    return( V3_FILE_ERROR_NAME );
  }

  my_err = FSMakeFSSpec( 0, 0, native_filename, &spec );
  if ( my_err != noErr )
  {
    return( V3_FILE_ERROR_NOTFOUND );
  }

  my_err = FSpSetFLock( &spec );
  if ( my_err != noErr )
  {
    file->last_error = V3_FILE_ERROR_WRITELOCK;
    return( file->last_error );
  }
#else /* no OS locks, we need to do them ourselves */
  /* open a file with ~ at the end, place ProcessID and lock type in it */



#endif

  file->lockmode = lock;
  return( V3_PASS );
}

s32 v3_FileUnLock( v3_FILE * file )
{
#if ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  FSSpec spec;
  v3_FILENAME native_filename;
#endif

#if ( OS_TYPE == OS_LINUX )
#if  ( defined( V3_FILE64 ) )
  if ( flock( (int) file->handle, LOCK_UN ) == -1 )
#else
  if ( flock( (int) v3u32u64( file->handle ), LOCK_UN ) == -1 )
#endif
  {
    file->last_error = V3_FILE_ERROR_UNLOCK;
    return( file->last_error );
  }
#elif ( OS_TYPE == OS_SOLARIS )
  struct flock lock_desc;

  /* With thus values, we should lock all the file */
  lock_desc.l_type = F_UNLCK;
  lock_desc.l_whence = 0;
  lock_desc.l_start = 0;
  lock_desc.l_len = 0;
#if ( defined( V3_FILE64 ) )
  if ( fcntl( (int) file->handle, F_SETLK, &lock_desc ) == -1 )
#else
  if ( fcntl( (int) v3u32u64( file->handle ), F_SETLK, &lock_desc ) == -1 )
#endif
  {
    file->last_error = V3_FILE_ERROR_READLOCK;
    return( file->last_error );
  }
#elif ( OS_TYPE == OS_MACOS )
  if ( v3_FileNativePath( native_filename, file->filename ) != V3_PASS )
  {
    return( V3_FILE_ERROR_NAME );
  }

  my_err = FSMakeFSSpec(0, 0, native_filename, &spec);
  if ( my_err != noErr )
  {
    return( V3_FILE_ERROR_NOTFOUND );
  }

  my_err = FSpRstFLock( &spec );
  if ( my_err != noErr )
  {
    file->last_error = V3_FILE_ERROR_WRITELOCK;
    return( file->last_error );
  }
#else /* no OS locks, we need to do them ourselves */

#endif

  file->lockmode = V3_FILE_LOCK_NONE;
  return( V3_PASS );
}

s32 v3_FileWipe( v3_FILE * file, u64 length )
{
  u64 file_size;       /* Size of the file to wipe */
  void * zero_land;    /* Zero full memory to wiped the file with */
  u64 zero_land_size;  /* Size of zero_land space */
  u64 zero;            /* Use to wipe if can't allocate zero_land */
  u64 to_wipe;
  u64 really_write;
  s32 error;

  if ( ( error = v3_FileLength( &file_size, file ) ) != V3_PASS )
  {
    file->last_error = error;
    return( file->last_error );
  }

  if ( ( error = v3_FileSeek( file, length ) ) != V3_PASS )
  {
    file->last_error = error;
    return( file->last_error );
  }

  _V3_SET64( zero_land_size, 00000000, 00001000 );
  if ( ( zero_land = v3MemAlloc( zero_land_size, V3_MEM_NORMAL ) ) == NULL )
  {
    _V3_SET64( zero_land_size, 00000000, 00000008 );
    zero = v3u64u32( (u32) 0 );
    zero_land = &zero;
  }

/* !!! fix error handling */

  /* seek to length, while < file_size write data */
  to_wipe = v3u64Sub( file_size, length );
  do
  {
    /* !!! problem with lock status and write... */
    if ( v3u64Gt( zero_land_size, to_wipe ) )
    {
      really_write = v3_FileWrite( file, zero_land, to_wipe );
    }
    else
    {
      really_write = v3_FileWrite( file, zero_land, zero_land_size );
    }
    to_wipe = v3u64Sub( to_wipe, really_write );
  } while ( !v3u64Eq( really_write, v3u64u32( (u32) 0 ) ) );

  v3MemFree( zero_land );
  return( V3_PASS );
}

/* testing */

s32 v3_TestCosmFile( void )
{
  v3_DIR * testdir;
  v3_FILE * testfile;
  u8 * testfileimg;  /* Image of the file as it is wrote */
  u8   testpattern1[31] = "1234568790\nabcdefghi\nABCDEFGHI";
  u8   buffer[100];
  s32  i, j;
  u64  token_len;  /* len of the tokens we are reading/writting */
  u64  real_write;
  u64  real_read;
  u64  real_length;
  u64  real_offset;
  u8 * hello = (u8*) "Hello World !";
  v3_FILENAME testfilenames [16];
  s32 error;

  const u8 bytes1[16] = /* the number 0x0102030405060708090A0B0C0D0E0F10 */
  {
    0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
    0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x10
  };
  u8 bytes2[16];
  u16  b_u16;
  u32  b_u32;
  u64  b_u64, e_u64;
  u128 b_u128, e_u128;

/***********************************************************************/
/* File test 1                                                         */
/*                                                                     */
/* Test functions : v3FileOpen, v3FileRead, v3FileWrite, v3FileSeek,   */
/*   v3FileTell, v3FileEOF, v3FileLength, v3FileTruncate, v3FileClose  */
/*                                                                     */
/* Methodology : ;)                                                    */
/* -------------                                                       */
/* 01. Create the file cosmfile.tst (read and write, exclusive lock)   */
/* 02. Write 20 patterns of 10 bytes (bytes = u8)                      */
/* 03. Check we are at end of file                                     */
/* 04. Check current size of the file (should be 200)                  */
/* 05. Move to the begining of the file                                */
/* 06. Read and compare 8 patterns of 25 bytes                         */
/* 07. Check we are at end of file                                     */
/* 08. Truncate file at 100 bytes                                      */
/* 09. Check current size of the file (should be 100)                  */
/* 10. Move to the begining of the file                                */
/* 11. Read and compare 1 pattern of 100 bytes                         */
/* 12. Check we are at end of file                                     */
/* 13. Write 26 bytes                                                  */
/* 14. Check current size of the file (should be 126)                  */
/* 15. Check we are at end of file                                     */
/* 16. Read 25 patterns of 6 bytes, with an offset of (-5), from EOF   */
/* 17. Checking position in file (offset should be 5)                  */
/* 18. Close file                                                      */
/* 19. Open file cosmfile.tst (write only, exclusive lock)             */
/* 20. Write the string 'Hello world !' to file                        */
/* 21. Close file                                                      */
/* 22. Open file cosmfile.tst (read only, shared read lock)            */
/* 23. Read and compare 10 patterns of 12 bytes                        */
/* 24. Read and compare 1 pattern of 6 bytes                           */
/* 25. Check we are at end of file                                     */
/* 26. Close file                                                      */
/* 27. Delete file                                                     */
/***********************************************************************/

  /* Allocate memory for all tests */

#if ( defined( CPU_64BIT ) )
  testfile = (v3_FILE *) v3MemAlloc( sizeof( v3_FILE ), V3_MEM_NORMAL );
#else
  testfile = (v3_FILE *) v3MemAlloc( v3u64u32( sizeof( v3_FILE ) ),
    V3_MEM_NORMAL );
#endif

  if ( testfile == NULL )
  {
    return( 1 ); /* we already tested v3MemAlloc! */
  }

#if ( defined( CPU_64BIT ) )
  testfileimg = (u8 *) v3MemAlloc( 200 , V3_MEM_NORMAL );
#else
  testfileimg = (u8 *) v3MemAlloc( v3u64u32( 200 ), V3_MEM_NORMAL );
#endif
  if ( testfileimg == NULL )
  {
    return( 1 );
  }

#if ( defined( CPU_64BIT ) )
  testdir = (v3_DIR *) v3MemAlloc( sizeof( v3_DIR ), V3_MEM_NORMAL );
#else
  testdir = (v3_DIR *) v3MemAlloc( v3u64u32( sizeof( v3_DIR ) ),
    V3_MEM_NORMAL );
#endif
  if ( testdir == NULL )
  {
    return( 1 );
  }

/* Test 1. Phase 1. Open file 'cosmfile.tst' */

  error = v3FileOpen( testfile, (ascii *) "cosmfile.tst",
    V3_FILE_MODE_CREATE | V3_FILE_MODE_READ | V3_FILE_MODE_WRITE,
    V3_FILE_LOCK_WRITE );
  if ( error != V3_PASS )
  { /* Open was not succesfull */
    return( -1 );
  }

/* Test 1. Phase 2. Writing tokens */

#if ( defined( CPU_64BIT ) )
  token_len = 10;
#else
  token_len = v3u64u32( 10 );
#endif

  for ( i = 0 ; i < 20 ; i++ )
  {
    real_write = v3FileWrite( testfile, (testpattern1 + i), token_len );
    if ( !v3u64Eq( real_write, token_len ) )
    { /* Write was not succesful */
      return( -2 );
    }
    v3MemCopy ( &testfileimg[i * 10], &testpattern1[i], token_len );
  }

/* Test 1. Phase 3. Check we are at end of file */

  error = v3FileEOF( testfile );
  if ( error != V3_FILE_ERROR_EOF )
  { /* Even a real error, or the library think there is more to read */
    return( -3 );
  }

/* Test 1. Phase 4. Check current size of the file (should be 200) */

  error = v3FileLength( &real_length, testfile );
  if ( error != V3_PASS )
  { /* Length was not succesfull */
    return( -4 );
  }
#if ( defined( CPU_64BIT ) )
  if ( real_length != 200 )
#else
  if ( !v3u64Eq( real_length, v3u64u32( 200 ) ) )
#endif
  { /* Wrong length returned */
    return( -5 );
  }

/* Test 1. Phase 5. Move to the beginning of file */

#if ( defined( CPU_64BIT) )
  error = v3FileSeek( testfile, 0 );
#else
  error = v3FileSeek( testfile, v3u64u32( 0 ) );
#endif
  if ( error != V3_PASS )
  { /* Seek was not successfull */
    return( -6 );
  }


/* Test 1. Phase 6. Read and compare tokens */
#if ( defined( CPU_64BIT ) )
  token_len = 25;
#else
  token_len = v3u64u32( 25 );
#endif

  for ( i = 0 ; i < 8 ; i++ )
  {
    real_read = v3FileRead( buffer, token_len, testfile );
    if ( !v3u64Eq( real_read, token_len ) )
    { /* Read was not succesfull or read wrong size */
       return( -7 );
    }
    for ( j = 0 ; j < 25 ; j++ )
    {
      if ( buffer[j]!= testfileimg[j + (i * 25)] )
      {
        v3FileClose( testfile );
        return( -8 );
      }
    }
  }

/* Test 1. Phase 7. Check we are at end of file */

  error = v3FileEOF( testfile );
  if ( error != V3_FILE_ERROR_EOF )
  { /* Even a real error, or the library think there is more to read */
    return( -9 );
  }

/* Test 1. Phase 8. Truncate file at 100 bytes */

#if ( defined( CPU_64BIT ) )
  error = v3FileTruncate( testfile, 100 );
#else
  error = v3FileTruncate( testfile, v3u64u32( 100 ) );
#endif
  if ( error != V3_PASS )
  { /* Truncate was not succesfull */
    return( -10 );
  }

/* Test 1. Phase 9. Check current size of the file (should be 100) */

  error = v3FileLength( &real_length, testfile );
  if ( error != V3_PASS )
  { /* Length was not succesfull */
    return( -11 );
  }
#if ( defined( CPU_64BIT ) )
  if ( real_length != 100 )
#else
  if ( !v3u64Eq( real_length, v3u64u32( 100 ) ) )
#endif
  { /* Wrong length returned */
    return( -12 );
  }

/* Test 1. Phase 10. Move to the beginning of file */

#if ( defined( CPU_64BIT) )
  error = v3FileSeek( testfile, 0 );
#else
  error = v3FileSeek( testfile, v3u64u32( 0 ) );
#endif
  if ( error != V3_PASS )
  { /* Seek was not successfull */
    return( -13 );
  }

/* Test 1. Phase 11. Read and compare tokens */

#if ( defined( CPU_64BIT ) )
  token_len = 100;
#else
  token_len = v3u64u32( 100 );
#endif

  real_read = v3FileRead( buffer, token_len, testfile );
  if ( !v3u64Eq( real_read, token_len ) )
  { /* Read was not succesfull or read wrong size */
     return( -14 );
  }
  for ( j = 0 ; j < 100 ; j++ )
  {
    if ( buffer[j] != testfileimg[j] )
    {
      v3FileClose( testfile );
      return( -15 );
    }
  }

/* Test 1. Phase 12. Check we are at end of file */

  error = v3FileEOF( testfile );
  if ( error != V3_FILE_ERROR_EOF )
  { /* Even a real error, or the library think there is more to read */
    return( -16 );
  }

/* Test 1. Phase 13. Writing 26 bytes */

#if ( defined( CPU_64BIT ) )
  token_len = 26;
#else
  token_len = v3u64u32( 26 );
#endif

  real_write = v3FileWrite( testfile, &testfileimg[100], token_len );
  if ( !v3u64Eq( real_write, token_len ) )
  { /* Write was not succesful */
    return( -17 );
  }

/* Test 1. Phase 14. Check current size of the file (should be 126) */

  error = v3FileLength( &real_length, testfile );
  if ( error != V3_PASS )
  { /* Length was not succesfull */
    return( -18 );
  }
#if ( defined( CPU_64BIT ) )
  if ( real_length != 126 )
#else
  if ( !v3u64Eq( real_length, v3u64u32( 126 ) ) )
#endif
  { /* Wrong length returned */
    return( -19 );
  }

/* Test 1. Phase 15. Check we are at end of file */

  error = v3FileEOF( testfile );
  if ( error != V3_FILE_ERROR_EOF )
  { /* Even a real error, or the library think there is more to read */
    return( -20 );
  }

/* Test 1. Phase 16. Read and compare tokens from end of file */

#if ( defined( CPU_64BIT ) )
  token_len = 5;
#else
  token_len = v3u64u32( 5 );
#endif

  for ( i = 24 ; i > -1 ; i-- )
  {
#if ( defined( CPU_64BIT) )
    error = v3FileSeek( testfile, i * 5 );
#else
    error = v3FileSeek( testfile, v3u64u32( i * 5 ) );
#endif
    if ( error != V3_PASS )
    { /* Seek was not successfull */
      return( -21 );
    }

    real_read = v3FileRead( buffer, token_len, testfile );
    if ( !v3u64Eq( real_read, token_len ) )
    { /* Read was not succesfull or read wrong size */
       return( -22 );
    }
    for ( j = 0 ; j < 5 ; j++ )
    {
      if ( buffer[j]!= testfileimg[j + (i * 5)] )
      {
        v3FileClose( testfile );
        return( -23 );
      }
    }
  }

/* Test 1. Phase 17. Checking position (offset should be 5) */

  error = v3FileTell( &real_offset , testfile );
  if ( error != V3_PASS )
  { /* Tell was not succesfull */
    return( -24 );
  }

#if ( defined( CPU_64BIT ) )
  if ( !v3u64Eq( real_offset, 5 ) )
#else
  if ( !v3u64Eq( real_offset, v3u64u32( 5 ) ) )
#endif
  { /* Returned position is wrong */
    return( -25 );
  }

/* Test 1. Phase 18. Close file */

  if ( v3FileClose( testfile ) != V3_PASS )
  {
    return( -26 );
  }

/* Test 1. Phase 19. Open file 'cosmfile.tst' */

  error = v3FileOpen( testfile, (ascii *) "cosmfile.tst", V3_FILE_MODE_WRITE,
    V3_FILE_LOCK_WRITE );
  if ( error != V3_PASS )
  { /* Open was not succesfull */
    return( -27 );
  }

/* Test 1. Phase 20. Write the Hello phrase */

#if ( defined( CPU_64BIT ) )
  token_len = 13;
#else
  token_len = v3u64u32( 13 );
#endif

  real_write = v3FileWrite ( testfile, hello, token_len );
  if ( !v3u64Eq( real_write, token_len ) )
  { /* Write was not succesful */
    return( -28 );
  }
  v3MemCopy( testfileimg, hello, token_len );

/* Test 1. Phase 21. Close file */

  if ( v3FileClose( testfile ) != V3_PASS )
  {
    return( -29 );
  }

/* Test 1. Phase 22. Open file 'cosmfile.tst' */

  error = v3FileOpen( testfile, (ascii *) "cosmfile.tst", V3_FILE_MODE_READ,
    V3_FILE_LOCK_READ );
  if ( error != V3_PASS )
  { /* Open was not succesfull */
    return( -30 );
  }

/* Test 1. Phase 23. Read and compare tokens */

#if ( defined( CPU_64BIT ) )
  token_len = 12;
#else
  token_len = v3u64u32( 12 );
#endif

  for ( i = 0 ; i < 10 ; i++ )
  {
    real_read = v3FileRead( buffer, token_len, testfile );
    if ( !v3u64Eq( real_read, token_len ) )
    { /* Read was not succesfull or read wrong size */
       return( -31 );
    }
    for ( j = 0 ; j < 12 ; j++ )
    {
      if ( buffer[j]!= testfileimg[j + (i * 12)] )
      {
        v3FileClose( testfile );
        return( -32 );
      }
    }
  }

/* Test 1. Phase 24. Read and compare 6 bytes */

#if ( defined( CPU_64BIT ) )
  token_len = 6;
#else
  token_len = v3u64u32( 6 );
#endif

  real_read = v3FileRead( buffer, token_len, testfile );
  if ( !v3u64Eq( real_read, token_len ) )
  { /* Read was not succesfull or read wrong size */
     return( -33 );
  }
  for ( j = 0 ; j < 6 ; j++ )
  {
    if ( buffer[j] != testfileimg[j + 120] )
    {
      v3FileClose( testfile );
      return( -34 );
    }
  }

/* Test 1. Phase 25. Check we are at end of file */

  error = v3FileEOF( testfile );
  if ( error != V3_FILE_ERROR_EOF )
  { /* Even a real error, or the library think there is more to read */
    return( -35 );
  }

/* Test 1. Phase 26. Close file */

  if ( v3FileClose( testfile ) != V3_PASS )
  {
    return( -36 );
  }

/* Test 1. Phase 27. Delete file 'cosmfile.tst' */

  if ( v3FileDelete( (ascii *) "cosmfile.tst" ) != V3_PASS )
  { /* Delete was not succesfull */
    return( -37 );
  }

/***********************************************************************/
/* File test 2                                                         */
/*                                                                     */
/* Test functions : v3DirOpen, v3DirRead, v3DirClose                   */
/*                                                                     */
/* Methodology :                                                       */
/* -------------                                                       */
/* 01. Open current directory                                          */
/* 02. Read current directory content                                  */
/* 03. Close the directory                                             */
/* 04. Create a directory 'testdir.tmp'                                */
/* 05. Create a directory 'testdir.tmp/level2'                         */
/* 06. Open directory 'testdir.tmp'                                    */
/* 07. Create file 'testdir.tmp/tiny.tmp' (write only, exclusive lock) */
/* 08. Write the 'Hello world !' string in it                          */
/* 09. Close the file                                                  */
/* 10. Read content of the directory and check                         */
/* 11. Close directory 'testdir.tmp'                                   */
/* 12. Delete directory 'testdir.tmp/level2'                           */
/* 13. Try to delete directory 'testdir.tmp', should fail (not empty)  */
/* 14. Delete file 'testdir.tmp/tiny.tmp'                              */
/* 15. Delete directory 'testdir.tmp'                                  */
/***********************************************************************/

/* Test 2. Phase 1. Open current directory */

  if ( v3DirOpen( testdir, (ascii *) ".", V3_DIR_MODE_READ ) != V3_PASS )
  { /* Open a directory was not succesfull */
    return( -38 );
  }

/* Test 2. Phase 2. Read current directory content */

#if ( defined( CPU_64BIT ) )
  do
  {
    real_read = v3DirRead( testfilenames, 16, testdir );
    if ( ( real_read == 0 ) && ( testdir->last_error != V3_DIR_ERROR_EOF ) )
    { /* We didn't reach EOF but get an error instead */
      return( -39 );
    }
  } while ( real_read != 0 );
#else
  do
  {
    real_read = v3DirRead( testfilenames, v3u64u32( 16 ), testdir );
    if ( v3u64Eq( real_read, v3u64u32( 0 ) ) &&
      ( testdir->last_error != V3_DIR_ERROR_EOF ) )
    { /* We didn't reach EOF but get an error instead */
      return( -40 );
    }
  } while ( !v3u64Eq( real_read, v3u64u32( 0 ) ) );
#endif

/* Test 2. Phase 3. Close directory */

  if ( v3DirClose( testdir ) != V3_PASS )
  { /* Close a directory was not succesfull */
    return( -41 );
  }

/* Test 2. Phase 4. Create directory 'testdir.tmp' */

  if ( v3DirOpen( testdir, (ascii *) "testdir.tmp", V3_DIR_MODE_CREATE )
    != V3_PASS )
  { /* Create a directory was not succesfull */
    return( -42 );
  }

/* Test 2. Phase 5. Create directory 'testdir.tmp/level2/level3/level4/' */

  if ( v3DirOpen( testdir, (ascii *) "testdir.tmp/level2/level3/level4/",
    V3_DIR_MODE_CREATE ) != V3_PASS )
  { /* Create a directory was not succesfull */
    return( -43 );
  }

/* Test 2. Phase 6. Open directory 'testdir.tmp' */

  if ( v3DirOpen( testdir, (ascii *) "testdir.tmp", V3_DIR_MODE_READ )
    != V3_PASS )
  { /* Open a directory was not succesfull */
    return( -44 );
  }

/* Test 2. Phase 7. Open file 'testdir.tmp/tiny.tmp' */

  error = v3FileOpen( testfile, (ascii *) "testdir.tmp/tiny.tmp",
    V3_FILE_MODE_CREATE | V3_FILE_MODE_WRITE, V3_FILE_LOCK_WRITE );
  if ( error != V3_PASS )
  { /* Open was not succesfull */
    return( -45 );
  }

/* Test 2. Phase 8. Write the Hello phrase */

#if ( defined( CPU_64BIT ) )
  token_len = 13;
#else
  token_len = v3u64u32( 13 );
#endif

  real_write = v3FileWrite ( testfile, hello, token_len );
  if ( !v3u64Eq( real_write, token_len ) )
  { /* Write was not succesful */
    return( -46 );
  }

/* Test 2. Phase 9. Close file */

  if ( v3FileClose( testfile ) != V3_PASS )
  {
    return( -47 );
  }

/* Test 2. Phase 10. Read content of the directory and check */

/* Test 2. Phase 11. Close directory */
  if ( v3DirClose( testdir ) != V3_PASS )
  { /* Close a directory was not succesfull */
    return( -48 );
  }

/* Test 2. Phase 12. Delete directory 'testdir.tmp/level2/level3/level4' */

  if ( v3DirDelete( (ascii *) "testdir.tmp/level2/level3/level4/" )
    != V3_PASS )
  { /* Delete a directory was not succesfull */
    return( -49 );
  }

/* Test 2. Phase 13. Try to delete directory 'testdir.tmp', should fail */

  if ( v3DirDelete( (ascii *) "testdir.tmp" ) == V3_PASS )
  { /* Delete a directory was succesfull -> Wrong thing */
    return( -50 );
  }

/* Test 2. Phase 14. Delete file 'testdir.tmp/tiny.tmp' */

  if ( v3FileDelete( (ascii *) "testdir.tmp/tiny.tmp" ) != V3_PASS )
  { /* Delete was not succesfull */
    return( -51 );
  }

/* Test 2. Phase 15. Delete 'testdir.tmp tree' */

  if ( ( v3DirDelete( (ascii *) "testdir.tmp/level2/level3" ) != V3_PASS ) ||
    ( v3DirDelete( (ascii *) "testdir.tmp/level2" ) != V3_PASS ) ||
    ( v3DirDelete( (ascii *) "testdir.tmp" ) != V3_PASS ) )
  { /* Delete a directory was not succesfull */
    return( -52 );
  }

  v3MemFree( testdir );
  v3MemFree( testfile );
  v3MemFree( testfileimg );

/*
  Test functions : v3*Load, v3*Save
*/

  v3u16Load( &b_u16, bytes1 );
  if ( b_u16 != 0x0102 )
  {
    return( -53 );
  }

  v3u16Save( bytes2, &b_u16 );
  for ( i = 0 ; i < 2 ; i++ )
  {
    if ( bytes1[i] != bytes2[i] )
    {
      return( -54 );
    }
  }

  v3u32Load( &b_u32, bytes1 );
  if ( b_u32 != 0x01020304 )
  {
    return( -55 );
  }

  v3u32Save( bytes2, &b_u32 );
  for ( i = 0 ; i < 4 ; i++ )
  {
    if ( bytes1[i] != bytes2[i] )
    {
      return( -56 );
    }
  }

  v3u64Load( &b_u64, bytes1 );
  _V3_SET64( e_u64, 01020304, 05060708 );
  if ( !v3u64Eq( b_u64, e_u64 ) )
  {
    return( -57 );
  }

  v3u64Save( bytes2, &b_u64 );
  for ( i = 0 ; i < 8 ; i++ )
  {
    if ( bytes1[i] != bytes2[i] )
    {
      return( -58 );
    }
  }

  v3u128Load( &b_u128, bytes1 );
  _V3_SET128( e_u128, 01020304, 05060708, 090A0B0C, 0D0E0F10 );
  if ( !v3u128Eq( b_u128, e_u128 ) )
  {
    return( -59 );
  }

  v3u128Save( bytes2, &b_u128 );
  for ( i = 0 ; i < 16 ; i++ )
  {
    if ( bytes1[i] != bytes2[i] )
    {
      return( -60 );
    }
  }

  return( V3_PASS );
}
