/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

/* CPU/OS Layer - CPU and OS specific code is allowed */

#ifndef V3_COSMIO_H
#define V3_COSMIO_H

#include "cputypes.h"
#include "cosmmath.h"
#include "cosmfile.h"
#include <stdarg.h>

/*
  Standard input and output functions.
*/

#define V3_IO_ECHO     0
#define V3_IO_NOECHO   1

#define V3_IO_NOWAIT   13
#define V3_IO_WAIT     14

#define V3_IO_UNSIGNED 0
#define V3_IO_SIGNED   1

/*
  The following are defines/macros for the ANSI escape codes. Keep in mind
  that not all devices are ANSI, so use them only in fitting applications.
  If you're doing large amounts of ANSI, embedding them into strings
  will be slightly faster (if less readable), and this table will be a
  convienient reference.
*/
#define ANSI_CLEAR         "\033[0m"  /* clear all modes (rst to last save) */
#define ANSI_HIGH          "\033[1m"  /* high intensity */
#define ANSI_LOW           "\033[2m"  /* low intensity */
#define ANSI_UNDERLINE_ON  "\033[4m"
#define ANSI_UNDERLINE_OFF "\033[24m"
#define ANSI_BLINK_ON      "\033[5m"
#define ANSI_BLINK_OFF     "\033[25m"
#define ANSI_INVERSE_ON    "\033[7m"
#define ANSI_INVERSE_OFF   "\033[27m"
#define ANSI_INVISIBLE     "\033[8m"  /* no display, this is NOT security */
#define ANSI_SAVE          "\033[s"   /* Saves cursor position */
#define ANSI_RESTORE       "\033[u"   /* Return to saved cursor position */
#define ANSI_CLEARSCREEN   "\033[2J"  /* Clear screen and home cursor */
#define ANSI_CLEARLINE     "\033[K"   /* Clear to end of line */
#define ANSI_FG_BLACK      "\033[30m" /* foreground colors */
#define ANSI_FG_RED        "\033[31m"
#define ANSI_FG_GREEN      "\033[32m"
#define ANSI_FG_YELLOW     "\033[33m"
#define ANSI_FG_BLUE       "\033[34m"
#define ANSI_FG_MAGENTA    "\033[35m"
#define ANSI_FG_CYAN       "\033[36m"
#define ANSI_FG_WHITE      "\033[37m"
#define ANSI_BG_BLACK      "\033[40m" /* background colors */
#define ANSI_BG_RED        "\033[41m"
#define ANSI_BG_GREEN      "\033[42m"
#define ANSI_BG_YELLOW     "\033[43m"
#define ANSI_BG_BLUE       "\033[44m"
#define ANSI_BG_MAGENTA    "\033[45m"
#define ANSI_BG_CYAN       "\033[46m"
#define ANSI_BG_WHITE      "\033[47m"
#define _ANSI_MOVE( line, column ) \
  v3PrintA( (ascii *) "\033[%u;%uH", line, column )
#define _ANSI_UP( count ) \
  v3PrintA( (ascii *) "\033[%uA", count )
#define _ANSI_DOWN( count ) \
  v3PrintA( (ascii *) "\033[%uB", count )
#define _ANSI_FORWARD( count ) \
  v3PrintA( (ascii *) "\033[%uC", count )
#define _ANSI_BACK( count ) \
  v3PrintA( (ascii *) "\033[%uD", count )

u64 v3Input( u8 * buffer, u64 length, u32 echo, u32 wait );
  /*
    Read up to length bytes from stdin into the buffer. If echo is
    V3_IO_NOECHO then the typed characters will not be echoed in any way,
    otherwise echo should be V3_IO_ECHO.
    If wait is V3_IO_NOWAIT, then return immediately if not enough input is
    available. If wait is V3_IO_WAIT then do not return until length bytes
    are available.
    Returns: Number of bytes read, 0xFFFFFFFFFFFFFFFF (-1) indicates
      EOF.
  */

u64 v3InputA( ascii * buffer, u64 max_chars, u32 echo );
  /*
    Reads up to max_chars-1 ascii characters from standard input device
    until EOF or '\n' is reached. buffer will always be a terminated
    string if input is read. If echo is V3_IO_NOECHO then the typed
    characters will not be echoed in any way (for password entry etc),
    otherwise echo should be V3_IO_ECHO. Any input beyond max_chars-1
    will be discarded, so use a large enough buffer.
    Returns: Number of characters input, 0xFFFFFFFFFFFFFFFF (-1) indicates
      EOF.
  */

u64 v3InputU( unicode * buffer, u64 max_chars, u32 echo );
  /*
    Reads up to max_chars-1 unicode characters from standard input device
    until EOF or '\n' is reached. buffer will always be a terminated
    string if input is read. If echo is V3_IO_NOECHO then the typed
    characters will not be echoed in any way (for password entry etc),
    otherwise echo should be V3_IO_ECHO.
    Returns: Number of characters input, 0xFFFFFFFFFFFFFFFF (-1) indicates
      EOF.
  */

u64 v3StrLengthA( const ascii * string );
  /*
    Get the length of the ascii string.
    Returns: the length of the ascii string in characters.
  */

u64 v3StrLengthU( const unicode * string );
  /*
    Get the length of the unicode string.
    Returns: the length of the unicode string in characters.
  */

s32 v3StrCopyA( ascii * dest, const ascii * src, u64 max_chars );
  /*
    Copy a string of up to max_length-1 ascii characters from src to dest,
    including the terminating \0. If the src string is longer then
    max_chars-1 the function will fail and do nothing.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3StrCopyU( unicode * dest, const unicode * src, u64 max_chars );
  /*
    Copy a string of up to max_length-1 unicode characters from src to dest,
    including the terminating \0. If the src string is longer then
    max_chars-1 the function will fail and do nothing.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3StrCopyUA( unicode * dest, const ascii * src, u64 max_chars );
  /*
    Converts the src ascii string to dest unicode using the trivial mapping,
    up to and including the terminating \0. Only strings up to
    ( max_chars - 1 ) characters long will be copied.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3StrCmpA( const ascii * stringA, const ascii * stringB, u64 max_chars );
  /*
    Compare the 2 ascii strings, up to max_chars or the end of one string.
    Returns: 0 if the strings are identical, a postive number if the ascii
      code of the first different character in stringA is greater than the
      ascii code of that character in stringB, or a negative number if the
      ascii code of the first different character in stringA is less than the
      ascii code of that character in stringB.
  */

s32 v3StrCmpU( const unicode * stringA, const unicode * stringB,
  u64 max_chars );
  /*
    Compare the 2 unicode strings, up to max_chars or the end of one string.
    Returns: 0 if the strings are identical, a postive number if the unicode
      code of the first different character in stringA is greater than the
      unicode code of that character in stringB, or a negative number if the
      unicode code of the first different character in stringA is less than
      the unicode code of that character in stringB.
  */

ascii * v3StrCharA( const ascii * string, ascii match_char, u64 max_chars );
  /*
    Locate match_char in the string, which is up to max_chars long.
    Returns: pointer to the first ascii character matching match_char in the
      string, or NULL if it does not occur in the string.
  */

unicode * v3StrCharU( const unicode * string, unicode match_char,
  u64 max_chars );
  /*
    Locate match_char in the string, which is up to max_chars long.
    Returns: pointer to the first unicode character matching match_char in
      the string, or NULL if it does not occur in the string.
  */

ascii * v3StrStrA( const ascii * string, const ascii * substring,
  u64 max_chars );
  /*
    Locate substring in the string, which is up to max_chars-1 long.
    Returns: pointer to the first match of substring in string,
      or NULL if it does not occur in the string.
  */

unicode * v3StrStrU( const unicode * string, const unicode * substring,
  u64 max_chars );
  /*
    Locate substring in the string, which is up to max_chars-1 long.
    Returns: pointer to the first match of substring in string,
      or NULL if it does not occur in the string.
  */

/*
  The v3{type}(A|U) family of functions. Macro's for speed.
*/
#define v3u32A( result, end, string, radix ) \
  v3_ParseTextA( (void *) result, 4, V3_IO_UNSIGNED, end, string, radix )
#define v3s32A( result, end, string, radix ) \
  v3_ParseTextA( (void *) result, 4, V3_IO_SIGNED, end, string, radix )
#define v3u64A( result, end, string, radix ) \
  v3_ParseTextA( (void *) result, 8, V3_IO_UNSIGNED, end, string, radix )
#define v3s64A( result, end, string, radix ) \
  v3_ParseTextA( (void *) result, 8, V3_IO_SIGNED, end, string, radix )
#define v3u128A( result, end, string, radix ) \
  v3_ParseTextA( (void *) result, 16, V3_IO_UNSIGNED, end, string, radix )
#define v3s128A( result, end, string, radix ) \
  v3_ParseTextA( (void *) result, 16, V3_IO_SIGNED, end, string, radix )
#define v3u32U( result, end, string, radix ) \
  v3_ParseTextU( (void *) result, 4, V3_IO_UNSIGNED, end, string, radix )
#define v3s32U( result, end, string, radix ) \
  v3_ParseTextU( (void *) result, 4, V3_IO_SIGNED, end, string, radix )
#define v3u64U( result, end, string, radix ) \
  v3_ParseTextU( (void *) result, 8, V3_IO_UNSIGNED, end, string, radix )
#define v3s64U( result, end, string, radix ) \
  v3_ParseTextU( (void *) result, 8, V3_IO_SIGNED, end, string, radix )
#define v3u128U( result, end, string, radix ) \
  v3_ParseTextU( (void *) result, 16, V3_IO_UNSIGNED, end, string, radix )
#define v3s128U( result, end, string, radix ) \
  v3_ParseTextU( (void *) result, 16, V3_IO_SIGNED, end, string, radix )
  /*
    Convert the ascii or unicode string written in radix to the number type.
    ascii numbers are of the form: [space][+|-][0|0x|0X][0-9a-zA-Z]*
    unicode numbers accept numbers/letters in any language form.
    radix must be 2-36 or 0. If radix is 0, numbers starting with 0x|0X will
    be read as base 16, numbers starting with 0 will be interpreted as base 8,
    and all others will be base 10. If end is not NULL, it will be set to the
    character after the last character used in the number.
    For u8 and u16 types use the u32 function and typecast the result.
    Note that use of radixes other then 2, 8, 10, or 16 are generally useless.
    Returns: sets result to number and returns V3_PASS on success, or sets
      result to 0 and returns V3_FAIL on failure.
  */

s32 v3f32A( f32 * result, ascii ** end, const ascii * string );
s32 v3f64A( f64 * result, ascii ** end, const ascii * string );
s32 v3f32U( f32 * result, unicode ** end, const unicode * string );
s32 v3f64U( f64 * result, unicode ** end, const unicode * string );
  /*
    Convert the ascii or unicode base-10 string to a floating point number.
    ascii numbers are of the form: [space][+|-][0-9]*.[0-9]*[e|E[+|-][0-9]*]
    unicode numbers accept the base-10 numbers of any language.
    If end is not NULL, it will be set to the character after the last
    character used in the number.
    Returns: sets result to the number and returns V3_PASS on success, or
      sets result to one of ERANGE or +/- HUGE_VAL and returns V3_FAIL
      on failure.
  */

/*
  Output functions. printf
*/

u64 v3PrintA( const ascii * format, ... );
  /*
    Prints the formatted ascii text to the standard output device.
    All output buffers will be flushed before return.
    %[width][.prec]type_char
    [width]
      0n = min n characters are printed, pad 0
      -n = min n characters are printed, left justify
       n = min n characters are printed
      *  = next argument (u32) will specify width
    [.prec]
      .n = max n decimals places are printed for type f|g|F|G
           Should always be set, default = 6.
      .n = max n characters printed for type s|b. Default = 0.
      .n = no effect on all other types
      .* = next argument (u32) will specify precision
    [type_char]
      c = single character
      s = zero terminated string (ascii in A versions, unicode in U versions)
      b = raw byte buffer, precision flag is length in bytes
      f = f32 [-]dddd.dddddd
      g = f64 [-]dddd.dddddd
      F = f32 [-]d.dddddd[Edd]
      G = f64 [-]d.dddddd[Edd]
      u = u32 dec
      v = u64 dec
      w = u128 dec
      i = s32 dec
      j = s64 dec
      k = s128 dec
      X = u32 hex (hex = [0-9A-F], always uppercase, no automatic "0x" prefix)
      Y = u64 hex
      Z = u128 hex
      p = pointer zero-padded hex - native length 32bit|64bit
      Note the mnemonics for the 32bit (root) variants:
          c = character, s = string, f = float, b = buffer
          u = unsigned, i = integer, X = hex, p = pointer
    Returns: Number of characters output.
  */

u64 v3PrintU( const unicode * format, ... );
  /*
    Prints the formatted unicode text to the standard output device.
    All output buffers will be flushed before return.
    See v3PrintA for format usage.
    Returns: Number of characters output.
  */

u64 v3PrintAStr( ascii * string, u64 max_chars, const ascii * format, ... );
  /*
    Prints the formatted ascii text to the string. No more the max_chars-1
    characters will be written to the string. See v3PrintA for format usage.
    Returns: Number of characters written to string, or -1 if truncated due
      to max_chars.
  */

u64 v3PrintUStr( unicode * string, u64 max_chars,
                 const unicode * format, ... );
  /*
    Prints the formatted unicode text to the string. No more the max_chars-1
    characters will be written to the string. See v3PrintA for format usage.
    Returns: Number of characters written to string, or -1 if truncated due
      to max_chars.
  */

u64 v3PrintAFile( v3_FILE * file, const ascii * format, ... );
  /*
    Prints the formatted ascii text to the file.
    See v3PrintA for format usage.
    Returns: Number of characters written to file.
  */

u64 v3PrintUFile( v3_FILE * file, const unicode * format, ... );
  /*
    Prints the formatted unicode text to the file.
    See v3PrintA for format usage.
    Returns: Number of characters written to file.
  */

/* low level */

s32 v3_ParseTextA( void * number, u8 number_width, u32 sign, ascii ** end,
  const ascii * string, u32 radix );
  /*
    Worker function for v3{type}A functions, places the parsed value into
    number, when number_width is the bytes in the expected value (4, 8, or 16)
    If sign is V3_IO_SIGNED, then a signed value will be parsed and produced,
    if sign is V3_IO_UNSIGNED, then an unsigned is parsed and producted.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3_ParseTextU( void * number, u8 number_width, u32 sign, unicode ** end,
  const unicode * string, u32 radix );
  /*
    Worker function for v3{type}U functions, places the parsed value into
    number, when number_width is the bytes in the expected value (4, 8, or 16)
    If sign is V3_IO_SIGNED, then a signed value will be parsed and produced,
    if sign is V3_IO_UNSIGNED, then an unsigned is parsed and producted.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s8 v3_GetCharValU( unicode ch );
  /*
    Helper function for v3_ParseTextU. Maps characters into a value similar
    to what the table does in the v3_ParseTextA function.
  */

u64 v3_Print( v3_FILE * file, void * string, u8 char_width, u64 max_chars,
  const void * format, va_list args );
  /*
    Output the formatted character string to the file or string.
    If file is not NULL, then we write to the file.
    If string_ptr is not NULL, then we output to the string.
    If both are NULL, we write to the standard output.
    Returns: Number of characters written to file or string.
  */

s32 v3_PrintChar( v3_FILE * file, void * string_ptr, u8 char_width,
  u16 character );
  /*
    Output the character to the file or string.
    If file is not NULL, then we write to the file.
    If string_ptr is not NULL, then we output to the string.
    If both are NULL, we write to the standard output.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3_isnan( f64 number );
  /*
    Test if number is NAN.
    Returns: 1 if number is NAN.
  */

s32 v3_isinf( f64 number );
  /*
    Test if number is +INF or -INF.
    Returns: 1 if number is +INF or -INF.
  */

/* testing */

s32 v3_TestCosmIO( void );
  /*
    Test functions in this header.
    Returns: V3_PASS on success, or a negative number corresponding to the
      test that failed.
  */

#endif
