/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

/* CPU/OS Layer - CPU and OS specific code is allowed */

#include "cosmmath.h"

/* bigger */

#if ( defined( CPU_32BIT ) )

u64 v3u64u32( u32 a )
{
  u64 r;

  r.hi = 0;
  r.lo = a;
  return( r );
}

s64 v3s64s32( s32 a )
{
  s64 r;

  if ( a < 0 )
  {
    r.hi = 0xFFFFFFFF;
  }
  else
  {
    r.hi = 0;
  }
  r.lo = a;
  return( r );
}
#endif /* CPU_32BIT */

u128 v3u128u64( u64 a )
{
#if ( defined( CPU_64BIT ) )
  u128 r;

  r.hi = 0;
  r.lo = a;
  return( r );
#else
  u128 r;

  r.hi.hi = 0;
  r.hi.lo = 0;
  r.lo.hi = a.hi;
  r.lo.lo = a.lo;
  return( r );
#endif
}

s128 v3s128s64( s64 a )
{
#if ( defined( CPU_64BIT ) )
  s128 r;

  if ( a < 0 )
  {
    r.hi = 0xFFFFFFFFFFFFFFFF;
    r.lo = a;
  }
  else
  {
    r.hi = 0;
    r.lo = a;
  }
  return( r );
#else
  s128 r;

  if ( a.hi < 0 )
  {
    r.hi.hi = r.hi.lo = 0xFFFFFFFF;
  }
  else
  {
    r.hi.hi = r.hi.lo = 0;
  }
  r.lo.hi = a.hi;
  r.lo.lo = a.lo;
  return( r );
#endif
}

u128 v3u128u32( u32 a )
{
#if ( defined( CPU_64BIT ) )
  u128 r;

  r.hi = 0;
  r.lo = (u64) a;
  return( r );
#else
  u128 r;

  r.hi.hi = 0;
  r.hi.lo = 0;
  r.lo.hi = 0;
  r.lo.lo = a;
  return( r );
#endif
}

s128 v3s128s32( s32 a )
{
#if ( defined( CPU_64BIT ) )
  s128 r;

  if ( a < 0 )
  {
    r.hi = 0xFFFFFFFFFFFFFFFF;
    r.lo = a;
  }
  else
  {
    r.hi = 0;
    r.lo = a;
  }
  return( r );
#else
  s128 r;

  if ( a < 0 )
  {
    r.hi.hi = r.hi.lo = r.lo.hi = 0xFFFFFFFF;
  }
  else
  {
    r.hi.hi = r.hi.lo = r.lo.hi = 0;
  }
  r.lo.lo = a;
  return( r );
#endif
}

/* smaller - truncation will occur */

#if ( defined( CPU_32BIT ) )

u64 v3u64u128( u128 a )
{
  return( a.lo );
}

s64 v3s64s128( s128 a )
{
  s64 r;

  r.hi = (s32) a.lo.hi;
  r.lo = (u32) a.lo.lo;
  return( r );
}

s32 v3s32s128( s128 a )
{
  return( (s32) a.lo.lo );
}

#endif /* CPU_32BIT */

/*
  signed <-> unsigned functions, sign may change
*/

#if ( defined( CPU_32BIT ) )

u64 v3u64s64( s64 a )
{
  u64 r;

  r.hi = (u32) a.hi;
  r.lo = (u32) a.lo;
  return( r );
}

s64 v3s64u64( u64 a )
{
  s64 r;

  r.hi = (s32) a.hi;
  r.lo = (u32) a.lo;
  return( r );
}

#endif

u128 v3u128s128( s128 a )
{
  u128 r;

#if ( defined( CPU_64BIT ) )
  r.hi = (u64) a.hi;
  r.lo = (u64) a.lo;
#else
  r.hi.hi = (u32) a.hi.hi;
  r.hi.lo = (u32) a.hi.lo;
  r.lo.hi = (u32) a.lo.hi;
  r.lo.lo = (u32) a.lo.lo;
#endif

  return( r );
}

s128 v3s128u128( u128 a )
{
  s128 r;

#if ( defined( CPU_64BIT ) )
  r.hi = (s64) a.hi;
  r.lo = (u64) a.lo;
#else
  r.hi.hi = (s32) a.hi.hi;
  r.hi.lo = (u32) a.hi.lo;
  r.lo.hi = (u32) a.lo.hi;
  r.lo.lo = (u32) a.lo.lo;
#endif

  return( r );
}

#if ( defined( CPU_32BIT ) )

/*
  u64 Functions
*/

u64 v3u64Add( u64 a, u64 b )
{
  u64 r;

  r.lo = a.lo + b.lo;
  r.hi = a.hi + b.hi;

  if ( r.lo < a.lo ) /* overflow in the lower order */
  {
    r.hi++;
  }
  return( r );
}

u64 v3u64Sub( u64 a, u64 b )
{
  /* c = a + ~b + 1 */
  return( v3u64Add( a, v3u64Add( v3u64Not( b ), v3u64u32( 1 ) ) ) );
}

u64 v3u64Mul( u64 a, u64 b )
{
  u64 r, tmp;
  u32 a0, a1, a2, a3, b0, b1, b2, b3;

  a0 = a.lo & 0xFFFF;
  a1 = ( a.lo >> 16 ) & 0xFFFF;
  a2 = a.hi & 0xFFFF;
  a3 = ( a.hi >> 16 ) & 0xFFFF;
  b0 = b.lo & 0xFFFF;
  b1 = ( b.lo >> 16 ) & 0xFFFF;
  b2 = b.hi & 0xFFFF;
  b3 = ( b.hi >> 16 ) & 0xFFFF;

  /* calculate subresults */
  /* efficiency = 10 mul32, 3 shift64, 9 add64, 10 promotes */

  r = v3u64u32( a0 * b0 );

  tmp = v3u64u32( a0 * b1 );
  tmp = v3u64Add( tmp, v3u64u32( a1 * b0 ) );
  r = v3u64Add( r, v3u64Lsh( tmp, (u32) 16 ) );

  tmp = v3u64u32( a0 * b2 );
  tmp = v3u64Add( tmp, v3u64u32( a1 * b1 ) );
  tmp = v3u64Add( tmp, v3u64u32( a2 * b0 ) );
  r = v3u64Add( r, v3u64Lsh( tmp, (u32) 32 ) );

  tmp = v3u64u32( a0 * b3 );
  tmp = v3u64Add( tmp, v3u64u32( a1 * b2 ) );
  tmp = v3u64Add( tmp, v3u64u32( a2 * b1 ) );
  tmp = v3u64Add( tmp, v3u64u32( a3 * b0 ) );
  r = v3u64Add( r, v3u64Lsh( tmp, (u32) 48 ) );

  return( r );
}

u64 v3u64Div( u64 a, u64 b )
{
  u64 div;
  u64 mod;

  v3_u64DivMod( a, b, &div, &mod );

  return( div );
}

u64 v3u64Mod( u64 a, u64 b )
{
  u64 div;
  u64 mod;

  v3_u64DivMod( a, b, &div, &mod );

  return( mod );
}

u64 v3u64Inc( u64 * a )
{
  u64 r;

  r = *a;
  *a = v3u64Add( r, v3u64u32( 1 ) );

  return( r );
}

u64 v3u64Dec( u64 * a )
{
  u64 r;

  r = *a;
  *a = v3u64Sub( r, v3u64u32( 1 ) );

  return( r );
}


u64 v3u64And( u64 a, u64 b )
{
  u64 r;

  r.hi = a.hi & b.hi;
  r.lo = a.lo & b.lo;
  return( r );
}

u64 v3u64Or ( u64 a, u64 b )
{
  u64 r;

  r.hi = a.hi | b.hi;
  r.lo = a.lo | b.lo;
  return( r );
}

u64 v3u64Xor( u64 a, u64 b )
{
  u64 r;

  r.hi = a.hi ^ b.hi;
  r.lo = a.lo ^ b.lo;
  return( r );
}

u64 v3u64Not( u64 a )
{
  u64 r;

  r.hi = ~a.hi;
  r.lo = ~a.lo;
  return( r );
}

u64 v3u64Lsh( u64 a, u32 x )
{
  u64 r;

  /* Take care of shifts by offset > 31 */
  if ( x <= 32 )
  {
    if ( x == 32 )
    {
      r.lo = 0;
      r.hi = a.lo;
    }
    else
    {
      r.lo = a.lo << x;
      r.hi = ( a.hi << x ) | ( a.lo >> (32 - x) );
    }
  }
  else
  {
    r.lo = 0;
    r.hi = a.lo << (x - 32);
  }

  return( r );
}

u64 v3u64Rsh( u64 a, u32 x )
{
  u64 r;

  /* Take care of shifts by offset > 31 */
  if ( x <= 32 )
  {
    if ( x == 32 )
    {
      r.lo = a.hi;
      r.hi = 0;
    }
    else
    {
      r.lo = ( a.hi << (32 - x) ) | ( a.lo >> x );
      r.hi = a.hi >> x;
    }
  }
  else
  {
    r.lo = a.hi >> (x - 32);
    r.hi = 0;
  }

  return( r );
}

u32 v3u64Eq( u64 a, u64 b )
{
  return( ( a.hi == b.hi ) && ( a.lo == b.lo ) );
}

u32 v3u64Gt( u64 a, u64 b )
{
  if ( ( a.hi > b.hi ) || ( ( a.hi == b.hi ) && ( a.lo > b.lo ) ) )
  {
    return( 1 );
  }
  return( 0 );
}

u32 v3u64Lt( u64 a, u64 b )
{
  if ( ( a.hi < b.hi ) || ( ( a.hi == b.hi ) && ( a.lo < b.lo ) ) )
  {
    return( 1 );
  }
  return( 0 );
}

/*
  s64 Functions
*/

s64 v3s64Add( s64 a, s64 b )
{
  u64 uA, uB, uResult;
  s64 result;

  uA.hi = (u32) a.hi;
  uA.lo = (u32) a.lo;
  uB.hi = (u32) b.hi;
  uB.lo = (u32) b.lo;

  uResult = v3u64Add( uA, uB );
  result.hi = (s32) uResult.hi;
  result.lo = (u32) uResult.lo;
  return( result );
}

s64 v3s64Sub( s64 a, s64 b )
{
  u64 uA, uB, uResult;
  s64 result;

  uA.hi = (u32) a.hi;
  uA.lo = (u32) a.lo;
  uB.hi = (u32) b.hi;
  uB.lo = (u32) b.lo;

  uResult = v3u64Sub( uA, uB );
  result.hi = (s32) uResult.hi;
  result.lo = (u32) uResult.lo;
  return( result );
}

s64 v3s64Mul( s64 a, s64 b )
{
  u64 uA, uB, uResult;
  s64 result;
  s32 sign;       /* 0 = no, else yes */

  uA.hi = (u32) a.hi;
  uA.lo = (u32) a.lo;
  uB.hi = (u32) b.hi;
  uB.lo = (u32) b.lo;

  sign = 0;
  /* Get the absolute value of a and b */
  if ( a.hi < 0 )
  {
    sign++;
    uA = v3u64Add( v3u64Not( uA ), v3u64u32( 1 ) );
  }
  if ( b.hi < 0 )
  {
    sign--;
    uB = v3u64Add( v3u64Not( uB ), v3u64u32( 1 ) );
  }

  uResult = v3u64Mul( uA, uB );

  if ( sign != 0 )
  {
    /* The result must be negative */
    uResult = v3u64Add( v3u64Not( uResult ), v3u64u32( 1 ) );
  }
  result.hi = (s32) uResult.hi;
  result.lo = (u32) uResult.lo;

  return( result );
}

s64 v3s64Div( s64 a, s64 b )
{
  u8 sign;
  u64 uA, uB, uR;
  s64 result;

  sign = 0;
  uA.hi = (u32) a.hi;
  uA.lo = (u32) a.lo;
  uB.hi = (u32) b.hi;
  uB.lo = (u32) b.lo;

  /* remove the sign */
  /* result is negative if and only if one of a or b is negative */
  if ( a.hi < 0 )
  {
    sign++;
    /* uA = -a */
    uA = v3u64Add( v3u64Not( uA ), v3u64u32( 1 ) );
  }
  if ( b.hi < 0 )
  {
    sign--;
    /* uB = -b */
    uB = v3u64Add( v3u64Not( uB ), v3u64u32( 1 ) );
  }
  uR = v3u64Div( uA, uB );

  /* Modify the sign if we need to */
  if ( sign != 0 )
  {
    /* result = -uR */
    uR = v3u64Add( v3u64Not( uR ), v3u64u32( 1 ) );
  }
  result.hi = (s32) uR.hi;
  result.lo = (u32) uR.lo;

  return( result );
}

#endif /* CPU_32BIT - v3s64Mod needed by both */

s64 v3s64Mod( s64 a, s64 b )
{
  u8 sign;
  s64 result;

#if ( defined( CPU_64BIT ) )
  sign = 0;
  if ( a < 0 )
  {
    sign++;
    a = -a;
  }
  if ( b < 0 )
  {
    sign--;
    b = -b;
  }

  result = v3u64Mod( a, b );

  /* Modify the sign if we need to */
  if ( ( sign != 0 ) && ( result != 0 ) )
  {
    /* result = -result + b */
    result = -result + b;
  }
#else
  u64 uA, uB, uR;

  uA.hi = (u32) a.hi;
  uA.lo = (u32) a.lo;
  uB.hi = (u32) b.hi;
  uB.lo = (u32) b.lo;

  /* remove the sign */
  /* result is negative if and only if a is negative */
  sign = 0;
  if ( a.hi < 0 )
  {
    sign++;
    /* uA = -a */
    uA = v3u64Add( v3u64Not( uA ), v3u64u32( 1 ) );
  }
  if ( b.hi < 0 )
  {
    sign--;
    /* uB = -b */
    uB = v3u64Add( v3u64Not( uB ), v3u64u32( 1 ) );
  }

  uR = v3u64Mod( uA, uB );

  /* Modify the sign if we need to */
  if ( ( sign != 0 ) && ( !v3u64Eq( uR, v3u64u32( 0 ) ) ) )
  {
    /* result = -result + b */
    uR = v3u64Add( v3u64Not( uR ), v3u64u32( 1 ) );
    uR = v3u64Add( uR, uB );
  }
  result.hi = (s32) uR.hi;
  result.lo = (u32) uR.lo;
#endif

  return( result );
}

#if ( defined( CPU_32BIT ) )

s64 v3s64Inc( s64 * a )
{
  s64 r;

  r = *a;
  *a = v3s64Add( r, v3s64s32( 1 ) );

  return( r );
}

s64 v3s64Dec( s64 * a )
{
  s64 r;

  r = *a;
  *a = v3s64Sub ( r, v3s64s32( 1 ) );

  return( r );
}

u32 v3s64Eq( s64 a, s64 b )
{
  return( ( a.hi == b.hi ) && ( a.lo == b.lo ) );
}

u32 v3s64Gt( s64 a, s64 b )
{
  if ( ( a.hi > b.hi ) || ( ( a.hi == b.hi ) && ( a.lo > b.lo ) ) )
  {
    return( 1 );
  }
  return( 0 );
}

u32 v3s64Lt( s64 a, s64 b )
{
  if ( ( a.hi < b.hi ) || ( ( a.hi == b.hi ) && ( a.lo < b.lo ) ) )
  {
    return( 1 );
  }
  return( 0 );
}
#endif /* CPU_32BIT */

/*
  u128 Functions
*/

u128 v3u128Add( u128 a, u128 b )
{
  u128 r;

#if ( defined( CPU_64BIT ) )
  r.lo = a.lo + b.lo;
  r.hi = a.hi + b.hi;
  if ( r.lo < a.lo ) /* Deal with the overflow in the lower order */
  {
    r.hi++;
  }
#else
  r.lo = v3u64Add( a.lo, b.lo );
  r.hi = v3u64Add( a.hi, b.hi );
  if ( v3u64Gt( a.lo, r.lo ) )
  {
    /* Deal with the overflow in the lower order */
    v3u64Inc( &r.hi );
  }
#endif
  return( r );
}

u128 v3u128Sub( u128 a, u128 b )
{
  /* c = a + ~b + 1 */
#if ( defined( CPU_64BIT ) )
  return( v3u128Add( a, v3u128Add( v3u128Not( b ), ( v3u128u64( 1 ) ) ) ) );
#else
  return( v3u128Add( a, v3u128Add( v3u128Not( b ), ( v3u128u32( 1 ) ) ) ) );
#endif
}

u128 v3u128Mul( u128 a, u128 b )
{
#if ( defined( CPU_64BIT ) )
  u128 r, tmp;
  u64 a0, a1, a2, a3, b0, b1, b2, b3;

  a0 = a.lo & 0xFFFFFFFF;
  a1 = ( a.lo >> 32 ) & 0xFFFFFFFF;
  a2 = a.hi & 0xFFFFFFFF;
  a3 = ( a.hi >> 32 ) & 0xFFFFFFFF;
  b0 = b.lo & 0xFFFFFFFF;
  b1 = ( b.lo >> 32 ) & 0xFFFFFFFF;
  b2 = b.hi & 0xFFFFFFFF;
  b3 = ( b.hi >> 32 ) & 0xFFFFFFFF;

  /* calculate subresults */
  /* efficiency = 10 mul64, 3 shift128, 9 add128, 10 promotes */

  r = v3u128u64( a0 * b0 );

  tmp = v3u128u64( a0 * b1 );
  tmp = v3u128Add( tmp, v3u128u64( a1 * b0 ) );
  r = v3u128Add( r, v3u128Lsh( tmp, (u32) 32 ) );

  tmp = v3u128u64( a0 * b2 );
  tmp = v3u128Add( tmp, v3u128u64( a1 * b1 ) );
  tmp = v3u128Add( tmp, v3u128u64( a2 * b0 ) );
  r = v3u128Add( r, v3u128Lsh( tmp, (u32) 64 ) );

  tmp = v3u128u64( a0 * b3 );
  tmp = v3u128Add( tmp, v3u128u64( a1 * b2 ) );
  tmp = v3u128Add( tmp, v3u128u64( a2 * b1 ) );
  tmp = v3u128Add( tmp, v3u128u64( a3 * b0 ) );
  r = v3u128Add( r, v3u128Lsh( tmp, (u32) 96 ) );


  return( r );
#else
  u128 r;
  u64 tmp;
  u32 a0, a1, a2, a3, a4, a5, a6, a7;
  u32 b0, b1, b2, b3, b4, b5, b6, b7;

  a0 = a.lo.lo & 0xFFFF;
  a1 = ( a.lo.lo >> 16 ) & 0xFFFF;
  a2 = a.lo.hi & 0xFFFF;
  a3 = ( a.lo.hi >> 16 ) & 0xFFFF;
  a4 = a.hi.lo & 0xFFFF;
  a5 = ( a.hi.lo >> 16 ) & 0xFFFF;
  a6 = a.hi.hi & 0xFFFF;
  a7 = ( a.hi.hi >> 16 ) & 0xFFFF;
  b0 = b.lo.lo & 0xFFFF;
  b1 = ( b.lo.lo >> 16 ) & 0xFFFF;
  b2 = b.lo.hi & 0xFFFF;
  b3 = ( b.lo.hi >> 16 ) & 0xFFFF;
  b4 = b.hi.lo & 0xFFFF;
  b5 = ( b.hi.lo >> 16 ) & 0xFFFF;
  b6 = b.hi.hi & 0xFFFF;
  b7 = ( b.hi.hi >> 16 ) & 0xFFFF;

  /* calculate subresults */
  /* efficiency = 36 mul32, 2 shift64,  5 shift128,  7 add128, 28 add64
     36 promote32,  7 promote64 */

  r = v3u128u32( a0 * b0 );

  tmp = v3u64u32( a0 * b1 );
  tmp = v3u64Add( tmp, v3u64u32( a1 * b0 ) );
  r = v3u128Add( r, v3u128u64( v3u64Lsh( tmp, (u32) 16 ) ) );

  tmp = v3u64u32( a0 * b2 );
  tmp = v3u64Add( tmp, v3u64u32( a1 * b1 ) );
  tmp = v3u64Add( tmp, v3u64u32( a2 * b0 ) );
  r = v3u128Add( r, v3u128u64( v3u64Lsh( tmp, (u32) 32 ) ) );

  tmp = v3u64u32( a0 * b3 );
  tmp = v3u64Add( tmp, v3u64u32( a1 * b2 ) );
  tmp = v3u64Add( tmp, v3u64u32( a2 * b1 ) );
  tmp = v3u64Add( tmp, v3u64u32( a3 * b0 ) );
  r = v3u128Add( r, v3u128Lsh( v3u128u64( tmp ), (u32) 48 ) );

  tmp = v3u64u32( a0 * b4 );
  tmp = v3u64Add( tmp, v3u64u32( a1 * b3 ) );
  tmp = v3u64Add( tmp, v3u64u32( a2 * b2 ) );
  tmp = v3u64Add( tmp, v3u64u32( a3 * b1 ) );
  tmp = v3u64Add( tmp, v3u64u32( a4 * b0 ) );
  r = v3u128Add( r, v3u128Lsh( v3u128u64( tmp ), (u32) 64 ) );

  tmp = v3u64u32( a0 * b5 );
  tmp = v3u64Add( tmp, v3u64u32( a1 * b4 ) );
  tmp = v3u64Add( tmp, v3u64u32( a2 * b3 ) );
  tmp = v3u64Add( tmp, v3u64u32( a3 * b2 ) );
  tmp = v3u64Add( tmp, v3u64u32( a4 * b1 ) );
  tmp = v3u64Add( tmp, v3u64u32( a5 * b0 ) );
  r = v3u128Add( r, v3u128Lsh( v3u128u64( tmp ), (u32) 80 ) );

  tmp = v3u64u32( a0 * b6 );
  tmp = v3u64Add( tmp, v3u64u32( a1 * b5 ) );
  tmp = v3u64Add( tmp, v3u64u32( a2 * b4 ) );
  tmp = v3u64Add( tmp, v3u64u32( a3 * b3 ) );
  tmp = v3u64Add( tmp, v3u64u32( a4 * b2 ) );
  tmp = v3u64Add( tmp, v3u64u32( a5 * b1 ) );
  tmp = v3u64Add( tmp, v3u64u32( a6 * b0 ) );
  r = v3u128Add( r, v3u128Lsh( v3u128u64( tmp ), (u32) 96 ) );

  tmp = v3u64u32( a0 * b7 );
  tmp = v3u64Add( tmp, v3u64u32( a1 * b6 ) );
  tmp = v3u64Add( tmp, v3u64u32( a2 * b5 ) );
  tmp = v3u64Add( tmp, v3u64u32( a3 * b4 ) );
  tmp = v3u64Add( tmp, v3u64u32( a4 * b3 ) );
  tmp = v3u64Add( tmp, v3u64u32( a5 * b2 ) );
  tmp = v3u64Add( tmp, v3u64u32( a6 * b1 ) );
  tmp = v3u64Add( tmp, v3u64u32( a7 * b0 ) );
  r = v3u128Add( r, v3u128Lsh( v3u128u64( tmp ), (u32) 112 ) );

  return( r );
#endif
}

u128 v3u128Div( u128 a, u128 b )
{
  u128 div;
  u128 mod;

  v3_u128DivMod( a, b, &div, &mod );

  return( div );
}

u128 v3u128Mod( u128 a, u128 b )
{
  u128 div;
  u128 mod;

  v3_u128DivMod( a, b, &div, &mod );

  return( mod );
}

u128 v3u128Inc( u128 * a )
{
  u128 r;

  r = *a;
#if ( defined( CPU_64BIT ) )
  *a = v3u128Add( r, v3u128u64( 1 ) );
#else
  *a = v3u128Add( r, v3u128u32( 1 ) );
#endif

  return( r );
}

u128 v3u128Dec( u128 * a )
{
  u128 r;

  r = *a;
#if ( defined( CPU_64BIT ) )
  *a = v3u128Sub( r, v3u128u64( 1 ) );
#else
  *a = v3u128Sub( r, v3u128u32( 1 ) );
#endif

  return( r );
}

u128 v3u128And( u128 a, u128 b )
{
  u128 r;

#if ( defined( CPU_64BIT ) )
  r.hi = a.hi & b.hi;
  r.lo = a.lo & b.lo;
#else
  r.hi.hi = a.hi.hi & b.hi.hi;
  r.hi.lo = a.hi.lo & b.hi.lo;
  r.lo.hi = a.lo.hi & b.lo.hi;
  r.lo.lo = a.lo.lo & b.lo.lo;
#endif

  return( r );
}

u128 v3u128Or ( u128 a, u128 b )
{
  u128 r;

#if ( defined( CPU_64BIT ) )
  r.hi = a.hi | b.hi;
  r.lo = a.lo | b.lo;
#else
  r.hi.hi = a.hi.hi | b.hi.hi;
  r.hi.lo = a.hi.lo | b.hi.lo;
  r.lo.hi = a.lo.hi | b.lo.hi;
  r.lo.lo = a.lo.lo | b.lo.lo;
#endif

  return( r );
}

u128 v3u128Xor( u128 a, u128 b )
{
  u128 r;

#if ( defined( CPU_64BIT ) )
  r.hi = a.hi ^ b.hi;
  r.lo = a.lo ^ b.lo;
#else
  r.hi.hi = a.hi.hi ^ b.hi.hi;
  r.hi.lo = a.hi.lo ^ b.hi.lo;
  r.lo.hi = a.lo.hi ^ b.lo.hi;
  r.lo.lo = a.lo.lo ^ b.lo.lo;
#endif

  return( r );
}

u128 v3u128Not( u128 a )
{
  u128 r;

#if ( defined( CPU_64BIT ) )
  r.hi = ~a.hi;
  r.lo = ~a.lo;
#else
  r.hi.hi = ~a.hi.hi;
  r.hi.lo = ~a.hi.lo;
  r.lo.hi = ~a.lo.hi;
  r.lo.lo = ~a.lo.lo;
#endif

  return( r );
}

u128 v3u128Lsh( u128 a, u32 x )
{
  u128 r;

  if ( x > 127 )
    return( v3u128u32( 0 ) );

#if ( defined( CPU_64BIT ) )
  /* Take care of shifts by offset > 63 */
  if ( x <= 64 )
  {
    if ( x == 64 )
    {
      r.lo = 0;
      r.hi = a.lo;
    }
    else
    {
      r.hi = ( a.hi << x ) | ( a.lo >> (64 - x) );
      r.lo = a.lo << x;
    }
  }
  else
  {
    r.hi = a.lo << (x - 64);
    r.lo = 0;
  }
#else
  /* Take care of shifts by offset > 31 */
  if ( x >= 96 )
  {
    if ( x == 96 )
    {
      r.lo.lo = 0;
      r.lo.hi = 0;
      r.hi.lo = 0;
      r.hi.hi = a.lo.lo;
    }
    else
    {
      r.lo.lo = 0;
      r.lo.hi = 0;
      r.hi.lo = 0;
      r.hi.hi = a.lo.lo << (x - 96);
    }
  }
  else if ( x >= 64 )
  {
    if ( x == 64 )
    {
      r.lo.lo = 0;
      r.lo.hi = 0;
      r.hi.lo = a.lo.lo;
      r.hi.hi = a.lo.hi;
    }
    else
    {
      r.lo.lo = 0;
      r.lo.hi = 0;
      r.hi.lo = a.lo.lo << (x - 64);
      r.hi.hi = ( ( a.lo.hi << (x - 64) ) | ( a.lo.lo >> (96 - x) ) );
    }
  }
  else if ( x >= 32 )
  {
    if ( x == 32 )
    {
      r.lo.lo = 0;
      r.lo.hi = a.lo.lo;
      r.hi.lo = a.lo.hi;
      r.hi.hi = a.hi.lo;
    }
    else
    {
      r.lo.lo = 0;
      r.lo.hi = a.lo.lo << (x - 32);
      r.hi.lo = ( a.lo.hi << (x - 32) ) | ( a.lo.lo >> (64 - x) );
      r.hi.hi = ( a.hi.lo << (x - 32) ) | ( a.lo.hi >> (64 - x) );
    }
  }
  else
  {
    if ( x == 0 )
    {
      r = a;
    }
    else
    {
      r.lo.lo = a.lo.lo << x;
      r.lo.hi = ( a.lo.hi << x ) | ( a.lo.lo >> (32 - x) );
      r.hi.lo = ( a.hi.lo << x ) | ( a.lo.hi >> (32 - x) );
      r.hi.hi = ( a.hi.hi << x ) | ( a.hi.lo >> (32 - x) );
    }
  }
#endif
  return( r );
}

u128 v3u128Rsh( u128 a, u32 x )
{
  u128 r;

  if ( x > 127 )
  {
    return( v3u128u32( 0 ) );
  }

#if ( defined( CPU_64BIT ) )
  /* Take care of shifts by offset > 63 */
  if ( x <= 64 )
  {
    if ( x == 64 )
    {
      r.lo = a.hi;
      r.hi = 0;
    }
    else
    {
      r.lo = ( a.hi << (64 - x) ) | (a.lo >> x);
      r.hi = a.hi >> x;
    }
  }
  else
  {
    r.lo = a.hi >> (x - 64);
    r.hi = 0;
  }
#else
  /* Take care of shifts by offset > 31 */
  if ( x <= 32 )
  {
    if ( x == 32 )
    {
      r.lo.lo = a.lo.hi;
      r.lo.hi = a.hi.lo;
      r.hi.lo = a.hi.hi;
      r.hi.hi = 0;
    }
    else
    {
      r.lo.lo = ( a.lo.hi << (32 - x) ) | ( a.lo.lo >> x );
      r.lo.hi = ( a.hi.lo << (32 - x) ) | ( a.lo.hi >> x );
      r.hi.lo = ( a.hi.hi << (32 - x) ) | ( a.hi.lo >> x );
      r.hi.hi = ( a.hi.hi >> x );
    }
  }
  else if ( x <= 64 )
  {
    if ( x == 64 )
    {
      r.lo = a.hi;
      r.hi.lo = 0;
      r.hi.hi = 0;
    }
    else
    {
      r.lo.lo = ( a.hi.lo << (64 - x) ) | ( a.lo.hi >> (x - 32) );
      r.lo.hi = ( a.hi.hi << (64 - x) ) | ( a.hi.lo >> (x - 32) );
      r.hi.lo = ( a.hi.hi >> (x - 32) );
      r.hi.hi = 0;
    }
  }
  else if ( x <= 96 )
  {
    if ( x == 96 )
    {
      r.lo.lo = a.hi.hi;
      r.lo.hi = 0;
      r.hi.lo = 0;
      r.hi.hi = 0;
    }
    else
    {
      r.lo.lo = ( a.hi.hi << (96 - x) ) | ( a.hi.lo >> (x - 64) );
      r.lo.hi = ( a.hi.hi >> (x - 64) );
      r.hi.lo = 0;
      r.hi.hi = 0;
    }
  }
  else
  {
    r.lo.lo = ( a.hi.hi >> (x - 96) ) ;
    r.lo.hi = 0;
    r.hi.lo = 0;
    r.hi.hi = 0;
  }
#endif

  return( r );
}

u32 v3u128Eq( u128 a, u128 b )
{
#if ( defined( CPU_64BIT ) )
  return( ( a.hi == b.hi ) && ( a.lo == b.lo ) );
#else
  return( ( a.hi.hi == b.hi.hi ) && ( a.hi.lo == b.hi.lo )
    && ( a.lo.hi == b.lo.hi ) && ( a.lo.lo == b.lo.lo ) );
#endif
}

u32 v3u128Gt( u128 a, u128 b )
{
#if ( defined( CPU_64BIT ) )
  if ( ( a.hi > b.hi ) || ( ( a.hi == b.hi ) && ( a.lo > b.lo ) ) )
  {
    return( 1 );
  }
  return( 0 );
#else
  if ( v3u64Gt( a.hi, b.hi )
    || ( v3u64Eq( a.hi, b.hi ) && v3u64Gt( a.lo, b.lo ) ) )
  {
    return( 1 );
  }
  return( 0 );
#endif
}

u32 v3u128Lt( u128 a, u128 b )
{
#if ( defined( CPU_64BIT ) )
  if ( ( a.hi < b.hi ) || ( ( a.hi == b.hi ) && ( a.lo < b.lo ) ) )
  {
    return( 1 );
  }
  return( 0 );
#else
  if ( v3u64Lt( a.hi, b.hi )
    || ( v3u64Eq( a.hi, b.hi ) && v3u64Lt( a.lo, b.lo ) ) )
  {
    return( 1 );
  }
  return( 0 );
#endif
}

/*
  s128 Functions
*/

s128 v3s128Add( s128 a, s128 b )
{
#if ( defined( CPU_64BIT ) )
  u128 uA, uB, uResult;
  s128 result;

  uA.hi = (u64) a.hi;
  uA.lo = (u64) a.lo;
  uB.hi = (u64) b.hi;
  uB.lo = (u64) b.lo;

  uResult = v3u128Add( uA, uB );
  result.hi = (s64) uResult.hi;
  result.lo = (u64) uResult.lo;

  return( result );
#else
  u128 uA, uB, uResult;
  s128 result;

  uA.hi.hi = (u32) a.hi.hi;
  uA.hi.lo = (u32) a.hi.lo;
  uA.lo.hi = (u32) a.lo.hi;
  uA.lo.lo = (u32) a.lo.lo;
  uB.hi.hi = (u32) b.hi.hi;
  uB.hi.lo = (u32) b.hi.lo;
  uB.lo.hi = (u32) b.lo.hi;
  uB.lo.lo = (u32) b.lo.lo;

  uResult = v3u128Add( uA, uB );
  result.hi.hi = (s32) uResult.hi.hi;
  result.hi.lo = (u32) uResult.hi.lo;
  result.lo.hi = (u32) uResult.lo.hi;
  result.lo.lo = (u32) uResult.lo.lo;

  return( result );
#endif
}

s128 v3s128Sub( s128 a, s128 b )
{
#if ( defined( CPU_64BIT ) )
  u128 uA, uB, uResult;
  s128 result;

  uA.hi = (u64) a.hi;
  uA.lo = (u64) a.lo;
  uB.hi = (u64) b.hi;
  uB.lo = (u64) b.lo;

  uResult = v3u128Sub( uA, uB );
  result.hi = (s64) uResult.hi;
  result.lo = (u64) uResult.lo;

  return( result );
#else
  u128 uA, uB, uResult;
  s128 result;

  uA.hi.hi = (u32) a.hi.hi;
  uA.hi.lo = (u32) a.hi.lo;
  uA.lo.hi = (u32) a.lo.hi;
  uA.lo.lo = (u32) a.lo.lo;
  uB.hi.hi = (u32) b.hi.hi;
  uB.hi.lo = (u32) b.hi.lo;
  uB.lo.hi = (u32) b.lo.hi;
  uB.lo.lo = (u32) b.lo.lo;

  uResult = v3u128Sub( uA, uB );
  result.hi.hi = (s32) uResult.hi.hi;
  result.hi.lo = (u32) uResult.hi.lo;
  result.lo.hi = (u32) uResult.lo.hi;
  result.lo.lo = (u32) uResult.lo.lo;

  return( result );
#endif
}

s128 v3s128Mul( s128 a, s128 b )
{
  u128 uA, uB, uR;
  s128 result;
  s32 sign;       /* 0 = no, else yes */

#if ( defined( CPU_64BIT ) )
  uA.hi = (u64) a.hi;
  uA.lo = (u64) a.lo;
  uB.hi = (u64) b.hi;
  uB.lo = (u64) b.lo;
#else
  uA.hi.hi = (u32) a.hi.hi;
  uA.hi.lo = (u32) a.hi.lo;
  uA.lo.hi = (u32) a.lo.hi;
  uA.lo.lo = (u32) a.lo.lo;
  uB.hi.hi = (u32) b.hi.hi;
  uB.hi.lo = (u32) b.hi.lo;
  uB.lo.hi = (u32) b.lo.hi;
  uB.lo.lo = (u32) b.lo.lo;
#endif

  sign = 0;
  /* Get the absolute value of a and b */
#if ( defined( CPU_64BIT ) )
  if ( a.hi < 0 )
#else
  if ( a.hi.hi < 0 )
#endif
  {
    sign++;
    uA = v3u128Add( v3u128Not( uA ), v3u128u32( 1 ) );
  }
#if ( defined( CPU_64BIT ) )
  if ( b.hi < 0 )
#else
  if ( b.hi.hi < 0 )
#endif
  {
    sign--;
    uB = v3u128Add( v3u128Not( uB ), v3u128u32( 1 ) );
  }

  uR = v3u128Mul( uA, uB );

  if ( sign != 0 )
  {
    /* The result must be negative */
    uR = v3u128Add( v3u128Not( uR ), v3u128u32( 1 ) );
  }
#if ( defined( CPU_64BIT ) )
  result.hi = (s64) uR.hi;
  result.lo = (u64) uR.lo;
#else
  result.hi.hi = (s32) uR.hi.hi;
  result.hi.lo = (u32) uR.hi.lo;
  result.lo.hi = (u32) uR.lo.hi;
  result.lo.lo = (u32) uR.lo.lo;
#endif

  return( result );
}

s128 v3s128Div( s128 a, s128 b )
{
  u8 sign;
  u128 uA, uB, uR;
  s128 result;

  sign = 0;
#if ( defined( CPU_64BIT ) )
  uA.hi = (u64) a.hi;
  uA.lo = (u64) a.lo;
  uB.hi = (u64) b.hi;
  uB.lo = (u64) b.lo;
#else
  uA.hi.hi = (u32) a.hi.hi;
  uA.hi.lo = (u32) a.hi.lo;
  uA.lo.hi = (u32) a.lo.hi;
  uA.lo.lo = (u32) a.lo.lo;
  uB.hi.hi = (u32) b.hi.hi;
  uB.hi.lo = (u32) b.hi.lo;
  uB.lo.hi = (u32) b.lo.hi;
  uB.lo.lo = (u32) b.lo.lo;
#endif

  /* remove the sign */
  /* result is negative if and only if one of a or b is negative */
#if ( defined( CPU_64BIT ) )
  if ( a.hi < 0 )
#else
  if ( a.hi.hi < 0 )
#endif
  {
    sign++;
    /* uA = -a */
    uA = v3u128Add( v3u128Not( uA ), v3u128u32( 1 ) );
  }
#if ( defined( CPU_64BIT ) )
  if ( b.hi < 0 )
#else
  if ( b.hi.hi < 0 )
#endif
  {
    sign--;
    /* uB = -b */
    uB = v3u128Add( v3u128Not( uB ), v3u128u32( 1 ) );
  }

  uR = v3u128Div( uA, uB );

  /* Modify the sign if we need to */
  if ( sign != 0 )
  {
    /* result = -uR */
    uR = v3u128Add( v3u128Not( uR ), v3u128u32( 1 ) );
  }

#if ( defined( CPU_64BIT ) )
  result.hi = (s64) uR.hi;
  result.lo = (u64) uR.lo;
#else
  result.hi.hi = (s32) uR.hi.hi;
  result.hi.lo = (u32) uR.hi.lo;
  result.lo.hi = (u32) uR.lo.hi;
  result.lo.lo = (u32) uR.lo.lo;
#endif

  return( result );
}

s128 v3s128Mod( s128 a, s128 b )
{
  u8 sign;
  u128 uA, uB, uR;
  s128 result;
  u128 zero;

  sign = 0;
#if ( defined( CPU_64BIT ) )
  uA.hi = (u64) a.hi;
  uA.lo = (u64) a.lo;
  uB.hi = (u64) b.hi;
  uB.lo = (u64) b.lo;
#else
  uA.hi.hi = (u32) a.hi.hi;
  uA.hi.lo = (u32) a.hi.lo;
  uA.lo.hi = (u32) a.lo.hi;
  uA.lo.lo = (u32) a.lo.lo;
  uB.hi.hi = (u32) b.hi.hi;
  uB.hi.lo = (u32) b.hi.lo;
  uB.lo.hi = (u32) b.lo.hi;
  uB.lo.lo = (u32) b.lo.lo;
#endif

  /* remove the sign */
  /* result is negative if and only if a is negative */
#if ( defined( CPU_64BIT ) )
  if ( a.hi < 0 )
#else
  if ( a.hi.hi < 0 )
#endif
  {
    sign++;
    /* uA = -a */
    uA = v3u128Add( v3u128Not( uA ), v3u128u32( 1 ) );
  }
#if ( defined( CPU_64BIT ) )
  if ( b.hi < 0 )
#else
  if ( b.hi.hi < 0 )
#endif
  {
    sign--;
    /* uB = -b */
    uB = v3u128Add( v3u128Not( uB ), v3u128u32( 1 ) );
  }

  uR = v3u128Mod( uA, uB );

  /* Modify the sign if we need to */
  _V3_SET128( zero, 00000000, 00000000, 00000000, 00000000 );
  if ( ( sign != 0 ) && ( !v3u128Eq( uR, zero ) ) )
  {
    /* result = -uR */
    uR = v3u128Add( v3u128Not( uR ), v3u128u32( 1 ) );
    uR = v3u128Add( uR, uB );
  }

#if ( defined( CPU_64BIT ) )
  result.hi = (s64) uR.hi;
  result.lo = (u64) uR.lo;
#else
  result.hi.hi = (s32) uR.hi.hi;
  result.hi.lo = (u32) uR.hi.lo;
  result.lo.hi = (u32) uR.lo.hi;
  result.lo.lo = (u32) uR.lo.lo;
#endif

  return( result );
}

s128 v3s128Inc( s128 * a )
{
  s128 r;

  r = *a;
#if ( defined( CPU_64BIT ) )
  *a = v3s128Add( r, v3s128s64( 1 ) );
#else
  *a = v3s128Add( r, v3s128s32( 1 ) );
#endif

  return( r );
}

s128 v3s128Dec( s128 * a )
{
  s128 r;

  r = *a;
#if ( defined( CPU_64BIT ) )
  *a = v3s128Sub( r, v3s128s64( 1 ) );
#else
  *a = v3s128Sub( r, v3s128s32( 1 ) );
#endif

  return( r );
}

u32 v3s128Eq( s128 a, s128 b )
{
#if ( defined( CPU_64BIT ) )
  return( ( a.hi == b.hi ) && ( a.lo == b.lo ) );
#else
  return( ( a.hi.hi == b.hi.hi ) && ( a.hi.lo == b.hi.lo )
    && ( a.lo.hi == b.lo.hi ) && ( a.lo.lo == b.lo.lo ) );
#endif
}

u32 v3s128Gt( s128 a, s128 b )
{
#if ( defined( CPU_64BIT ) )
  if ( ( a.hi > b.hi ) || ( ( a.hi == b.hi ) && ( a.lo > b.lo ) ) )
  {
    return( 1 );
  }
  return( 0 );
#else
  if ( v3s64Gt( a.hi, b.hi )
    || ( v3s64Eq( a.hi, b.hi ) && v3u64Gt( a.lo, b.lo ) ) )
  {
    return( 1 );
  }
  return( 0 );
#endif
}

u32 v3s128Lt( s128 a, s128 b )
{
#if ( defined( CPU_64BIT ) )
  if ( ( a.hi < b.hi ) || ( ( a.hi == b.hi ) && ( a.lo < b.lo ) ) )
  {
    return( 1 );
  }
  return( 0 );
#else
  if ( v3s64Lt( a.hi, b.hi )
    || ( v3s64Eq( a.hi, b.hi ) && v3u64Lt( a.lo, b.lo ) ) )
  {
    return( 1 );
  }
  return( 0 );
#endif
}

void v3_u64DivMod( u64 a, u64 b, u64 * div, u64 * mod )
{
  u64 result;
  u64 rest;
  u64 mask;
  u64 temp;
  int nb;

  result = v3u64u32( 0 );
  rest = a;
  mask = b;
  temp = v3u64Lsh( v3u64u32( 1 ), 63 );
  nb = 0;

  if ( ( v3u64Eq( b, v3u64u32( 0 ) ) ) || ( v3u64Eq( a, v3u64u32( 0 ) ) ) )
  {
    /* 0/x or x/0 - Return 0 */
    *(div) = v3u64u32( (u32) 0 );
    *(mod) = v3u64u32( (u32) 0 );
    return;
  }

  /* First left shift the 'b' so it have the same number of bit then 'a' */

  /* while ( ( a > mask ) && ( ( 64th bit of mask ) == 0 ) */
  while ( v3u64Gt( a, mask ) && ( v3u64Eq( v3u64And( temp, mask ),
    v3u64u32( (u32) 0 ) ) ) )
  {
    mask = v3u64Lsh( mask, 1 );
    nb = nb + 1;
  }

  /* Now right shift step by step and compare rest and divisor */

  while ( nb > 0 )
  {
    if ( !v3u64Gt( mask, rest ) )
    {
      /* rest >= mask */
      rest = v3u64Sub( rest, mask );
      v3u64Inc( &result );
    }
    result = v3u64Lsh( result, 1 );
    mask = v3u64Rsh( mask, 1 );
    nb = nb - 1;
  }

  if ( !v3u64Gt( mask, rest ) )
  {
    /* rest >= mask */
    rest = v3u64Sub( rest, mask );
    v3u64Inc( &result );
  }

  *(div) = result;
  *(mod) = rest;
}

void v3_u128DivMod( u128 a, u128 b, u128 * div, u128 * mod )
{
  u128 result;
  u128 rest;
  u128 mask;
  u128 temp;
  int nb;

  result = v3u128u32( 0 );
  rest = a;
  mask = b;
  temp = v3u128Lsh( v3u128u32( 1 ), 127 );
  nb = 0;

  if ( ( v3u128Eq( b, v3u128u32( 0 ) ) )
    || ( v3u128Eq( a, v3u128u32( 0 ) ) ) )
  {
    /* 0/x or x/0 - Return 0 */
    *(div) = v3u128u32( 0 );
    *(mod) = v3u128u32( 0 );
    return;
  }

  /* First left shift the 'b' so it have the same number of bit then 'a' */

  /* while ( ( a > mask ) && ( ( 64th bit of mask ) == 0 ) */
  while ( v3u128Gt( a, mask ) && ( v3u128Eq( v3u128And( temp, mask ),
    v3u128u32( 0 ) ) ) )
  {
    mask = v3u128Lsh( mask, 1 );
    nb = nb + 1;
  }

  /* Now right shift step by step and compare rest and divisor */

  while ( nb > 0 )
  {
    if ( !v3u128Gt( mask, rest ) )
    {
      /* rest >= mask */
      rest = v3u128Sub( rest, mask );
      v3u128Inc( &result );
    }
    result = v3u128Lsh( result, 1 );
    mask = v3u128Rsh( mask, 1 );
    nb = nb - 1;
  }

  if ( !v3u128Gt( mask, rest ) )
  {
    /* rest >= mask */
    rest = v3u128Sub( rest, mask );
    v3u128Inc( &result );
  }

  *(div) = result;
  *(mod) = rest;
}

/* testing */

s32 v3_TestCosmMath( void )
{
  u32   u_32a,  u_32b,  u_32r;
  s32   s_32a,  s_32b,  s_32r;
  u64   u_64a,  u_64b,  u_64r;
  s64   s_64a,  s_64b,  s_64r;
  u128 u_128a, u_128b, u_128r;
  s128 s_128a, s_128b, s_128r;

  /*
    Basic methods: Test each function for valid results. Be sure to test
    all border conditions - positive to negative, rollovers, 0, etc.
    a small (+/+), b large (+/-)
    Add: a+b/b+a ordering, a+a no carry, b+b carry
    Sub:

    Base test numbers - here for reference
    u_32a = 0x01234567;
    u_32b = 0xFEDC89AB;
    s_32a = 0x01234567;
    s_32b = 0xFEDC89AB;
    _V3_SET64( u_64a, 01234567, 89ABCDEF );
    _V3_SET64( u_64b, FEDC89AB, 76543210 );
    _V3_SET64( s_64a, 01234567, 89ABCDEF );
    _V3_SET64( s_64b, FEDC89AB, 76543210 );
    _V3_SET128( u_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
    _V3_SET128( u_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );
    _V3_SET128( s_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
    _V3_SET128( s_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );
  */

  /* Conversions bigger and smaller */

  u_32a = 0x01234567;
  u_32b = 0xFEDC89AB;
  s_32a = 0x01234567;
  s_32b = 0xFEDC89AB;
  _V3_SET64( u_64a, 01234567, 89ABCDEF );
  _V3_SET64( u_64b, FEDC89AB, 76543210 );
  _V3_SET64( s_64a, 01234567, 89ABCDEF );
  _V3_SET64( s_64b, FEDC89AB, 76543210 );
  _V3_SET128( u_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
  _V3_SET128( u_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );
  _V3_SET128( s_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
  _V3_SET128( s_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );

  /* u32 -> u64 */
  u_64r = v3u64u32( u_32a );
  if ( !_V3_EQ64( u_64r, 00000000, 01234567 ) )
  {
    return( -1 );
  }
  u_64r = v3u64u32( u_32b );
  if ( !_V3_EQ64( u_64r, 00000000, FEDC89AB ) )
  {
    return( -2 );
  }

  /* s32 ->  s64 */
  s_64r = v3s64s32( s_32a );
  if ( !_V3_EQ64( s_64r, 00000000, 01234567 ) )
  {
    return( -3 );
  }
  s_64r = v3s64s32( s_32b );
  if ( !_V3_EQ64( s_64r, FFFFFFFF, FEDC89AB ) )
  {
    return( -4 );
  }

  /* u64 -> u128 */
  u_128r = v3u128u64( u_64a );
  if ( !_V3_EQ128( u_128r, 00000000, 00000000, 01234567, 89ABCDEF ) )
  {
    return( -5 );
  }
  u_128r = v3u128u64( u_64b );
  if ( !_V3_EQ128( u_128r, 00000000, 00000000, FEDC89AB, 76543210 ) )
  {
    return( -6 );
  }

  /* s64 -> s128 */
  s_128r = v3s128s64( s_64a );
  if ( !_V3_EQ128( s_128r, 00000000, 00000000, 01234567, 89ABCDEF ) )
  {
    return( -7 );
  }
  s_128r = v3s128s64( s_64b );
  if ( !_V3_EQ128( s_128r, FFFFFFFF, FFFFFFFF, FEDC89AB, 76543210 ) )
  {
    return( -8 );
  }

  /* u32 -> u128 */
  u_128r = v3u128u32( u_32a );
  if ( !_V3_EQ128( u_128r, 00000000, 00000000, 00000000, 01234567 ) )
  {
    return( -9 );
  }
  u_128r = v3u128u32( u_32b );
  if ( !_V3_EQ128( u_128r, 00000000, 00000000, 00000000, FEDC89AB ) )
  {
    return( -10 );
  }

  /* s32 -> s128 */
  s_128r = v3s128s32( s_32a );
  if ( !_V3_EQ128( s_128r, 00000000, 00000000, 00000000, 01234567 ) )
  {
    return( -11 );
  }
  s_128r = v3s128s32( s_32b );
  if ( !_V3_EQ128( s_128r, FFFFFFFF, FFFFFFFF, FFFFFFFF, FEDC89AB ) )
  {
    return( -12 );
  }

  /* u64 -> u32 */
  u_32r = v3u32u64( u_64a );
  if ( u_32r != 0x89ABCDEF )
  {
    return( -13 );
  }
  u_32r = v3u32u64( u_64b );
  if ( u_32r != 0x76543210 )
  {
    return( -14 );
  }

  /* s64 ->  s32 */
  s_32r = v3s32s64( s_64a );
  if ( s_32r != 0x89ABCDEF )
  {
    return( -15 );
  }
  s_32r = v3s32s64( s_64b );
  if ( s_32r != 0x76543210 )
  {
    return( -16 );
  }

  /* u128 -> u64 */
  u_64r = v3u64u128( u_128a );
  if ( !_V3_EQ64( u_64r, 45670123, CDEF89AB ) )
  {
    return( -17 );
  }
  u_64r = v3u64u128( u_128b );
  if ( !_V3_EQ64( u_64r, 89ABFEDC, 32107654 ) )
  {
    return( -18 );
  }

  /* s128 -> s64 */
  s_64r = v3s64s128( s_128a );
  if ( !_V3_EQ64( s_64r, 45670123, CDEF89AB ) )
  {
    return( -19 );
  }
  s_64r = v3s64s128( s_128b );
  if ( !_V3_EQ64( s_64r, 89ABFEDC, 32107654 ) )
  {
    return( -20 );
  }

  /* u128 -> u32 */
  u_32r = v3u32u128( u_128a );
  if ( u_32r != 0xCDEF89AB )
  {
    return( -21 );
  }
  u_32r = v3u32u128( u_128b );
  if ( u_32r != 0x32107654 )
  {
    return( -22 );
  }

  /* s128 -> s32 */
  s_32r = v3s32s128( s_128a );
  if ( s_32r != 0xCDEF89AB )
  {
    return( -23 );
  }
  s_32r = v3s32s128( s_128b );
  if ( s_32r != 0x32107654 )
  {
    return( -24 );
  }

  /* u64 functions ====================================================== */

  /* u64 Add */
  _V3_SET64( u_64a, 01234567, 89ABCDEF );
  _V3_SET64( u_64b, FEDC89AB, 76543210 );

  u_64r = v3u64Add( u_64a, u_64b );
  if ( !_V3_EQ64( u_64r, FFFFCF12, FFFFFFFF ) )
  {
    return( -25 );
  }
  u_64r = v3u64Add( u_64b, u_64a );
  if ( !_V3_EQ64( u_64r, FFFFCF12, FFFFFFFF ) )
  {
    return( -26 );
  }
  u_64r = v3u64Add( u_64a, u_64a );
  if ( !_V3_EQ64( u_64r, 02468ACF, 13579BDE ) )
  {
    return( -27 );
  }
  u_64r = v3u64Add( u_64b, u_64b );
  if ( !_V3_EQ64( u_64r, FDB91356, ECA86420 ) )
  {
    return( -28 );
  }

  /* u64 Sub */
  _V3_SET64( u_64a, 01234567, 89ABCDEF );
  _V3_SET64( u_64b, FEDC89AB, 76543210 );

  u_64r = v3u64Sub( u_64a, u_64b );
  if ( !_V3_EQ64( u_64r, 0246BBBC, 13579BDF ) )
  {
    return( -29 );
  }

  u_64r = v3u64Sub( u_64b, u_64a );
  if ( !_V3_EQ64( u_64r, FDB94443, ECA86421 ) )
  {
    return( -30 );
  }

  u_64r = v3u64Sub( u_64a, u_64a );
  if ( !_V3_EQ64( u_64r, 00000000, 00000000 ) )
  {
    return( -31 );
  }

  u_64r = v3u64Sub( u_64b, u_64b );
  if ( !_V3_EQ64( u_64r, 00000000, 00000000 ) )
  {
    return( -32 );
  }

  /* u64 Mul */
  _V3_SET64( u_64a, 01234567, 89ABCDEF );
  _V3_SET64( u_64b, FEDC89AB, 76543210 );

  u_64r = v3u64Mul( u_64a, u_64a );
  if ( !_V3_EQ64( u_64r, DCA5E208, 90F2A521 ) )
  {
    return( -33 );
  }

  u_64r = v3u64Mul( u_64b, u_64b );
  if ( !_V3_EQ64( u_64r, 3441BB37, A44A4100 ) )
  {
    return( -34 );
  }

  u_64r = v3u64Mul( u_64a, u_64b );
  if ( !_V3_EQ64( u_64r, 778C624C, E5618CF0 ) )
  {
    return( -35 );
  }

  u_64r = v3u64Mul( u_64b, u_64a );
  if ( !_V3_EQ64( u_64r, 778C624C, E5618CF0 ) )
  {
    return( -36 );
  }

  /* u64 Div, Mod */
  _V3_SET64( u_64a, 01234567, 89ABCDEF );
  _V3_SET64( u_64b, FEDC89AB, 76543210 );

  u_64r = v3u64Div( u_64a, u_64b );
  if ( !_V3_EQ64( u_64r, 00000000, 00000000 ) )
  {
    return( -37 );
  }

  u_64r = v3u64Mod( u_64a, u_64b );
  if ( !_V3_EQ64( u_64r, 01234567, 89ABCDEF ) )
  {
    return( -38 );
  }

  u_64r = v3u64Div( u_64b, u_64a );
  if ( !_V3_EQ64( u_64r, 00000000, 000000DF ) )
  {
    return( -39 );
  }

  u_64r = v3u64Mod( u_64b, u_64a );
  if ( !_V3_EQ64( u_64r, 0123147A, 89ABCEDF ) )
  {
    return( -40 );
  }

  /* u64 Inc */
  _V3_SET64( u_64r, FFFFFFFF, FFFFFFFE );

  v3u64Inc( &u_64r );
  if ( !_V3_EQ64( u_64r, FFFFFFFF, FFFFFFFF ) )
  {
    return( -41 );
  }
  v3u64Inc( &u_64r );
  if ( !_V3_EQ64( u_64r, 00000000, 00000000 ) )
  {
    return( -42 );
  }

  /* u64 Dec */
  v3u64Dec( &u_64r );
  if ( !_V3_EQ64( u_64r, FFFFFFFF, FFFFFFFF ) )
  {
    return( -43 );
  }
  v3u64Dec( &u_64r );
  if ( !_V3_EQ64( u_64r, FFFFFFFF, FFFFFFFE ) )
  {
    return( -44 );
  }

  /* u64 And */
  _V3_SET64( u_64a, 01234567, 89ABCDEF );
  _V3_SET64( u_64b, FEDC89AB, 76543210 );

  u_64r = v3u64And( u_64a, u_64a );
  if ( !_V3_EQ64( u_64r, 01234567, 89ABCDEF ) )
  {
    return( -45 );
  }
  u_64r = v3u64And( u_64a, u_64b );
  if ( !_V3_EQ64( u_64r, 00000123, 00000000 ) )
  {
    return( -46 );
  }

  /* u64 Or */
  u_64r = v3u64Or( u_64a, u_64a );
  if ( !_V3_EQ64( u_64r, 01234567, 89ABCDEF ) )
  {
    return( -47 );
  }
  u_64r = v3u64Or( u_64a, u_64b );
  if ( !_V3_EQ64( u_64r, FFFFCDEF, FFFFFFFF ) )
  {
    return( -48 );
  }

  /* u64 Xor */
  /* r = a ^ b ^ a ^ b ^ a = a */
  u_64r = v3u64Xor( u_64a, u_64b );
  u_64r = v3u64Xor( u_64r, u_64a );
  u_64r = v3u64Xor( u_64r, u_64b );
  u_64r = v3u64Xor( u_64r, u_64a );
  if ( !_V3_EQ64( u_64r, 01234567, 89ABCDEF ) )
  {
    return( -49 );
  }

  /* u64 Not */
  u_64r = v3u64Not( u_64a );
  if ( !_V3_EQ64( u_64r, FEDCBA98, 76543210 ) )
  {
    return( -50 );
  }
  u_64r = v3u64Not( u_64b );
  if ( !_V3_EQ64( u_64r, 01237654, 89ABCDEF ) )
  {
    return( -51 );
  }

  /* u64 Lsh */
  u_64r = v3u64Lsh( u_64a, 4 );
  if ( !_V3_EQ64( u_64r, 12345678, 9ABCDEF0 ) )
  {
    return( -52 );
  }
  u_64r = v3u64Lsh( u_64a, 36 );
  if ( !_V3_EQ64( u_64r, 9ABCDEF0, 00000000 ) )
  {
    return( -53 );
  }

  /* u64 Rsh */
  u_64r = v3u64Rsh( u_64a, 4 );
  if ( !_V3_EQ64( u_64r, 00123456, 789ABCDE ) )
  {
    return( -54 );
  }
  u_64r = v3u64Rsh( u_64a, 36 );
  if ( !_V3_EQ64( u_64r, 00000000, 00123456 ) )
  {
    return( -55 );
  }

  /* u64 Eq */
  _V3_SET64( u_64a, 01234567, 89ABCDEF );
  _V3_SET64( u_64b, FEDC89AB, 76543210 );
  if ( !v3u64Eq( u_64a, u_64a ) )
  {
    return( -56 );
  }
  if ( !v3u64Eq( u_64b, u_64b ) )
  {
    return( -57 );
  }
  if ( v3u64Eq( u_64a, u_64b ) )
  {
    return( -58 );
  }
  if ( v3u64Eq( u_64b, u_64a ) )
  {
    return( -59 );
  }

  /* u64 Gt */
  if ( v3u64Gt( u_64a, u_64a ) )
  {
    return( -60 );
  }
  if ( v3u64Gt( u_64b, u_64b ) )
  {
    return( -61 );
  }
  if ( v3u64Gt( u_64a, u_64b ) )
  {
    return( -62 );
  }
  if ( !v3u64Gt( u_64b, u_64a ) )
  {
    return( -63 );
  }

  /* u64 Lt */
  if ( v3u64Lt( u_64a, u_64a ) )
  {
    return( -64 );
  }
  if ( v3u64Lt( u_64b, u_64b ) )
  {
    return( -65 );
  }
  if ( !v3u64Lt( u_64a, u_64b ) )
  {
    return( -66 );
  }
  if ( v3u64Lt( u_64b, u_64a ) )
  {
    return( -67 );
  }

  /* s64 functions ====================================================== */

  /* s64 Add */
  _V3_SET64( s_64a, 01234567, 89ABCDEF );
  _V3_SET64( s_64b, FEDC89AB, 76543210 );

  s_64r = v3s64Add( s_64a, s_64b );
  if ( !_V3_EQ64( s_64r, FFFFCF12, FFFFFFFF ) )
  {
    return( -68 );
  }
  s_64r = v3s64Add( s_64b, s_64a );
  if ( !_V3_EQ64( s_64r, FFFFCF12, FFFFFFFF ) )
  {
    return( -69 );
  }
  s_64r = v3s64Add( s_64a, s_64a );
  if ( !_V3_EQ64( s_64r, 02468ACF, 13579BDE ) )
  {
    return( -70 );
  }
  s_64r = v3s64Add( s_64b, s_64b );
  if ( !_V3_EQ64( s_64r, FDB91356, ECA86420 ) )
  {
    return( -71 );
  }

  /* s64Sub */
  _V3_SET64( s_64a, 01234567, 89ABCDEF );
  _V3_SET64( s_64b, FEDC89AB, 76543210 );

  s_64r = v3s64Sub( s_64a, s_64b );
  if ( !_V3_EQ64( s_64r, 0246BBBC, 13579BDF ) )
  {
    return( -72 );
  }

  s_64r = v3s64Sub( s_64b, s_64a );
  if ( !_V3_EQ64( s_64r, FDB94443, ECA86421 ) )
  {
    return( -73 );
  }

  s_64r = v3s64Sub( s_64a, s_64a );
  if ( !_V3_EQ64( s_64r, 00000000, 00000000 ) )
  {
    return( -74 );
  }

  s_64r = v3s64Sub( s_64b, s_64b );
  if ( !_V3_EQ64( s_64r, 00000000, 00000000 ) )
  {
    return( -75 );
  }

  /* s64 Mul */
  _V3_SET64( s_64a, 01234567, 89ABCDEF );
  _V3_SET64( s_64b, FEDC89AB, 76543210 );

  s_64r = v3s64Mul( s_64a, s_64a );
  if ( !_V3_EQ64( s_64r, DCA5E208, 90F2A521 ) )
  {
    return( -76 );
  }

  s_64r = v3s64Mul( s_64b, s_64b );
  if ( !_V3_EQ64( s_64r, 3441BB37, A44A4100 ) )
  {
    return( -77 );
  }

  s_64r = v3s64Mul( s_64a, s_64b );
  if ( !_V3_EQ64( s_64r, 778C624C, E5618CF0 ) )
  {
    return( -78 );
  }

  s_64r = v3s64Mul( s_64b, s_64a );
  if ( !_V3_EQ64( s_64r, 778C624C, E5618CF0 ) )
  {
    return( -79 );
  }

  /* s64 Div, Mod */
  _V3_SET64( s_64a, 01234567, 89ABCDEF );
  _V3_SET64( s_64b, FEDC89AB, 76543210 );

  s_64r = v3s64Div( s_64a, s_64b );
  if ( !_V3_EQ64( s_64r, 00000000, 00000000 ) )
  {
    return( -80 );
  }

  s_64r = v3s64Mod( s_64a, s_64b );
  if ( !_V3_EQ64( s_64r, 000030ED, 00000001 ) )
  {
    return( -81 );
  }

  s_64r = v3s64Div( s_64b, s_64a );
  if ( !_V3_EQ64( s_64r, FFFFFFFF, FFFFFFFF ) )
  {
    return( -82 );
  }

  s_64r = v3s64Mod( s_64b, s_64a );
  if ( !_V3_EQ64( s_64r, 0123147A, 89ABCDEE ) )
  {
    return( -83 );
  }

  /* s64 Inc */
  _V3_SET64( s_64r, FFFFFFFF, FFFFFFFE );

  v3s64Inc( &s_64r );
  if ( !_V3_EQ64( s_64r, FFFFFFFF, FFFFFFFF ) )
  {
    return( -84 );
  }
  v3s64Inc( &s_64r );
  if ( !_V3_EQ64( s_64r, 00000000, 00000000 ) )
  {
    return( -85 );
  }

  /* s64 Dec */
  v3s64Dec( &s_64r );
  if ( !_V3_EQ64( s_64r, FFFFFFFF, FFFFFFFF ) )
  {
    return( -86 );
  }
  v3s64Dec( &s_64r );
  if ( !_V3_EQ64( s_64r, FFFFFFFF, FFFFFFFE ) )
  {
    return( -87 );
  }

  /* s64 Eq */
  _V3_SET64( s_64a, 01234567, 89ABCDEF );
  _V3_SET64( s_64b, FEDC89AB, 76543210 );

  if ( !v3s64Eq( s_64a, s_64a ) )
  {
    return( -88 );
  }
  if ( !v3s64Eq( s_64b, s_64b ) )
  {
    return( -89 );
  }
  if ( v3s64Eq( s_64a, s_64b ) )
  {
    return( -90 );
  }
  if ( v3s64Eq( s_64b, s_64a ) )
  {
    return( -91 );
  }

  /* s64 Gt */
  if ( v3s64Gt( s_64a, s_64a ) )
  {
    return( -92 );
  }
  if ( v3s64Gt( s_64b, s_64b ) )
  {
    return( -93 );
  }
  if ( !v3s64Gt( s_64a, s_64b ) )
  {
    return( -94 );
  }
  if ( v3s64Gt( s_64b, s_64a ) )
  {
    return( -95 );
  }

  /* s64 Lt */
  if ( v3s64Lt( s_64a, s_64a ) )
  {
    return( -96 );
  }
  if ( v3s64Lt( s_64b, s_64b ) )
  {
    return( -97 );
  }
  if ( v3s64Lt( s_64a, s_64b ) )
  {
    return( -98 );
  }
  if ( !v3s64Lt( s_64b, s_64a ) )
  {
    return( -99 );
  }

  /* u128 functions ===================================================== */

  /* u128 Add */
  _V3_SET128( u_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
  _V3_SET128( u_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );

  u_128r = v3u128Add( u_128a, u_128b );
  if ( !_V3_EQ128( u_128r, FFFFCF12, FFFFFFFF, CF12FFFF, FFFFFFFF ) )
  {
    return( -100 );
  }
  u_128r = v3u128Add( u_128b, u_128a );
  if ( !_V3_EQ128( u_128r, FFFFCF12, FFFFFFFF, CF12FFFF, FFFFFFFF ) )
  {
    return( -101 );
  }
  u_128r = v3u128Add( u_128a, u_128a );
  if ( !_V3_EQ128( u_128r, 02468ACF, 13579BDE, 8ACE0247, 9BDF1356 ) )
  {
    return( -102 );
  }
  u_128r = v3u128Add( u_128b, u_128b );
  if ( !_V3_EQ128( u_128r, FDB91356, ECA86421, 1357FDB8, 6420ECA8 ) )
  {
    return( -103 );
  }

  /* u128 Sub */
  _V3_SET128( u_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
  _V3_SET128( u_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );

  u_128r = v3u128Sub( u_128a, u_128b );
  if ( !_V3_EQ128( u_128r, 0246BBBC, 13579BDE, BBBB0247, 9BDF1357 ) )
  {
    return( -104 );
  }

  u_128r = v3u128Sub( u_128b, u_128a );
  if ( !_V3_EQ128( u_128r, FDB94443, ECA86421, 4444FDB8, 6420ECA9 ) )
  {
    return( -105 );
  }

  u_128r = v3u128Sub( u_128a, u_128a );
  if ( !_V3_EQ128( u_128r, 00000000, 00000000, 00000000, 00000000 ) )
  {
    return( -106 );
  }

  u_128r = v3u128Sub( u_128b, u_128b );
  if ( !_V3_EQ128( u_128r, 00000000, 00000000, 00000000, 00000000 ) )
  {
    return( -107 );
  }

  /* u128 Mul */
  _V3_SET128( u_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
  _V3_SET128( u_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );

  u_128r = v3u128Mul( u_128a, u_128a );
  if ( !_V3_EQ128( u_128r, 4AF09DB0, 50050EDE, C9D67D60, 35527839 ) )
  {
    return( -108 );
  }

  u_128r = v3u128Mul( u_128b, u_128b );
  if ( !_V3_EQ128( u_128r, 7EC36D85, ECE1CBE2, BD1C7FA7, D1318B90 ) )
  {
    return( -109 );
  }

  u_128r = v3u128Mul( u_128a, u_128b );
  if ( !_V3_EQ128( u_128r, 9FD30906, 618C929F, 6D73817B, FCBDFE1C ) )
  {
    return( -110 );
  }

  u_128r = v3u128Mul( u_128b, u_128a );
  if ( !_V3_EQ128( u_128r, 9FD30906, 618C929F, 6D73817B, FCBDFE1C ) )
  {
    return( -111 );
  }

  /* u128 Div, Mod */
  _V3_SET128( u_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
  _V3_SET128( u_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );

  u_128r = v3u128Div( u_128a, u_128b );
  if ( !_V3_EQ128( u_128r, 00000000, 00000000, 00000000, 00000000 ) )
  {
    return( -112 );
  }

  u_128r = v3u128Mod( u_128a, u_128b );
  if ( !_V3_EQ128( u_128r, 01234567, 89ABCDEF, 45670123, CDEF89AB ) )
  {
    return( -113 );
  }

  u_128r = v3u128Div( u_128b, u_128a );
  if ( !_V3_EQ128( u_128r, 00000000, 00000000, 00000000, 000000DF ) )
  {
    return( -114 );
  }

  u_128r = v3u128Mod( u_128b, u_128a );
  if ( !_V3_EQ128( u_128r, 0123147A, 89ABCEA3, 14F200AB, CE678A5F ) )
  {
    return( -115 );
  }

  /* u128 Inc */
  _V3_SET128( u_128r, FFFFFFFF, FFFFFFFF, FFFFFFFF, FFFFFFFE );

  v3u128Inc( &u_128r );
  if ( !_V3_EQ128( u_128r, FFFFFFFF, FFFFFFFF, FFFFFFFF, FFFFFFFF ) )
  {
    return( -116 );
  }
  v3u128Inc( &u_128r );
  if ( !_V3_EQ128( u_128r, 00000000, 00000000, 00000000, 00000000 ) )
  {
    return( -117 );
  }

  /* u128 Dec */
  v3u128Dec( &u_128r );
  if ( !_V3_EQ128( u_128r, FFFFFFFF, FFFFFFFF, FFFFFFFF, FFFFFFFF ) )
  {
    return( -118 );
  }
  v3u128Dec( &u_128r );
  if ( !_V3_EQ128( u_128r, FFFFFFFF, FFFFFFFF, FFFFFFFF, FFFFFFFE ) )
  {
    return( -119 );
  }

  /* u128 And */
  _V3_SET128( u_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
  _V3_SET128( u_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );

  u_128r = v3u128And( u_128a, u_128a );
  if ( !_V3_EQ128( u_128r, 01234567, 89ABCDEF, 45670123, CDEF89AB ) )
  {
    return( -120 );
  }
  u_128r = v3u128And( u_128a, u_128b );
  if ( !_V3_EQ128( u_128r, 00000123, 00000000, 01230000, 00000000 ) )
  {
    return( -121 );
  }

  /* u128 Or */
  u_128r = v3u128Or( u_128a, u_128a );
  if ( !_V3_EQ128( u_128r, 01234567, 89ABCDEF, 45670123, CDEF89AB ) )
  {
    return( -122 );
  }
  u_128r = v3u128Or( u_128a, u_128b );
  if ( !_V3_EQ128( u_128r, FFFFCDEF, FFFFFFFF, CDEFFFFF, FFFFFFFF ) )
  {
    return( -123 );
  }

  /* u128 Xor */
  /* r = a ^ b ^ a ^ b ^ a = a */
  u_128r = v3u128Xor( u_128a, u_128b );
  u_128r = v3u128Xor( u_128r, u_128a );
  u_128r = v3u128Xor( u_128r, u_128b );
  u_128r = v3u128Xor( u_128r, u_128a );
  if ( !_V3_EQ128( u_128r, 01234567, 89ABCDEF, 45670123, CDEF89AB ) )
  {
    return( -124 );
  }

  /* u128 Not */
  u_128r = v3u128Not( u_128a );
  if ( !_V3_EQ128( u_128r, FEDCBA98, 76543210, BA98FEDC, 32107654 ) )
  {
    return( -125 );
  }
  u_128r = v3u128Not( u_128b );
  if ( !_V3_EQ128( u_128r, 01237654, 89ABCDEF, 76540123, CDEF89AB ) )
  {
    return( -126 );
  }

  /* u128 Lsh */
  u_128r = v3u128Lsh( u_128a, 4 );
  if ( !_V3_EQ128( u_128r, 12345678, 9ABCDEF4, 5670123C, DEF89AB0 ) )
  {
    return( -127 );
  }
  u_128r = v3u128Lsh( u_128a, 36 );
  if ( !_V3_EQ128( u_128r, 9ABCDEF4, 5670123C, DEF89AB0, 00000000 ) )
  {
    return( -128 );
  }
  u_128r = v3u128Lsh( u_128a, 68 );
  if ( !_V3_EQ128( u_128r, 5670123C, DEF89AB0, 00000000, 00000000 ) )
  {
    return( -129 );
  }
  u_128r = v3u128Lsh( u_128a, 100 );
  if ( !_V3_EQ128( u_128r, DEF89AB0, 00000000, 00000000, 00000000 ) )
  {
    return( -130 );
  }

  /* 128 Rsh */
  u_128r = v3u128Rsh( u_128a, 4 );
  if ( !_V3_EQ128( u_128r, 00123456, 789ABCDE, F4567012, 3CDEF89A ) )
  {
    return( -131 );
  }
  u_128r = v3u128Rsh( u_128a, 36 );
  if ( !_V3_EQ128( u_128r, 00000000, 00123456, 789ABCDE, F4567012 ) )
  {
    return( -132 );
  }
  u_128r = v3u128Rsh( u_128a, 68 );
  if ( !_V3_EQ128( u_128r, 00000000, 00000000, 00123456, 789ABCDE ) )
  {
    return( -133 );
  }
  u_128r = v3u128Rsh( u_128a, 100 );
  if ( !_V3_EQ128( u_128r, 00000000, 00000000, 00000000, 00123456 ) )
  {
    return( -134 );
  }

  /* u128 Eq */
  _V3_SET128( u_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
  _V3_SET128( u_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );

  if ( !v3u128Eq( u_128a, u_128a ) )
  {
    return( -135 );
  }
  if ( !v3u128Eq( u_128b, u_128b ) )
  {
    return( -136 );
  }
  if ( v3u128Eq( u_128a, u_128b ) )
  {
    return( -137 );
  }
  if ( v3u128Eq( u_128b, u_128a ) )
  {
    return( -138 );
  }

  /* u128 Gt */
  if ( v3u128Gt( u_128a, u_128a ) )
  {
    return( -139 );
  }
  if ( v3u128Gt( u_128b, u_128b ) )
  {
    return( -140 );
  }
  if ( v3u128Gt( u_128a, u_128b ) )
  {
    return( -141 );
  }
  if ( !v3u128Gt( u_128b, u_128a ) )
  {
    return( -142 );
  }

  /* u128 Lt */
  if ( v3u128Lt( u_128a, u_128a ) )
  {
    return( -143 );
  }
  if ( v3u128Lt( u_128b, u_128b ) )
  {
    return( -144 );
  }
  if ( !v3u128Lt( u_128a, u_128b ) )
  {
    return( -145 );
  }
  if ( v3u128Lt( u_128b, u_128a ) )
  {
    return( -146 );
  }

  /* s128 functions ===================================================== */

  /* s128 Add */
  _V3_SET128( s_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
  _V3_SET128( s_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );

  s_128r = v3s128Add( s_128a, s_128b );
  if ( !_V3_EQ128( s_128r, FFFFCF12, FFFFFFFF, CF12FFFF, FFFFFFFF ) )
  {
    return( -147 );
  }
  s_128r = v3s128Add( s_128b, s_128a );
  if ( !_V3_EQ128( s_128r, FFFFCF12, FFFFFFFF, CF12FFFF, FFFFFFFF ) )
  {
    return( -148 );
  }
  s_128r = v3s128Add( s_128a, s_128a );
  if ( !_V3_EQ128( s_128r, 02468ACF, 13579BDE, 8ACE0247, 9BDF1356 ) )
  {
    return( -149 );
  }
  s_128r = v3s128Add( s_128b, s_128b );
  if ( !_V3_EQ128( s_128r, FDB91356, ECA86421, 1357FDB8, 6420ECA8 ) )
  {
    return( -150 );
  }

  /* u128 Sub */
  _V3_SET128( s_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
  _V3_SET128( s_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );

  s_128r = v3s128Sub( s_128a, s_128b );
  if ( !_V3_EQ128( s_128r, 0246BBBC, 13579BDE, BBBB0247, 9BDF1357 ) )
  {
    return( -151 );
  }

  s_128r = v3s128Sub( s_128b, s_128a );
  if ( !_V3_EQ128( s_128r, FDB94443, ECA86421, 4444FDB8, 6420ECA9 ) )
  {
    return( -152 );
  }

  s_128r = v3s128Sub( s_128a, s_128a );
  if ( !_V3_EQ128( s_128r, 00000000, 00000000, 00000000, 00000000 ) )
  {
    return( -153 );
  }

  s_128r = v3s128Sub( s_128b, s_128b );
  if ( !_V3_EQ128( s_128r, 00000000, 00000000, 00000000, 00000000 ) )
  {
    return( -154 );
  }

  /* s128 Mul */
  _V3_SET128( s_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
  _V3_SET128( s_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );

  s_128r = v3s128Mul( s_128a, s_128a );
  if ( !_V3_EQ128( s_128r, 4AF09DB0, 50050EDE, C9D67D60, 35527839 ) )
  {
    return( -155 );
  }

  s_128r = v3s128Mul( s_128b, s_128b );
  if ( !_V3_EQ128( s_128r, 7EC36D85, ECE1CBE2, BD1C7FA7, D1318B90 ) )
  {
    return( -156 );
  }

  s_128r = v3s128Mul( s_128a, s_128b );
  if ( !_V3_EQ128( s_128r, 9FD30906, 618C929F, 6D73817B, FCBDFE1C ) )
  {
    return( -157 );
  }

  s_128r = v3s128Mul( s_128b, s_128a );
  if ( !_V3_EQ128( s_128r, 9FD30906, 618C929F, 6D73817B, FCBDFE1C ) )
  {
    return( -158 );
  }

  /* s128 Div, Mod */
  _V3_SET128( s_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
  _V3_SET128( s_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );

  s_128r = v3s128Div( s_128a, s_128b );
  if ( !_V3_EQ128( s_128r, 00000000, 00000000, 00000000, 00000000 ) )
  {
    return( -159 );
  }

  s_128r = v3s128Mod( s_128a, s_128b );
  if ( !_V3_EQ128( s_128r, 000030ED, 00000000, 30ED0000, 00000001 ) )
  {
    return( -160 );
  }

  s_128r = v3s128Div( s_128b, s_128a );
  if ( !_V3_EQ128( s_128r, FFFFFFFF, FFFFFFFF, FFFFFFFF, FFFFFFFF ) )
  {
    return( -161 );
  }

  s_128r = v3s128Mod( s_128b, s_128a );
  if ( !_V3_EQ128( s_128r, 0123147A, 89ABCDEF, 147A0123, CDEF89AA ) )
  {
    return( -162 );
  }

  /* s128 Inc */
  _V3_SET128( s_128r, FFFFFFFF, FFFFFFFF, FFFFFFFF, FFFFFFFE );

  v3s128Inc( &s_128r );
  if ( !_V3_EQ128( s_128r, FFFFFFFF, FFFFFFFF, FFFFFFFF, FFFFFFFF ) )
  {
    return( -163 );
  }
  v3s128Inc( &s_128r );
  if ( !_V3_EQ128( s_128r, 00000000, 00000000, 00000000, 00000000 ) )
  {
    return( -164 );
  }

  /* s128 Dec */
  v3s128Dec( &s_128r );
  if ( !_V3_EQ128( s_128r, FFFFFFFF, FFFFFFFF, FFFFFFFF, FFFFFFFF ) )
  {
    return( -165 );
  }
  v3s128Dec( &s_128r );
  if ( !_V3_EQ128( s_128r, FFFFFFFF, FFFFFFFF, FFFFFFFF, FFFFFFFE ) )
  {
    return( -166 );
  }

  /* s128 Eq */
  _V3_SET128( s_128a, 01234567, 89ABCDEF, 45670123, CDEF89AB );
  _V3_SET128( s_128b, FEDC89AB, 76543210, 89ABFEDC, 32107654 );
  if ( !v3s128Eq( s_128a, s_128a ) )
  {
    return( -167 );
  }
  if ( !v3s128Eq( s_128b, s_128b ) )
  {
    return( -168 );
  }
  if ( v3s128Eq( s_128a, s_128b ) )
  {
    return( -169 );
  }
  if ( v3s128Eq( s_128b, s_128a ) )
  {
    return( -170 );
  }

  /** compare ** Gt */
  if ( v3s128Gt( s_128a, s_128a ) )
  {
    return( -171 );
  }
  if ( v3s128Gt( s_128b, s_128b ) )
  {
    return( -172 );
  }
  if ( !v3s128Gt( s_128a, s_128b ) )
  {
    return( -173 );
  }
  if ( v3s128Gt( s_128b, s_128a ) )
  {
    return( -174 );
  }

  /* s128 Lt */
  if ( v3s128Lt( s_128a, s_128a ) )
  {
    return( -175 );
  }
  if ( v3s128Lt( s_128b, s_128b ) )
  {
    return( -176 );
  }
  if ( v3s128Lt( s_128a, s_128b ) )
  {
    return( -177 );
  }
  if ( !v3s128Lt( s_128b, s_128a ) )
  {
    return( -178 );
  }

/*
---------------------------------------------------------------------------
 Auditing mark
---------------------------------------------------------------------------
*/


  return( V3_PASS );
}
