/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

/* CPU/OS Layer - CPU and OS specific code is allowed */

#ifndef V3_COSMMEM_H
#define V3_COSMMEM_H

#include "cputypes.h"

#define V3_MEM_NORMAL 41
#define V3_MEM_SECURE 42

void * v3MemAlloc( u64 bytes, u32 secure );
  /*
    Allocate a block of zeroed memory. Will allocate 16*floor((bytes+15)/16))
    bytes, i.e. blocks of 16 bytes, and only on 16 byte boundries.
    If secure is V3_MEM_NORMAL no special requirements are needed.
    If secure is V3_MEM_SECURE the program will expect the following:
      - The memory will be protected from read or write from any other
        process including supervisor run programs or other processes owned
        by the same user, unless explicitly given access by the process.
      - The memory will not be swapped or otherwise copied out of live RAM
        including during a crash triggered dump.
      - The memory will not cross CPU boundries.
      - The parts of the kernel in supervisor space will not be modified
        after boot.
    If the secure flag cannot be enforced, v3MemWarning() will be called and
    the function will return NULL. The user will then be warned any time the
    program attempts to do secure operations that the system is insecure.
    Returns: An aligned pointer to the RAM block, or NULL on error.
  */

void * v3MemRealloc( void * memory, u64 bytes, u32 secure );
  /*
    Adjust the size of the ram block to bytes length, copying any old data.
    [ideally we would zero new/old memory in the process]
    The secure flag will be treated the same as it is in v3MemAlloc above.
    Memory should also be aligned on 16 byte boundries similar to v3MemAlloc.
    If bytes is zero, then free the memory, and return NULL.
    Returns: A pointer to the new memory space on success, or NULL if unable
      to change the size (the old memory will not be modified).
  */

s32 v3MemCopy( void * dest, const void * src, u64 length );
  /*
    Copy length bytes of memory from src to dest.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3MemSet( void * memory, u64 length, u8 value );
  /*
    Set length bytes of memory to value.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3MemCmp( const void * blockA, const void * blockB, u64 max_bytes );
  /*
    Compare the memory blocks blockA and blockB, up to max_bytes will
    be compared.
    Returns: 0 if the blocks are identical. If they are different then a
      positive or negative based on memA[x] - memB[x]. Due to endian issues
      the +/- result may or may not be relivant.
  */

void * v3MemOffset( const void * memory, u64 offset );
  /*
    Calculates the address of memory+offset.
    Returns: pointer to the data location, NULL if memory space is exceeded.
  */

void v3MemFree( void * memory );
  /*
    Free the allocated memory.
    Ideally, overwrite the memory with 0xFF and free the allocated memory.
    Returns: nothing.
  */

s32 v3MemSystem( u64 * amount );
  /*
    Set amount to the number of bytes of physical RAM in the system.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

void v3MemWarning( void );
  /*
    Force display of a warning message to the user, when secure memory is
    requested but not available.
    Returns: nothing.
  */

/* testing */

s32 v3_TestCosmMem( void );
  /*
    Test functions in this header.
    Returns: V3_PASS on success, or a negative number corresponding to the
      test that failed.
  */

#endif
