/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#include "security.h"
#ifdef RANDOM_DEBUG
#include "cosmio.h"
#endif

s32 v3HashBegin( v3_HASH_TMP * work, u32 type )
{
  if ( work == NULL )
    return( V3_HASH_ERROR_PARAM );

  v3MutexLock( &work->lock, V3_MUTEX_WAIT );

  switch ( type )
  {
    case V3_HASH_MD5:
      work->tmp_data = v3_MD5Begin();
      break;
    case V3_HASH_SHA1:
      work->tmp_data = v3_SHA1Begin();
      break;
    default:
      v3MutexUnlock( &work->lock );
      return( V3_HASH_ERROR_TYPE );
  }

  if ( work->tmp_data == NULL )
  {
    work->state = V3_HASH_STATE_ERROR;
    v3MutexUnlock( &work->lock );
    return ( V3_HASH_ERROR_MEMORY );
  }

  work->type = type;
  work->state = V3_HASH_STATE_INIT;

  v3MutexUnlock( &work->lock );
  return ( V3_PASS );
}

s32 v3HashUpdate( v3_HASH_TMP * work, const u8 * const data, u64 length )
{
  if ( work == NULL || data == NULL )
    return( V3_HASH_ERROR_PARAM );

  if ( ( work->state != V3_HASH_STATE_DATA &&
    work->state != V3_HASH_STATE_INIT ) || work->tmp_data == NULL )
    return( V3_HASH_ERROR_STATE );

  v3MutexLock( &work->lock, V3_MUTEX_WAIT );

  work->state = V3_HASH_STATE_DATA;

  switch( work->type )
  {
    case V3_HASH_MD5:
      v3_MD5Update( (v3_MD5_CONTEXT *) work->tmp_data, data, length );
      break;
    case V3_HASH_SHA1:
      v3_SHA1Update( (v3_SHA1_CONTEXT *) work->tmp_data, data, length );
      break;
    default:
      work->state = V3_HASH_STATE_ERROR;
      v3MutexUnlock( &work->lock );
      return( V3_HASH_ERROR_STATE );
  }

  v3MutexUnlock( &work->lock );
  return( V3_PASS );
}

s32 v3HashEnd( v3_HASH * hash, v3_HASH_TMP * work )
{
  if ( hash == NULL || work == NULL )
    return( V3_HASH_ERROR_MEMORY );

  if ( work->state != V3_HASH_STATE_DATA || work->tmp_data == NULL )
    return( V3_HASH_ERROR_STATE );

  v3MutexLock( &work->lock, V3_MUTEX_WAIT );

  work->state = V3_HASH_STATE_DONE;
  v3MemSet( hash, v3u64u32( sizeof( v3_HASH ) ), (u8) 0 );

  switch( work->type )
  {
    case V3_HASH_MD5:
      v3_MD5Final( (u8 *) hash->hash, (v3_MD5_CONTEXT *) work->tmp_data );
      hash->type = work->type;
      break;
    case V3_HASH_SHA1:
      v3_SHA1Final( (u8 *) hash->hash, (v3_SHA1_CONTEXT *) work->tmp_data );
      hash->type = work->type;
      break;
    default:
      work->state = V3_HASH_STATE_ERROR;
      v3MutexUnlock( &work->lock );
      return( V3_HASH_ERROR_STATE );
  }

  /* Free our scratch space. */
  v3MemFree( (void *) work->tmp_data );

  v3MutexUnlock( &work->lock );
  v3MutexFree( &work->lock );

  return( V3_PASS );
}

s32 v3HashEq( const v3_HASH * hashA, const v3_HASH * hashB )
{
  u32 i;

  if ( hashA->type != hashB->type )
  {
    return( 0 );
  }

  for( i = 0; i < 32; i++ )
  {
    if ( hashA->hash[i] != hashB->hash[i] )
    {
      return( 0 );
    }
  }

  return( 1 );
}

#if ( defined( NO_CRYPTO ) )
/* define stubs for crypto functions that return V3_FAIL */
#else /* have CRYPTO */
/* include file with cryptographic routines */
#include "crypto.c"
#endif /* NO_CRYPTO */

s32 v3Random( v3_RANDOM * random, u64 length, u8 * data,
  u64 salt_length, const u8 * salt )
{
#if ( defined( NO_CRYPTO ) )
  u32 i, j, k, st_idx, st_num;
  v3_HASH md;
  v3_HASH_TMP * work;

  /*
    Initialization of the random number generator is done by feeding it
    salt_length bytes of salt with data == NULL.
  */
  if ( ( salt != NULL ) && ( data == NULL ) )
  {
    st_idx = random->state_index;
    st_num = random->state_num;

    random->state_index = random->state_index + v3u32u64( salt_length );

    if ( random->state_index >= V3_RANDOM_STATE_SIZE )
    {
      random->state_index = random->state_index % V3_RANDOM_STATE_SIZE;
      random->state_num = V3_RANDOM_STATE_SIZE;
    }
    else if ( random->state_num < V3_RANDOM_STATE_SIZE )
    {
      if ( random->state_index > random->state_num )
        random->state_num = random->state_index;
    }

    /* Try to allocate memory with V3_MEM_SECURE if possible */
    if( ( work = v3MemAlloc( v3u64u32( sizeof( v3_HASH_TMP ) ),
      V3_MEM_SECURE ) ) == NULL )
      if( ( work = v3MemAlloc( v3u64u32( sizeof( v3_HASH_TMP ) ),
        V3_MEM_NORMAL ) ) == NULL )
        return( V3_FAIL );

    for ( i = 0; i < v3u32u64( salt_length ); i += MD_DIGEST_LENGTH )
    {
      j = ( v3u32u64( salt_length ) - i );
      j = ( j > MD_DIGEST_LENGTH ) ? MD_DIGEST_LENGTH : j;

      if ( v3HashBegin( work, V3_HASH_MD5 ) != V3_PASS )
      {
        v3MemFree( work );
        return( V3_FAIL );
      }
      if ( v3HashUpdate( work, (u8 *) md.hash, v3u64u32( MD_DIGEST_LENGTH ) )
        != V3_PASS )
      {
        v3MemFree( work );
        return( V3_FAIL );
      }

      k = ( st_idx + j ) - V3_RANDOM_STATE_SIZE;

      if ( k > 0 )
      {
        if ( v3HashUpdate( work, (u8 *) &random->state[st_idx],
          v3u64u32( j - k ) ) != V3_PASS )
        {
          v3MemFree( work );
          return( V3_FAIL );
        }
        if ( v3HashUpdate( work, (u8 *) &random->state[0], v3u64u32( j ) )
          != V3_PASS )
        {
          v3MemFree( work );
          return( V3_FAIL );
        }
      }
      else
      {
       if ( v3HashUpdate( work, (u8 *) &random->state[st_idx], v3u64u32( j ) )
          != V3_PASS )
        {
          v3MemFree( work );
          return( V3_FAIL );
        }
      }

      if ( v3HashUpdate( work, (u8 *) &salt, v3u64u32( j ) ) != V3_PASS )
      {
        v3MemFree( work );
        return( V3_FAIL );
      }

      if ( v3HashEnd( &md, work ) != V3_PASS )
      {
        v3MemFree( work );
        return( V3_FAIL );
      }

      salt = salt + j;

      for ( k = 0 ; k < j ; k++ )
      {
        random->state[st_idx++] ^= md.hash[k];

        if (st_idx >= V3_RANDOM_STATE_SIZE)
        {
          st_idx = 0;
          st_num = V3_RANDOM_STATE_SIZE;
        }
      }
      /* Zeroize work */
      v3MemSet( work, v3u64u32( sizeof( v3_HASH_TMP ) ), 0 );
    }
    v3MemFree( work );
    return( V3_PASS );
  }

  /*
    Generate length bytes of deterministic random bits if we call
    with salt == NULL, and place them into data;
  */
  else if ( ( salt == NULL ) && !( v3u64Eq( length, v3u64u32( 0  ) ) ) )
  {
    st_idx = random->state_index;
    st_num = random->state_num;

    random->state_index = random->state_index+v3u32u64( length );

    if (random->state_index > random->state_num)
      random->state_index = random->state_index % random->state_num;

    if( ( work = v3MemAlloc( v3u64u32( sizeof( v3_HASH_TMP ) ),
      V3_MEM_SECURE ) ) == NULL )
      if( ( work = v3MemAlloc( v3u64u32( sizeof( v3_HASH_TMP ) ),
        V3_MEM_NORMAL ) ) == NULL )
        return( V3_FAIL );

    while ( v3u64Gt(length, v3u64u32(0)) )
    {

      /* If length >= MD_DIGEST_LENGTH */
      if ( ( v3u64Gt( length, v3u64u32( MD_DIGEST_LENGTH / 2 ) ) ) ||
        ( v3u64Eq( length, v3u64u32( MD_DIGEST_LENGTH / 2 ) ) ) )
      {
        j = MD_DIGEST_LENGTH / 2;
      }
      else
      {
        j = v3u32u64( length );
      }

      length = v3u64Sub( length, v3u64u32( j ) );

      if ( v3HashBegin( work, V3_HASH_MD5 ) != V3_PASS )
      {
        v3MemFree( work );
        return( V3_FAIL );
      }
      if ( v3HashUpdate( work, (u8 *) &md.hash[MD_DIGEST_LENGTH/2],
        v3u64u32( MD_DIGEST_LENGTH/2 ) )
        != V3_PASS )
      {
        v3MemFree( work );
        return( V3_FAIL );
      }

      k = ( st_idx + j ) - st_num;

      if (k > 0)
      {
        if ( v3HashUpdate( work, (u8 *) &random->state[st_idx],
          v3u64u32( j - k ) ) != V3_PASS )
        {
          v3MemFree( work );
          return( V3_FAIL );
        }
        if ( v3HashUpdate( work, (u8 *) &random->state[0], v3u64u32(k) )
          != V3_PASS )
        {
          v3MemFree( work );
          return( V3_FAIL );
        }
      }
      else
      {
       if ( v3HashUpdate( work, (u8 *) &random->state[st_idx], v3u64u32(j) )
          != V3_PASS )
        {
          v3MemFree( work );
          return( V3_FAIL );
        }
      }
      if ( v3HashEnd( &md, work ) != V3_PASS )
      {
        v3MemFree( work );
        return( V3_FAIL );
      }

      for (i=0; i<j; i++)
      {
        if ( st_idx >= st_num )
          st_idx=0;
	random->state[st_idx++] ^= md.hash[i];
	*(data++) = md.hash[i+MD_DIGEST_LENGTH/2];
      }
    }
    v3MemFree(work);
    return( V3_PASS );
  }

  /*
    Generate a non-deterministic random sequence if we call with salt != NULL.
    The OS will also be asked for some random data if it supports that
    feature exists and salt != NULL.
  */
  else if ( ( salt != NULL ) && !( v3u64Eq( length, v3u64u32( 0  ) ) ) )
  {
    return( V3_PASS );
  }
#else
  /* Yarrow */
#endif

  return( V3_FAIL );
}

/* Some general math macros.  These are used by both hashes. */
#define _V3_TRUNC32( x )   ( (u32) ( x ) & 0xFFFFFFFF )
#define _V3_ROTL32( x, n ) _V3_TRUNC32( ( ( x << n ) | ( x >> ( 32 - n ) ) ) )

/*
  ---------------------------------------------------------
   RSA Data Security, Inc., MD5 message-digest algorithm.
  ---------------------------------------------------------
*/

/* F, G, H and I are basic MD5 functions. */
#define _V3_MD5_F( x, y, z )  ( ( (x) & (y) ) | ( (~x) & (z) ) )
#define _V3_MD5_G( x, y, z )  ( ( (x) & (z) ) | ( (y) & (~z) ) )
#define _V3_MD5_H( x, y, z )  ( (x) ^ (y) ^ (z) )
#define _V3_MD5_I( x, y, z )  ( (y) ^ ( (x) | (~z) ) )

/* Lots of constants in MD5. */
#define _V3_MD5_CONST11 7
#define _V3_MD5_CONST12 12
#define _V3_MD5_CONST13 17
#define _V3_MD5_CONST14 22
#define _V3_MD5_CONST21 5
#define _V3_MD5_CONST22 9
#define _V3_MD5_CONST23 14
#define _V3_MD5_CONST24 20
#define _V3_MD5_CONST31 4
#define _V3_MD5_CONST32 11
#define _V3_MD5_CONST33 16
#define _V3_MD5_CONST34 23
#define _V3_MD5_CONST41 6
#define _V3_MD5_CONST42 10
#define _V3_MD5_CONST43 15
#define _V3_MD5_CONST44 21

/*
  FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
  Rotation is separate from addition to prevent recomputation.
*/

#define _V3_MD5_FF( a, b, c, d, x, s, ac ) \
  { \
    a += _V3_MD5_F( b, c, d ) + x + ac; \
    a =  _V3_ROTL32( a, s ); \
    a += b; \
  }
#define _V3_MD5_GG( a, b, c, d, x, s, ac ) \
  { \
    a += _V3_MD5_G( b, c, d ) + x + ac; \
    a =  _V3_ROTL32( a, s ); \
    a += b; \
  }
#define _V3_MD5_HH( a, b, c, d, x, s, ac ) \
  { \
    a += _V3_MD5_H( b, c, d ) + x + ac; \
    a = _V3_ROTL32( a, s ); \
    a += b; \
  }
#define _V3_MD5_II( a, b, c, d, x, s, ac ) \
  { \
    a += _V3_MD5_I( b, c, d ) + x + ac; \
    a = _V3_ROTL32( a, s ); \
    a += (b); \
  }

/*
  v3_MD5Begin - Initalize a v3_MD5_CONTEXT variable for use in a new hash.
*/

v3_MD5_CONTEXT * v3_MD5Begin( void )
{
  v3_MD5_CONTEXT * context;
  u64 mem_size = v3u64u32( sizeof( v3_MD5_CONTEXT ) );

  if( ( context = v3MemAlloc( mem_size, V3_MEM_SECURE ) ) == NULL )
    if( ( context = v3MemAlloc( mem_size, V3_MEM_NORMAL ) ) == NULL )
      return( NULL );

  _V3_SET64( context->count, 00000000, 00000000 );

  /* Load magic initialization constants. */
  context->digest[0] = 0x67452301;
  context->digest[1] = 0xEFCDAB89;
  context->digest[2] = 0x98BADCFE;
  context->digest[3] = 0x10325476;

  return( context );
}

/*
  v3_MD5Update - Add data to an MD5 hash.
*/
void v3_MD5Update( v3_MD5_CONTEXT * context, const u8 * buffer, u64 length )
{
  u32 bytes;
  u32 offset;

  offset = v3u32u64( context->count ) & 0x3F;
  context->count = v3u64Add( context->count, length );

  while( v3u64Gt( length, v3u64u32( 0 ) ) )
  {
    bytes = 64 - offset;
    if ( v3u64Gt( v3u64u32( bytes ), length ) )
      bytes = v3u32u64( length );

    v3MemCopy( &context->buffer[offset], buffer, v3u64u32( bytes ) );

    length = v3u64Sub( length, v3u64u32( bytes ) );
    buffer += bytes;
    offset = ( offset + bytes ) & 0x3F;

    if ( offset == 0 )
      v3_MD5Transform( context );
  }
}

/*
  v3_MD5Transform - The core transform routine for MD5.
*/

void v3_MD5Transform( v3_MD5_CONTEXT * context )
{
  u32 a = context->digest[0], b = context->digest[1];
  u32 c = context->digest[2], d = context->digest[3];
  u32 x[16];
  u32 i;
  u8 * p;

  /* load the 64 u8's to 16 u32's from little endian */
  for ( i = 0 ; i < 16 ; i++ )
  {
    p = &context->buffer[( i * 4 )];
    x[i] = ( (u32) p[3] << 24 ) | ( (u32) p[2] << 16 ) |
      ( (u32) p[1] << 8 ) | (u32) p[0];
  }

  /* Round 1 */
  _V3_MD5_FF( a, b, c, d, x[ 0], _V3_MD5_CONST11, 0xD76AA478 ); /* 1 */
  _V3_MD5_FF( d, a, b, c, x[ 1], _V3_MD5_CONST12, 0xE8C7B756 ); /* 2 */
  _V3_MD5_FF( c, d, a, b, x[ 2], _V3_MD5_CONST13, 0x242070DB ); /* 3 */
  _V3_MD5_FF( b, c, d, a, x[ 3], _V3_MD5_CONST14, 0xC1BDCEEE ); /* 4 */
  _V3_MD5_FF( a, b, c, d, x[ 4], _V3_MD5_CONST11, 0xF57C0FAF ); /* 5 */
  _V3_MD5_FF( d, a, b, c, x[ 5], _V3_MD5_CONST12, 0x4787C62A ); /* 6 */
  _V3_MD5_FF( c, d, a, b, x[ 6], _V3_MD5_CONST13, 0xA8304613 ); /* 7 */
  _V3_MD5_FF( b, c, d, a, x[ 7], _V3_MD5_CONST14, 0xFD469501 ); /* 8 */
  _V3_MD5_FF( a, b, c, d, x[ 8], _V3_MD5_CONST11, 0x698098D8 ); /* 9 */
  _V3_MD5_FF( d, a, b, c, x[ 9], _V3_MD5_CONST12, 0x8B44F7AF ); /* 10 */
  _V3_MD5_FF( c, d, a, b, x[10], _V3_MD5_CONST13, 0xFFFF5BB1 ); /* 11 */
  _V3_MD5_FF( b, c, d, a, x[11], _V3_MD5_CONST14, 0x895CD7BE ); /* 12 */
  _V3_MD5_FF( a, b, c, d, x[12], _V3_MD5_CONST11, 0x6B901122 ); /* 13 */
  _V3_MD5_FF( d, a, b, c, x[13], _V3_MD5_CONST12, 0xFD987193 ); /* 14 */
  _V3_MD5_FF( c, d, a, b, x[14], _V3_MD5_CONST13, 0xA679438E ); /* 15 */
  _V3_MD5_FF( b, c, d, a, x[15], _V3_MD5_CONST14, 0x49B40821 ); /* 16 */

  /* Round 2 */
  _V3_MD5_GG (a, b, c, d, x[ 1], _V3_MD5_CONST21, 0xF61E2562 ); /* 17 */
  _V3_MD5_GG (d, a, b, c, x[ 6], _V3_MD5_CONST22, 0xC040B340 ); /* 18 */
  _V3_MD5_GG (c, d, a, b, x[11], _V3_MD5_CONST23, 0x265E5A51 ); /* 19 */
  _V3_MD5_GG (b, c, d, a, x[ 0], _V3_MD5_CONST24, 0xE9B6C7AA ); /* 20 */
  _V3_MD5_GG (a, b, c, d, x[ 5], _V3_MD5_CONST21, 0xD62F105D ); /* 21 */
  _V3_MD5_GG (d, a, b, c, x[10], _V3_MD5_CONST22, 0x02441453 ); /* 22 */
  _V3_MD5_GG (c, d, a, b, x[15], _V3_MD5_CONST23, 0xD8A1E681 ); /* 23 */
  _V3_MD5_GG (b, c, d, a, x[ 4], _V3_MD5_CONST24, 0xE7D3FBC8 ); /* 24 */
  _V3_MD5_GG (a, b, c, d, x[ 9], _V3_MD5_CONST21, 0x21E1CDE6 ); /* 25 */
  _V3_MD5_GG (d, a, b, c, x[14], _V3_MD5_CONST22, 0xC33707D6 ); /* 26 */
  _V3_MD5_GG (c, d, a, b, x[ 3], _V3_MD5_CONST23, 0xF4D50D87 ); /* 27 */
  _V3_MD5_GG (b, c, d, a, x[ 8], _V3_MD5_CONST24, 0x455A14ED ); /* 28 */
  _V3_MD5_GG (a, b, c, d, x[13], _V3_MD5_CONST21, 0xA9E3E905 ); /* 29 */
  _V3_MD5_GG (d, a, b, c, x[ 2], _V3_MD5_CONST22, 0xFCEFA3F8 ); /* 30 */
  _V3_MD5_GG (c, d, a, b, x[ 7], _V3_MD5_CONST23, 0x676F02D9 ); /* 31 */
  _V3_MD5_GG (b, c, d, a, x[12], _V3_MD5_CONST24, 0x8D2A4C8A ); /* 32 */

  /* Round 3 */
  _V3_MD5_HH (a, b, c, d, x[ 5], _V3_MD5_CONST31, 0xFFFA3942 ); /* 33 */
  _V3_MD5_HH (d, a, b, c, x[ 8], _V3_MD5_CONST32, 0x8771F681 ); /* 34 */
  _V3_MD5_HH (c, d, a, b, x[11], _V3_MD5_CONST33, 0x6D9D6122 ); /* 35 */
  _V3_MD5_HH (b, c, d, a, x[14], _V3_MD5_CONST34, 0xFDE5380C ); /* 36 */
  _V3_MD5_HH (a, b, c, d, x[ 1], _V3_MD5_CONST31, 0xA4BEEA44 ); /* 37 */
  _V3_MD5_HH (d, a, b, c, x[ 4], _V3_MD5_CONST32, 0x4BDECFA9 ); /* 38 */
  _V3_MD5_HH (c, d, a, b, x[ 7], _V3_MD5_CONST33, 0xF6BB4B60 ); /* 39 */
  _V3_MD5_HH (b, c, d, a, x[10], _V3_MD5_CONST34, 0xBEBFBC70 ); /* 40 */
  _V3_MD5_HH (a, b, c, d, x[13], _V3_MD5_CONST31, 0x289B7EC6 ); /* 41 */
  _V3_MD5_HH (d, a, b, c, x[ 0], _V3_MD5_CONST32, 0xEAA127FA ); /* 42 */
  _V3_MD5_HH (c, d, a, b, x[ 3], _V3_MD5_CONST33, 0xD4EF3085 ); /* 43 */
  _V3_MD5_HH (b, c, d, a, x[ 6], _V3_MD5_CONST34, 0x04881D05 ); /* 44 */
  _V3_MD5_HH (a, b, c, d, x[ 9], _V3_MD5_CONST31, 0xD9D4D039 ); /* 45 */
  _V3_MD5_HH (d, a, b, c, x[12], _V3_MD5_CONST32, 0xE6DB99E5 ); /* 46 */
  _V3_MD5_HH (c, d, a, b, x[15], _V3_MD5_CONST33, 0x1FA27CF8 ); /* 47 */
  _V3_MD5_HH (b, c, d, a, x[ 2], _V3_MD5_CONST34, 0xC4AC5665 ); /* 48 */

  /* Round 4 */
  _V3_MD5_II (a, b, c, d, x[ 0], _V3_MD5_CONST41, 0xF4292244 ); /* 49 */
  _V3_MD5_II (d, a, b, c, x[ 7], _V3_MD5_CONST42, 0x432AFF97 ); /* 50 */
  _V3_MD5_II (c, d, a, b, x[14], _V3_MD5_CONST43, 0xAB9423A7 ); /* 51 */
  _V3_MD5_II (b, c, d, a, x[ 5], _V3_MD5_CONST44, 0xFC93A039 ); /* 52 */
  _V3_MD5_II (a, b, c, d, x[12], _V3_MD5_CONST41, 0x655B59C3 ); /* 53 */
  _V3_MD5_II (d, a, b, c, x[ 3], _V3_MD5_CONST42, 0x8F0CCC92 ); /* 54 */
  _V3_MD5_II (c, d, a, b, x[10], _V3_MD5_CONST43, 0xFFEFF47D ); /* 55 */
  _V3_MD5_II (b, c, d, a, x[ 1], _V3_MD5_CONST44, 0x85845DD1 ); /* 56 */
  _V3_MD5_II (a, b, c, d, x[ 8], _V3_MD5_CONST41, 0x6FA87E4F ); /* 57 */
  _V3_MD5_II (d, a, b, c, x[15], _V3_MD5_CONST42, 0xFE2CE6E0 ); /* 58 */
  _V3_MD5_II (c, d, a, b, x[ 6], _V3_MD5_CONST43, 0xA3014314 ); /* 59 */
  _V3_MD5_II (b, c, d, a, x[13], _V3_MD5_CONST44, 0x4E0811A1 ); /* 60 */
  _V3_MD5_II (a, b, c, d, x[ 4], _V3_MD5_CONST41, 0xF7537E82 ); /* 61 */
  _V3_MD5_II (d, a, b, c, x[11], _V3_MD5_CONST42, 0xBD3AF235 ); /* 62 */
  _V3_MD5_II (c, d, a, b, x[ 2], _V3_MD5_CONST43, 0x2AD7D2BB ); /* 63 */
  _V3_MD5_II (b, c, d, a, x[ 9], _V3_MD5_CONST44, 0xEB86D391 ); /* 64 */

  context->digest[0] += a;
  context->digest[1] += b;
  context->digest[2] += c;
  context->digest[3] += d;

  /* Zeroize sensitive information. */
  v3MemSet( x, v3u64u32( sizeof( x ) ), 0 );
}

/*
  v3_MD5Final - The final routine to retrieve the MD5 digest.
*/
void v3_MD5Final( u8 * hash, v3_MD5_CONTEXT * context )
{
  u64 bit_count;
  u8 pad, bytes[8];
  s32 i, j;

  bit_count = v3u64Lsh( context->count, 3 );
  for ( i = 0 ; i < 8 ; i++ )
  {
    bytes[i] = (u8) v3u32u64( bit_count );
    bit_count = v3u64Rsh( bit_count, 8 );
  }

  /* Add 0x80 then pad the data with zeros until 8 bytes remain. */
  pad = 0x80;
  v3_MD5Update( context, &pad, v3u64u32( 1 ) );
  pad = 0x00;
  while( ( v3u32u64( context->count ) & 0x3F ) != 56 )
  {
    v3_MD5Update ( context, &pad, v3u64u32( 1 ) );
  }

  v3_MD5Update( context, bytes, v3u64u32( 8 ) );

  /* Copy the digest into the hash little endian and clear the context */
  for ( i = 0 ; i < 4 ; i++ )
  {
    for ( j = 0 ; j < 4 ; j++ )
    {
      hash[( i * 4 ) + j] = (u8) ( context->digest[i] >> ( j * 8 ) );
    }
  }
  v3MemSet( context, v3u64u32( sizeof( v3_MD5_CONTEXT ) ), 0 );
}

/*
  ------------------------------
   SHA1 - Secure Hash Algorithm
  ------------------------------
*/

/* Some basic macros and constants. */
#define _V3_SHA_F1(x, y, z)  ( ( x & y ) | ( ~x & z ) )
#define _V3_SHA_F2(x, y, z)  ( x ^ y ^ z )
#define _V3_SHA_F3(x, y, z)  ( ( x & y) | ( x & z ) | ( y & z ) )
#define _V3_SHA_F4(x, y, z)  ( x ^ y ^ z )
#define _V3_SHA_CONST1  0x5A827999
#define _V3_SHA_CONST2  0x6ED9EBA1
#define _V3_SHA_CONST3  0x8F1BBCDC
#define _V3_SHA_CONST4  0xCA62C1D6

/* The primary SHA hashing macro */
#define _V3_SHA_FG( n ) \
  { \
    T = _V3_TRUNC32( _V3_ROTL32( A, 5 ) + _V3_SHA_F##n( B, C, D ) + E + \
      *WP++ + _V3_SHA_CONST##n ); \
    E = D; \
    D = C; \
    C = _V3_ROTL32( B, 30 ); \
    B = A; \
    A = T; \
  }

/*
  v3_SHA1Begin - Initalize a v3_SHA1_CONTEXT variable for use in a new hash.
*/
v3_SHA1_CONTEXT * v3_SHA1Begin( void )
{
  v3_SHA1_CONTEXT * context;
  u64 mem_size = v3u64u32( sizeof( v3_SHA1_CONTEXT ) );

  if( ( context = v3MemAlloc( mem_size, V3_MEM_SECURE ) ) == NULL )
    if( ( context = v3MemAlloc( mem_size, V3_MEM_NORMAL ) ) == NULL )
      return( NULL );

  _V3_SET64( context->count, 00000000, 00000000 );

  /* Load magic initialization constants. */
  context->digest[0] = 0x67452301;
  context->digest[1] = 0xEFCDAB89;
  context->digest[2] = 0x98BADCFE;
  context->digest[3] = 0x10325476;
  context->digest[4] = 0xC3D2E1F0;

  return( context );
}

/*
  v3_SHA1Update - Add data to an SHA1 hash.
*/
void v3_SHA1Update( v3_SHA1_CONTEXT * context, const u8 * buffer, u64 length )
{
  u32 bytes;
  u32 offset;

  offset = v3u32u64( context->count ) & 0x3F;
  context->count = v3u64Add( context->count, length );

  while( v3u64Gt( length, v3u64u32( 0 ) ) )
  {
    bytes = 64 - offset;
    if ( v3u64Gt( v3u64u32( bytes ), length ) )
      bytes = v3u32u64( length );

    v3MemCopy( &context->buffer[offset], buffer, v3u64u32( bytes ) );

    length = v3u64Sub( length, v3u64u32( bytes ) );
    buffer += bytes;
    offset = ( offset + bytes ) & 0x3F;

    if ( offset == 0 )
      v3_SHA1Transform( context );
  }
}

/*
  v3_SHA1Transform - The core transform routine for SHA1.
*/
void v3_SHA1Transform( v3_SHA1_CONTEXT * context )
{
  u32 i, A, B, C, D, E, T;
  u32 W[80], *WP;

  for( i = 0; i < 16; i++ )
    v3u32Load( &W[ i ], &context->buffer[ i * 4 ] );

  for ( i = 16; i < 80; i++ )
  {
    W[ i ] = W[ i - 3 ] ^ W[ i - 8 ] ^ W[ i - 14 ] ^ W[ i - 16 ];
    W[ i ] = _V3_ROTL32( W[ i ], 1 );
  }

  A = context->digest[0];
  B = context->digest[1];
  C = context->digest[2];
  D = context->digest[3];
  E = context->digest[4];
  WP = W;

  for( i = 0; i < 20; i++ )
    _V3_SHA_FG( 1 );

  for( i = 0; i < 20; i++ )
    _V3_SHA_FG( 2 );

  for( i = 0; i < 20; i++ )
    _V3_SHA_FG( 3 );

  for( i = 0; i < 20; i++ )
    _V3_SHA_FG( 4 );

  context->digest[0] = _V3_TRUNC32( context->digest[0] + A );
  context->digest[1] = _V3_TRUNC32( context->digest[1] + B );
  context->digest[2] = _V3_TRUNC32( context->digest[2] + C );
  context->digest[3] = _V3_TRUNC32( context->digest[3] + D );
  context->digest[4] = _V3_TRUNC32( context->digest[4] + E );
}

/*
  v3_SHA1Final - The final routine to retrieve the SHA1 digest.
*/
void v3_SHA1Final( u8 * hash, v3_SHA1_CONTEXT * context )
{
  u64 bit_count;
  u8 pad, bits[8];
  s8 i;

  bit_count = v3u64Lsh( context->count, 3 );
  v3u64Save( bits, &bit_count );

  /* Add 0x80 then pad the data with zeros until 8 bytes remain. */
  pad = 0x80;
  v3_SHA1Update ( context, &pad, v3u64u32( 1 ) );

  pad = 0x00;
  while( ( v3u32u64( context->count ) & 0x3F ) != 56 )
    v3_SHA1Update ( context, &pad, v3u64u32( 1 ) );

  /* Fill those 8 bytes with the length in big-endian form. */
  v3_SHA1Update ( context, bits, v3u64u32( 8 ) );

  /* Copy the digest into the hash variable and zeroize the scratch space. */
  for( i = 0; i < 5; i++ )
    v3u32Save( &hash[i * 4], (u32 *) &context->digest[i] );
  v3MemSet( context, v3u64u32( sizeof( v3_SHA1_CONTEXT ) ), 0 );
}

/* testing */

s32 v3_TestSecurity( void )
{
  ascii string[2388] = V3_TEST_TEXT_BLOCK;
  v3_HASH_TMP * work;
  v3_HASH hash;
  u32 i;

  /* MD5 hash of V3_TEST_TEXT_BLOCK */
  const v3_HASH hash_block_md5 =
  {
    {
      0x75, 0xF9, 0x7B, 0x70, 0xED, 0x49, 0x8D, 0x5C,
      0x9E, 0xF9, 0xEA, 0xA3, 0x95, 0x38, 0x18, 0x9E,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    },
    V3_HASH_MD5
  };

  /* MD5 hash of the first 1000 bytes of V3_TEST_TEXT_BLOCK */
  const v3_HASH hash_block_1000_md5 =
  {
    {
      0x91, 0x99, 0x23, 0xC5, 0x74, 0x52, 0xAA, 0xE9,
      0xC6, 0xAE, 0x45, 0x1C, 0x8A, 0xD3, 0xBD, 0x5C,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    },
    V3_HASH_MD5
  };

  /* SHA1 hash of V3_TEST_TEXT_BLOCK */
  const v3_HASH hash_block_sha1 =
  {
    {
      0xF2, 0x2B, 0x84, 0x80, 0x8F, 0xF5, 0xCD, 0x92,
      0x59, 0x49, 0xF6, 0x5E, 0xA4, 0x0A, 0x1E, 0x10,
      0x9C, 0xBD, 0x26, 0xF7, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    },
    V3_HASH_SHA1
  };

  /* SHA1 hash of the first 1000 bytes of V3_TEST_TEXT_BLOCK */
  const v3_HASH hash_block_1000_sha1 =
  {
    {
      0xED, 0x43, 0xA1, 0xC7, 0x4A, 0x43, 0x5C, 0x9F,
      0x66, 0x52, 0x3E, 0x55, 0x7F, 0x48, 0xA8, 0x71,
      0x6F, 0xCB, 0xD8, 0x7C, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    },
    V3_HASH_SHA1
  };

  /*
    MD5 test section
    ****************
    A. Hash the test block and compare.
    B. Hash 1000 bytes of the test block and compare.
  */

  /* Test section A */

  if ( ( work = v3MemAlloc( v3u64u32( sizeof( v3_HASH_TMP ) ),
    V3_MEM_NORMAL ) ) == NULL )
  {
    /* Error in memory allocation. */
    return( -1 );
  }

  if ( v3HashBegin( work, V3_HASH_MD5 ) != V3_PASS )
  {
    return( -2 );
  }

  if ( v3HashUpdate( work, (ascii *) string, v3u64u32( sizeof( string ) ) )
    != V3_PASS )
  {
    return( -3 );
  }

  if ( v3HashEnd( &hash, work ) != V3_PASS )
  {
    return( -4 );
  }

  v3MemFree( work );

  /* Compare the hash */
  if( !v3HashEq( &hash, &hash_block_md5 ) )
  {
    return( -5 );
  }

  /* Test section B */

  if ( ( work = v3MemAlloc( v3u64u32( sizeof( v3_HASH_TMP ) ),
    V3_MEM_NORMAL ) ) == NULL )
  {
    /* Error in memory allocation. */
    return( -6 );
  }

  if ( v3HashBegin( work, V3_HASH_MD5 ) != V3_PASS )
  {
    return( -7 );
  }

  if ( v3HashUpdate( work, (ascii *) string, v3u64u32( 1000 ) )
    != V3_PASS )
  {
    return( -8 );
  }

  if ( v3HashEnd( &hash, work ) != V3_PASS )
  {
    return( -9 );
  }

  v3MemFree( work );

  /* Compare the hash */
  if( !v3HashEq( &hash, &hash_block_1000_md5 ) )
  {
    return( -10 );
  }

  /* Zeroize hash */
  for ( i = 0; i < 32; i++)
  {
    hash.hash[i] = 0;
  }
  hash.type = 0;

  /*
    SHA1 test section
    *****************
    C. Hash the test block and compare.
    D. Hash 1000 bytes of the test block and compare.
  */

  /* Test section C */

  if ( ( work = v3MemAlloc( v3u64u32( sizeof( v3_HASH_TMP ) ),
    V3_MEM_NORMAL ) ) == NULL )
  {
    /* Error in memory allocation. */
    return( -11 );
  }

  if ( v3HashBegin( work, V3_HASH_SHA1 ) != V3_PASS )
  {
    return( -12 );
  }

  if ( v3HashUpdate( work, (ascii *) string, v3u64u32( sizeof( string ) ) )
    != V3_PASS )
  {
    return( -13 );
  }

  if ( v3HashEnd( &hash, work ) != V3_PASS )
  {
    return( -14 );
  }

  v3MemFree( work );

  /* Compare the hash */
  if( !v3HashEq( &hash, &hash_block_sha1 ) )
  {
    return( -15 );
  }

  /* Test section D */

  if ( ( work = v3MemAlloc( v3u64u32( sizeof( v3_HASH_TMP ) ),
    V3_MEM_NORMAL ) ) == NULL )
  {
    /* Error in memory allocation. */
    return( -16 );
  }

  if ( v3HashBegin( work, V3_HASH_SHA1 ) != V3_PASS )
  {
    return( -17 );
  }

  if ( v3HashUpdate( work, (ascii *) string, v3u64u32( 1000 ) )
    != V3_PASS )
  {
    return( -18 );
  }

  if ( v3HashEnd( &hash, work ) != V3_PASS )
  {
    return( -19 );
  }

  v3MemFree( work );

  /* Compare the hash */
  if( !v3HashEq( &hash, &hash_block_1000_sha1 ) )
  {
    return( -20 );
  }

  return( V3_PASS );
}
