/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#ifndef V3_SECURITY_H
#define V3_SECURITY_H

#include "cputypes.h"
#include "cosmmem.h"
#include "cosmtime.h"
#include "cosmtask.h"
#include "cosmfile.h"

#define V3_HASH_NONE   0
#define V3_HASH_SHA1   1  /* 160 bits, preferred for signing hashes */
#define V3_HASH_MD5    2  /* 128 bits, preferred for checksums */

#define V3_HASH_STATE_INIT   42  /* Initialized */
#define V3_HASH_STATE_DATA   43  /* Feeding */
#define V3_HASH_STATE_DONE   44  /* Finished */
#define V3_HASH_STATE_ERROR  45  /* Error, data invalid */

#define V3_HASH_ERROR_NONE    0
#define V3_HASH_ERROR_TYPE    -1 /* invalid hash type*/
#define V3_HASH_ERROR_STATE   -2 /* invalid state, wrong order */
#define V3_HASH_ERROR_MEMORY  -3 /* memory problem */
#define V3_HASH_ERROR_FATAL   -4 /* fatal problem, hash state invalid */
#define V3_HASH_ERROR_PARAM   -5 /* The user passed a NULL pointer */

#define V3_RANDOM_STATE_SIZE 1023

#if ( defined( NO_CRYPTO ) )
#define MD_DIGEST_LENGTH 16
#else
#define MD_DIGEST_LENGTH 16
  /* Yarrow */
#endif


typedef struct
{
  u8 hash[32]; /* up to 256 bits of hash, big endian, aligned left */
  u32 type;
} v3_HASH;

typedef struct
{
  void * tmp_data;
  u32 state;
  u32 type;
  u32 mode;
  v3_MUTEX lock;
} v3_HASH_TMP;

/* Hash-specific internal datatypes */

typedef struct
{
  u8 buffer[64];  /* Data Buffer (Transform input) */
  u32 digest[4];  /* Message digest (Transform output) */
  u64 count;      /* Bits of data that have been hashed */
} v3_MD5_CONTEXT;

typedef struct
{
  u8 buffer[64];  /* Data Buffer (Transform input) */
  u32 digest[5];  /* Message digest (Transform output) */
  u64 count;      /* Bits of data that have been hashed */
} v3_SHA1_CONTEXT;

/* Random state */

typedef struct
{
  u8 state[V3_RANDOM_STATE_SIZE + MD_DIGEST_LENGTH];
  u32 state_index;
  u32 state_num;
} v3_RANDOM;

/*
  Use:
  ascii string[2388] = V3_TEST_TEXT_BLOCK;
*/
#define V3_TEST_TEXT_BLOCK \
"Why shouldn't I work for the NSA? That's a tough one.  But I'll "\
"take a shot. Say I'm working at the NSA, and somebody puts a cod"\
"e on my desk, somethin' no one else can break. Maybe I take a sh"\
"ot at it and maybe I break it. And I'm real happy with myself, c"\
"us' I did my job well. But maybe that code was the location of s"\
"ome rebel army in North Africa or the Middle East and once they "\
"have that location, they bomb the village where the rebels are h"\
"iding... Fifteen hundred people that I never met, never had no p"\
"roblem with get killed. Now the politicians are sayin', \"Oh, Sen"\
"d in the marines to secure the area\" cus' they don't give a shit"\
". It won't be their kid over there, gettin' shot. Just like it w"\
"asn't them when their number got called, cus' they were pullin' "\
"a tour in the National Guard. It'll be some kid from Southie ove"\
"r there takin' shrapnel in the ass. He comes back to find that t"\
"he plant he used to work at got exported to the country he just "\
"got back from. And the guy who put the shrapnel in his ass got h"\
"is old job, cus' he'll work for fifteen cents a day and no bathr"\
"oom breaks. Meanwhile he realizes the only reason he was over th"\
"ere in the first place was so that we could install a government"\
" that would sell us oil at a good price. And of course the oil c"\
"ompanies used the little skirmish over there to scare up domesti"\
"c oil prices. A cute little ancillary benefit for them but it ai"\
"n't helping my buddy at two-fifty a gallon. They're takin' their"\
" sweet time bringin' the oil back, of course, maybe even took th"\
"e liberty of hiring an alcoholic skipper who likes to drink mart"\
"inis and fsckin' play slalom with the icebergs, it ain't too lon"\
"g 'til he hits one, spills the oil and kills all the sea life in"\
" the North Atlantic. So now my buddy's out of work. He can't aff"\
"ord to drive, so he's walking to the fuckin' job interviews, whi"\
"ch sucks because the shrapnel in his ass is givin' him chronic h"\
"emorrhoids. And meanwhile he's starvin' cus' every time he tries"\
" to get a bite to eat the only blue plate special they're servin"\
"' is North Atlantic scrod with Quaker State. So what did I think"\
"? I'm holdin' out for somethin' better. I figure fuck it, while "\
"I'm at it why not just shoot my buddy, take his job, give it to "\
"his sworn enemy, hike up gas prices, bomb a village, club a baby"\
" seal, hit the hash pipe and join the National Guard? I could be"\
"elected President.\n"

/* Hash Functions */

s32 v3HashBegin( v3_HASH_TMP * work, u32 type );
  /*
    Initialize the work structure and setup the data needed to perform the
    type of hash needed.
    Returns: V3_PASS on success, or an error code on failure.
  */

s32 v3HashUpdate( v3_HASH_TMP * work, const u8 * const data, u64 length );
  /*
    Feed length bytes of data into the work hash.
    Note: Make sure your data is in big endian format before using.
    Returns: V3_PASS on success, or an error code on failure.
  */

s32 v3HashEnd( v3_HASH * hash, v3_HASH_TMP * work );
  /*
    Take the work data and generate the final hash.
    Returns: V3_PASS on success, or an error code on failure.
  */

s32 v3HashEq( const v3_HASH * hashA, const v3_HASH * hashB );
  /*
    Check if the hashes are equal. hashA == hashB ? 1 : 0;
    Returns: 1 if the hashes are equal, or 0 if they are not equal.
  */

/* Random function */

s32 v3Random( v3_RANDOM * random, u64 length, u8 * data,
  u64 salt_length, const u8 * salt );
  /*
    Initialization of the random number generator is done by feeding it
    salt_length bytes of salt with data == NULL. Any salt data fed to the
    random generator should be cumulative, so it can be called repeatedly
    with additional seed to provide more initial randomness.

    Generate length bytes of deterministic random bits if we call
    with salt == NULL, and place them into data;

    Generate a non-deterministic random sequence if we call with salt != NULL.
    The OS will also be asked for some random data if it supports that
    feature exists and salt != NULL.

    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

/* Low level MD5 functions */

v3_MD5_CONTEXT * v3_MD5Begin( void );
  /*
    Initalize a v3_MD5_CONTEXT variable for use in a new hash. Set bit
    count to 0 and buffer to initialization constants.
    Returns: Initalized v3_MD5_CONTEXT or NULL if an error occured.
  */

void v3_MD5Update( v3_MD5_CONTEXT * context, const u8 * buffer, u64 length );
  /*
    Update context to reflect the concatenation of another buffer full of
    bytes.
    Returns: Nothing.
  */

void v3_MD5Transform( v3_MD5_CONTEXT * context );
  /*
    The core transform routine for MD5. This alters an existing MD5 hash to
    reflect the addition of 16 u32's of new data. v3_MD5Update blocks the
    data and converts bytes into u32 for this routine.
    Returns: Nothing.
  */

void v3_MD5Final ( u8 * hash, v3_MD5_CONTEXT * context );
  /*
    The final routine to retrieve the MD5 digest. Take the work data and
    generate the final hash.
    Returns: Nothing.
  */

/* Low level SHA1 functions */

v3_SHA1_CONTEXT * v3_SHA1Begin( void );
  /*
    Initalize a v3_SHA1_CONTEXT variable for use in a new hash. Set bit
    count to 0 and buffer to initialization constants.
    Returns: Initalized v3_SHA1_CONTEXT or NULL if an error occured.
  */

void v3_SHA1Update( v3_SHA1_CONTEXT * context, const u8 * buffer,
  u64 length );
  /*
    Update context to reflect the concatenation of another buffer full of
    bytes.
    Returns: Nothing.
  */

void v3_SHA1Transform( v3_SHA1_CONTEXT * context );
  /*
    The core transform routine for SHA1.
    Returns: Nothing.
  */

void v3_SHA1Final( u8 * hash, v3_SHA1_CONTEXT * context );
  /*
    The final routine to retrieve the SHA1 digest. Take the work data and
    generate the final hash.
    Returns: Nothing.
  */

s32 v3_TestSecurity( void );
  /*
    Test functions in this header.
    Returns: V3_PASS on success, or a negative number corresponding to the
      test that failed.
  */

#endif
